{****************************************************************************

                   Copyright (c) 1993,94 by Florian Klmpfl

 ****************************************************************************}

unit codegen;

  interface

    uses
       cobjects,globals,tree,asmgen,symtable,tempad,types;

    { erzeugt fr den durch p bergebenen Ausdruck Assembler }
    { und erzeugt am Ende eine Assembler-Node                }

    procedure generatecode(var p : ptree);

    { fat Konstanten im Baum zusammen, fgt Typkonvertierungen ein... }
    function do_firstpass(var p : ptree) : boolean;

    { erzeugt den eigentlichen Code }
    function do_secondpass(var p : ptree) : boolean;

    { initilisiert bzw beendet den Codegenerator }
    procedure codegeninit;
    procedure codegendone;

    { erzeugt bei bool'schen Ausdrcken die Sprnge zu true bzw falselabel }
    procedure maketojumpbool(p : ptree);

    procedure concatcopy(source,dest : treferenz;size : longint;delsource : boolean);

    { siehe Implementation }
    procedure maybe_loadesi;

    type
       tprocinfo = record
          { aktuelle Klasse }
          class : pclassdef;
          { Returntyp }
          retdef : pdef;
          { true, falls Eceptions behandelt werden sollen }
          exceptions : boolean;
          { true, falls das Unterprogramm exportiert werden soll (OS/2) }
          exported : boolean;
          { Framepointeroffset }
          framepointer : longint;
          { Self-Pointer rel. zu EBP }
          ESI_offset : longint;
          { Resultatwertoffset }
          retoffset : longint;
          { Parameteroffset }
          call_offset : longint;
       end;

    var
       { Info ber das momentan geparste Unterprogramm }
       procinfo : tprocinfo;

       { Die Nummer der Label die bei BREAK bzw CONTINUE }
       { angesprungen werden sollen }
       aktbreaklabel,aktcontinuelabel : longint;

       { truelabel wird angesprungen, wenn ein Ausdruck true ist, falselabel }
       { entsprechend                                                        }
       truelabel,falselabel : longint;

       { Nr des Labels welches zum Verlassen eines Unterprogramm }
       { angesprungen wird                                       }
       aktexitlabel : longint;

       { Nummer des Labels zu dem bei einer Exception zurck gekehrt }
       { werden soll                                                 }
       aktexceptlabel : longint;

       { Der Code der beim Eintritt in ein Unterprogramm abgearbeitet }
       { wird                                                         }
       aktentrycode : tasmlist;

       { Der Code der beim Austritt aus einem Unterprogramm }
       { abgearbeitet wird                                  }
       aktexitcode : tasmlist;

       { Der Code der beim Austritt aus einem Unterprogramm }
       { mit einer Exception abgearbeitet wird              }
       aktexceptcode : tasmlist;

       { Der Code der aktuellen Procedure }
       aktproccode : tasmlist;

       { die asmlist fr das Datensegment ist in asmgen deklariert, }
       { da sie von symtable gebraucht wird                         }

       { Assemblerliste in der der Code fuer den aktuellen Ausdruck eingefuegt wird }
       exprasmlist : tasmlist;

       { Assemblerliste mit Debuggerinfos }
       debuginfos : tasmlist;

       { Boolean, wenn eine loadn kein Assembler erzeugt hat }
       simple_loadn : boolean;

  implementation

    {$I INNR.INC}

    procedure firstpass(var p : ptree);forward;
    procedure secondpass(var p : ptree);forward;

    var
       codegenerror : boolean;

    procedure error(const t : errorconst);

      begin
         if not(codegenerror) then
           globals.error(t);
         codegenerror:=true;
      end;

    type
       firstpassproc = procedure(var p : ptree);
       secondpassproc = procedure(var p : ptree);

    function ispowerof2(value : longint;var power : longint) : boolean;

      var
         hl : longint;
         i : longint;

      begin
         hl:=1;
         ispowerof2:=true;
         for i:=0 to 31 do
           begin
              if hl=value then
                begin
                   power:=i;
                   exit;
                end;
              hl:=hl shl 1;
           end;
         ispowerof2:=false;
      end;

    procedure emit(i : tasmop;s : topsize;const t : string);

      var
         p : pasmrec;

      begin
         if (i=PUSH) or (i=PUSHAD) then pushgened:=true;
         new(p);
         p^.linenr:=aktlinenr;
         p^.infile:=inputstack;
         p^.instruc:=i;
         p^.disposing:=D_STRING;
         p^.opsize:=s;
         p^.t:=strpnew(t);
         p^.next:=nil;
         if exprasmlist.wurzel=nil then exprasmlist.wurzel:=p
           else exprasmlist.last^.next:=p;
         exprasmlist.last:=p;
      end;

    procedure emitl(i : tasmop;l : longint);

      var
         p : pasmrec;

      begin
         new(p);
{         if jumpoptimize then
           begin
              if i<>A_LABEL then registerused(p,l)
              else getinforec(l)^.defined:=p;
           end; }
         p^.linenr:=aktlinenr;
         p^.infile:=inputstack;
         p^.instruc:=i;
         p^.opsize:=S_NO;
         p^.disposing:=D_LABEL;
         p^.l:=l;
         p^.next:=nil;
         if exprasmlist.wurzel=nil then exprasmlist.wurzel:=p
           else exprasmlist.last^.next:=p;
         exprasmlist.last:=p;
      end;

    procedure maketojumpbool(p : ptree);

      begin
         if (p^.resulttype^.deftype=grunddef) and
            (pgrunddef(p^.resulttype)^.typ=bool8bit) then
           begin
              if is_constboolnode(p) then
                begin
                   if p^.value<>0 then
                     emitl(JMP,truelabel)
                   else emitl(JMP,falselabel);
                end
              else
                begin
                   case p^.location.loc of
                      LOC_REGISTER : begin
                                        emit(A_OR,S_B,regid2str[p^.location.register]+
                                          ','+regid2str[p^.location.register]);
                                        emitl(JNZ,truelabel);
                                        emitl(JMP,falselabel);
                                     end;
                      LOC_MEM,LOC_REFERENZ : begin
                                        emit(CMP,S_B,'$0,'+getreferenzstring(p^.location.referenz));
                                        emitl(JNZ,truelabel);
                                        emitl(JMP,falselabel);
                                     end;
                      LOC_FLAGS : begin
                                     emitl(flag_2_jmp[p^.location.resflags],truelabel);
                                     emitl(JMP,falselabel);
                                  end;
                   end;
                end;
           end
         else
           error(type_mismatch);
      end;

    procedure emitoverflowcheck;

      var
         hl : longint;

      begin
         if check_overflow then
           begin
              hl:=getlabel;
              emitl(JNO,hl);
              emit(CALL,S_NO,'RE_OVERFLOW');
              maybe_loadesi;
              emitl(A_LABEL,hl);
           end;
      end;

    procedure emitpushreferenzaddr(const ref : treferenz);

      begin
         if ref.isintvalue then
           emit(PUSH,S_L,'$'+tostr(ref.offset))
         else
           begin
              if (ref.base=R_NO) and (ref.index=R_NO) then
                emit(PUSH,S_L,'$'+getreferenzstring(ref))
              else if (ref.base=R_NO) and (ref.index<>R_NO) and
                 (ref.offset=0) and (ref.scalefactor=0) and (ref.symbol=nil) then
                emit(PUSH,S_L,regid2str[ref.index])
              else if (ref.base<>R_NO) and (ref.index=R_NO) and
                 (ref.offset=0) and (ref.symbol=nil) then
                emit(PUSH,S_L,regid2str[ref.base])
              else
                begin
                   emit(LEA,S_L,getreferenzstring(ref)+',%edi');
                   emit(PUSH,S_L,'%edi');
                end;
           end;
        end;

    { ldt falls notwendig, ESI nach einem Call neu }

    procedure maybe_loadesi;

      begin
         if assigned(procinfo.class) then
           emit(MOV,S_NO,tostr(procinfo.ESI_offset)+'(%ebp),%esi');
      end;

    function isconvertable(def_from,def_to : pdef;var doconv : tconverttype;fromtreetype : ttreetyp) : boolean;

      { from_is_cstring mu true sein, wenn def_from die Definition einer }
      { Stringkonstanten ist, ntig wegen der Konvertierung von String-   }
      { konstante zu nullterminiertem String                              }

      { Hilfsliste: u8bit,s32bit,s64real,uvoid,
                    bool8bit,uchar,s8bit,s16bit,u16bit }

      const
         grunddefconverts : array[u8bit..u16bit,u8bit..u16bit] of tconverttype =
         ((tc_equal,tc_u8bit_2_s32bit,tc_int_2_real,tc_not_possible,
           tc_not_possible,tc_not_possible,tc_equal,tc_u8bit_2_s16bit,tc_u8bit_2_u16bit),

          (tc_s32bit_2_u8bit,tc_equal,tc_int_2_real,tc_not_possible,
           tc_not_possible,tc_not_possible,tc_s32bit_2_s8bit,tc_s32bit_2_s16bit,tc_s32bit_2_u16bit),

          (tc_not_possible,tc_not_possible,tc_equal,tc_not_possible,
           tc_not_possible,tc_not_possible,tc_not_possible,tc_not_possible,tc_not_possible),

          (tc_not_possible,tc_not_possible,tc_not_possible,tc_not_possible,
           tc_not_possible,tc_not_possible,tc_not_possible,tc_not_possible,tc_not_possible),

          (tc_not_possible,tc_not_possible,tc_not_possible,tc_not_possible,
           tc_equal,tc_not_possible,tc_not_possible,tc_not_possible,tc_not_possible),

          (tc_not_possible,tc_not_possible,tc_not_possible,tc_not_possible,
           tc_not_possible,tc_equal,tc_not_possible,tc_not_possible,tc_not_possible),

          (tc_equal,tc_s8bit_2_s32bit,tc_int_2_real,tc_not_possible,
           tc_not_possible,tc_not_possible,tc_equal,tc_s8bit_2_s16bit,tc_s8bit_2_u16bit),

          (tc_s16bit_2_u8bit,tc_s16bit_2_s32bit,tc_int_2_real,tc_not_possible,
           tc_not_possible,tc_not_possible,tc_s16bit_2_s8bit,tc_equal,tc_equal),

          (tc_u16bit_2_u8bit,tc_u16bit_2_s32bit,tc_int_2_real,tc_not_possible,
           tc_not_possible,tc_not_possible,tc_u16bit_2_s8bit,tc_equal,tc_equal));

      var
         b : boolean;

      begin
         b:=false;
         if (def_from^.deftype=grunddef) and (def_to^.deftype=grunddef) then
           begin
              doconv:=grunddefconverts[pgrunddef(def_from)^.typ,pgrunddef(def_to)^.typ];
              if doconv<>tc_not_possible then
                b:=true;
           end
         else if (def_from^.deftype=pointerdef) and (def_to^.deftype=arraydef) and
            (parraydef(def_to)^.lowrange=0) and
            is_equal(ppointerdef(def_from)^.definition,
              parraydef(def_to)^.definition) then
           begin
              doconv:=tc_pointer_to_array;
              b:=true;
           end
         else if (def_from^.deftype=arraydef) and (def_to^.deftype=pointerdef) and
            (parraydef(def_from)^.lowrange=0) and
            is_equal(parraydef(def_from)^.definition,
              ppointerdef(def_to)^.definition) then
           begin
              doconv:=tc_array_to_pointer;
              b:=true;
           end
         else
           if (def_from^.deftype=stringdef) and (def_to^.deftype=stringdef) then
             begin
                doconv:=tc_string_to_string;
                b:=true;
             end
         else
           { Char nach String: }
           if (def_from^.deftype=grunddef) and
              (pgrunddef(def_from)^.typ=uchar) and
              (def_to^.deftype=stringdef) then
             begin
                doconv:=tc_char_to_string;
                b:=true;
             end
         else
           { Stringkonstante zu nullterm. Stringkonstante }
           if (fromtreetype=stringconstn) and
                (
                  (def_to^.deftype=pointerdef) and
                  (ppointerdef(def_to)^.definition^.deftype=grunddef) and
                  (pgrunddef(ppointerdef(def_to)^.definition)^.typ=uchar)
                )
             then
             begin
                doconv:=tc_cstring_charpointer;
                b:=true;
             end
         else
           if (fromtreetype=stringconstn) and
                (
                  (def_to^.deftype=arraydef) and (parraydef(def_to)^.lowrange=0) and
                  (parraydef(def_to)^.definition^.deftype=grunddef) and
                  (pgrunddef(parraydef(def_to)^.definition)^.typ=uchar)
                )
             then
             begin
                doconv:=tc_cstring_chararray;
                b:=true;
             end
         else
           if (fromtreetype=ordconstn) and (def_from^.deftype=grunddef) and
              (pgrunddef(def_from)^.typ=uchar) and
                (
                  (def_to^.deftype=arraydef) and (parraydef(def_to)^.lowrange=0) and
                  (parraydef(def_to)^.definition^.deftype=grunddef) and
                  (pgrunddef(parraydef(def_to)^.definition)^.typ=uchar)
                )
             then
             begin
                doconv:=tc_cchar_chararray;
                b:=true;
             end;
         isconvertable:=b;
      end;

    type
       tpushed = array[R_EAX..R_EDX] of boolean;

    procedure pushusedregisters(var pushed : tpushed;b : byte);

      var
         r : tregister;

      begin
         for r:=R_EAX to R_EDX do
           begin
              pushed[r]:=false;
              { wenn das Register von dem aufgerufenen Unterprogramm }
              { verwendet wird:                                      }
              if ((b and ($80 shr byte(r)))<>0) then
                begin
                   { und momentan belegt ist }
                   if not(r in unused) then
                     begin
                        { dann retten... }
                        emit(PUSH,S_L,regid2str[r]);
                        unused:=unused-[r];
                        pushed[r]:=true;
                     end;
                   (*
                   { wenn nicht belegt, wird es nicht gerettet, d.h. es }
                   { wird von aktprocsym^.definition modifiziert        }
                   else
                     begin
                        { natrlich nur, wenn wir uns nicht im Hauptblock }
                        { befinden                                        }
                        if assigned(aktprocsym) then
                          aktprocsym^.definition^.usedregisters:=
                            aktprocsym^.definition^.usedregisters or ($80 shr byte(r));
                     end;
                   *)
                end;
           end;
      end;

    procedure genconstadd(size : topsize;l : longint;const str : string);

      begin
         if l=0 then
         else if l=1 then
           emit(A_INC,size,str)
         else if l=-1 then
           emit(A_INC,size,str)
         else
           emit(ADD,size,'$'+tostr(l)+','+str);
      end;

    procedure popusedregisters(const pushed : tpushed);

      var
         r : tregister;

      begin
         for r:=R_EDX downto R_EAX do
           if pushed[r] then
             emit(POP,S_L,regid2str[r]);
      end;

    procedure concatcopy(source,dest : treferenz;size : longint;delsource : boolean);

      var
         ecxpushed : boolean;
         helpsize : longint;
         i : byte;
         reg8,reg32 : tregister;
         swap : boolean;

      begin

         if delsource then
           del_referenz(source);

         { ab 12 Bytes wird movs verwendet }
         if (size<=8) or (not(littlesize) and (size<=12)) then
           begin
              helpsize:=size div 4;
              for i:=1 to helpsize do
                begin
                   emit(MOV,S_L,getreferenzstring(source)+',%edi');
                   emit(MOV,S_L,'%edi,'+getreferenzstring(dest));
                   inc(source.offset,4);
                   inc(dest.offset,4);
                   dec(size,4);
                end;
              if size>1 then
                begin
                   emit(MOV,S_W,getreferenzstring(source)+',%di');
                   emit(MOV,S_W,'%di,'+getreferenzstring(dest));
                   inc(source.offset,2);
                   inc(dest.offset,2);
                   dec(size,2);
                end;
              if size>0 then
                begin

                   { nun ein 8-Bit Register suchen }
                   swap:=false;
                   if R_EAX in unused then reg8:=R_AL
                   else if R_EBX in unused then reg8:=R_BL
                   else if R_ECX in unused then reg8:=R_CL
                   else if R_EDX in unused then reg8:=R_DL
                   else
                      begin
                         swap:=true;

                         { wir mssen nur 3 Register untersuchen, da dann }
                         { immer eines nicht Index oder Base ist          }
                         if (dest.base<>R_EAX) and (dest.index<>R_EAX) then
                           begin
                              reg8:=R_AL;
                              reg32:=R_EAX;
                           end
                         else if (dest.base<>R_EBX) and (dest.index<>R_EBX) then
                           begin
                              reg8:=R_BL;
                              reg32:=R_EBX;
                           end
                         else if (dest.base<>R_ECX) and (dest.index<>R_ECX) then
                           begin
                              reg8:=R_CL;
                              reg32:=R_ECX;
                           end;
                      end;
                   if swap=true then
                     { war frher XCHG, natrlich Quatsch }
                     emit(MOV,S_L,regid2str[reg32]+',%edi');
                   emit(MOV,S_B,getreferenzstring(source)+','+regid2str[reg8]);
                   emit(MOV,S_B,regid2str[reg8]+','+getreferenzstring(dest));
                   if swap=true then
                     emit(MOV,S_L,'%edi,'+regid2str[reg32]);
                end;
           end
         else
           begin
              emit(LEA,S_L,getreferenzstring(source)+',%esi');
              emit(LEA,S_L,getreferenzstring(dest)+',%edi');
              del_referenz(dest);
              if not(R_ECX in unused) then
                begin
                   emit(PUSH,S_L,'%ecx');
                   ecxpushed:=true;
                end
              else ecxpushed:=false;
              emit(CLD,S_NO,'');
              if littlesize then
                begin

                   if size mod 4=0 then
                     begin
                        emit(MOV,S_L,'$'+tostr(size div 4)+',%ecx');
                        emit(REP,S_NO,'');
                        emit(MOVS,S_L,'');
                     end
                   else if size mod 2=0 then
                     begin
                        emit(MOV,S_L,'$'+tostr(size div 2)+',%ecx');
                        emit(REP,S_NO,'');
                        emit(MOVS,S_W,'');
                     end
                   else
                     begin
                        emit(MOV,S_L,'$'+tostr(size)+',%ecx');
                        emit(REP,S_NO,'');
                        emit(MOVS,S_B,'');
                     end;
                end
              else
                begin
                   helpsize:=size-size mod 4;
                   size:=size mod 4;
                   emit(MOV,S_L,'$'+tostr(helpsize div 4)+',%ecx');
                   emit(REP,S_NO,'');
                   emit(MOVS,S_L,'');
                   if size>1 then
                     begin
                        dec(size,2);
                        emit(MOVS,S_W,'');
                     end;
                   if size=1 then
                     emit(MOVS,S_B,'');
                end;
              if ecxpushed then
                emit(POP,S_L,'%ecx');

              { SELF-Referenz wieder laden }
              if assigned(procinfo.class) then
                emit(MOV,S_NO,tostr(procinfo.ESI_offset)+'(%ebp),%esi');
           end;
      end;

    procedure copystring(const dref,sref : treferenz;len : byte);

      var
         pushed : tpushed;

      begin
         pushusedregisters(pushed,$ff);
         emitpushreferenzaddr(dref);
         emitpushreferenzaddr(sref);
         emit(PUSH,S_L,'$'+tostr(len));
         emit(CALL,S_NO,'STRCOPY');
         maybe_loadesi;
         popusedregisters(pushed);
      end;

    { berechnet fr einen binren Operator die bentigten }
    { Register                                            }

    procedure calcregisters(p : ptree;r32,fpu : word);

      begin
         p^.registers32:=p^.left^.registers32;
         if p^.right^.registers32>p^.registers32 then
           p^.registers32:=p^.right^.registers32;

         p^.registersfpu:=p^.left^.registersfpu;
         if p^.right^.registersfpu>p^.registersfpu then
           p^.registersfpu:=p^.right^.registersfpu;

         { Nur wenn links und rechts ein Unterschied < bentige Anzahl ist, }
         { wird ein zustzliches Register bentigt, da es dann keinen       }
         { schwierigeren Ast gibt, welcher erst ausgewertet werden kann     }

         if (abs(p^.left^.registers32-p^.right^.registers32)<r32) then inc(p^.registers32,r32);
         if (abs(p^.left^.registersfpu-p^.right^.registersfpu)<fpu) then inc(p^.registersfpu,fpu);

         { Fehlermeldung wenn mehr als 8 FPU-Register bentigt werden }
         if p^.registersfpu>8 then error(too_complex_expr);
      end;

    function both_rm(p : ptree) : boolean;

      begin
         if ((p^.left^.location.loc=LOC_MEM) or
           (p^.left^.location.loc=LOC_REFERENZ))
           and ((p^.right^.location.loc=LOC_MEM) or
           (p^.right^.location.loc=LOC_REFERENZ)) then
           both_rm:=true else both_rm:=false;
      end;

    procedure firstcomplex(p : ptree);

      var
         hp : ptree;

      begin
         { boolsches AND und OR immer von links nach rechts auswerten }
         if ((p^.treetype=orn) or (p^.treetype=andn)) and
           (p^.left^.resulttype^.deftype=grunddef) and
           (pgrunddef(p^.left^.resulttype)^.typ=bool8bit) then
           exit;

         if (p^.left^.registers32<p^.right^.registers32)

           { folgende Abfrage ist sinnvoll, da  }
           { selten alle 4 Register gebraucht werden und   }
           { damit erreicht wird, da der zustzliche Code }
           { bei vertauschten nichtkommutativen Operatoren }
           { entfllt                                      }
           and (p^.right^.registers32<=4) then
           begin
              hp:=p^.left;
              p^.left:=p^.right;
              p^.right:=hp;
              p^.swaped:=true;
           end
         else p^.swaped:=false;
      end;

    procedure restore(p : ptree);

      var
         hregister :  tregister;

      begin
         hregister:=getregister32;
         emit(POP,S_L,regid2str[hregister]);
         if p^.location.loc=LOC_REGISTER then
           begin
              p^.location.register:=hregister;
           end
         else
           begin
              clear_referenz(p^.location.referenz);
              p^.location.referenz.index:=hregister;
           end;
      end;

    function maybe_push(needed : byte;p : ptree) : boolean;

      var
         pushed : boolean;
         hregister : tregister;

      begin
         if needed>usablereg32 then
           begin
              if (p^.location.loc=LOC_REGISTER) then
                begin
                   pushed:=true;
                   emit(PUSH,S_L,regid2str[p^.location.register]);
                   ungetregister32(p^.location.register);
                end
              else if (p^.location.loc=LOC_MEM) or
                (p^.location.loc=LOC_REFERENZ) then
                  begin
                     hregister:=getregister32;
                     emit(LEA,S_L,getreferenzstring(p^.location.referenz)+','
                       +regid2str[hregister]);
                     emit(PUSH,S_L,regid2str[hregister]);
                     ungetregister32(hregister);
                     del_referenz(p^.location.referenz);
                     pushed:=true;
                  end
              else pushed:=false;
           end
         else pushed:=false;
         maybe_push:=pushed;
      end;

    procedure firsterror(var p : ptree);far;

      begin
         p^.error:=true;
         codegenerror:=true;
      end;

    procedure seconderror(var p : ptree);far;

      begin
         p^.error:=true;
         codegenerror:=true;
      end;

    procedure firstload(var p : ptree);far;

      begin
         p^.location.loc:=LOC_REFERENZ;
         p^.registers32:=0;
         p^.registersfpu:=0;
         clear_referenz(p^.location.referenz);
         case p^.symtableentry^.typ of
            varsym :
                begin
                   p^.resulttype:=pvarsym(p^.symtableentry)^.definition;
                   if ((p^.symtable^.symtabletype and $c000)<>0) and
                      (lexlevel>(p^.symtable^.symtabletype and $3fff)) then
                     begin
                        { sollte sich die Variable in einem anderen Stackframe       }
                        { befinden, so brauchen wir ein Register zum Dereferenzieren }
                        if (p^.symtable^.symtabletype and $3fff)<>0 then
                          p^.registers32:=1;
                     end;
                   if (pvarsym(p^.symtableentry)^.varspez=vs_const) then
                     p^.location.loc:=LOC_MEM;
                   { Bei einem Call by Referenz brauchen wir ein Register }
                   if (pvarsym(p^.symtableentry)^.varspez=vs_var) or
                      ((pvarsym(p^.symtableentry)^.varspez=vs_const) and
                       (
                         (pvarsym(p^.symtableentry)^.definition^.deftype=stringdef) or
                         (pvarsym(p^.symtableentry)^.definition^.deftype=arraydef) or
                         (pvarsym(p^.symtableentry)^.definition^.deftype=recorddef) or
                         (pvarsym(p^.symtableentry)^.definition^.deftype=classdef)
                       )
                      ) then
                     p^.registers32:=1;
                end;
            typedconstsym :
                p^.resulttype:=ptypedconstsym(p^.symtableentry)^.definition;
            procsym :
                begin
                   if assigned(pprocsym(p^.symtableentry)^.definition^.nextoverloaded) then
                     error(no_overloaded_procvars);
                   p^.resulttype:=pprocsym(p^.symtableentry)^.definition;
                end;
            else internalerror(3);
         end;
      end;

    procedure secondload(var p : ptree);far;

      var
         hregister : tregister;
         symtabletype,i : longint;

      begin
         simple_loadn:=true;
         clear_referenz(p^.location.referenz);
         case p^.symtableentry^.typ of
              varsym :
                 begin
                    hregister:=R_NO;
                    symtabletype:=p^.symtable^.symtabletype;
                    { erst lokale und Uebergabevariablen behandeln }
                    if (symtabletype and $c000)<>0 then
                      begin
                         p^.location.referenz.base:=R_EBP;
                         p^.location.referenz.offset:=pvarsym(p^.symtableentry)^.adresse;
                         if (symtabletype and $8000<>0) then
                           p^.location.referenz.offset:=-p^.location.referenz.offset;
                         if (symtabletype and $4000<>0) then
                           inc(p^.location.referenz.offset,longint(p^.symtable^.name));
                         if (lexlevel>(p^.symtable^.symtabletype and $3fff)) then
                           begin
                              hregister:=getregister32;
                              emit(MOV,S_L,tostr(procinfo.framepointer)+'(%ebp),'+regid2str[hregister]);
                              simple_loadn:=false;
                              i:=lexlevel-1;
                              while i>(p^.symtable^.symtabletype and $3fff) do
                                begin
                                   if procinfo.exceptions then
                                     emit(MOV,S_L,'12('+regid2str[hregister]+'),'+regid2str[hregister])
                                   else
                                     emit(MOV,S_L,'8('+regid2str[hregister]+'),'+regid2str[hregister]);
                                   dec(i);
                                end;
                              p^.location.referenz.base:=hregister;
                           end;
                      end
                    else
                      case p^.symtable^.symtabletype of
                         staticsymtable : begin
                                             p^.location.referenz.symbol:=stringdup('_'+p^.symtableentry^.name);
                                          end;
                         unitsymtable,globalsymtable : begin
                                             p^.location.referenz.symbol:=stringdup('U_'+
                                             p^.symtable^.name^+'_'+p^.symtableentry^.name);
                                          end;
                         objectsymtable : begin
                                             p^.location.referenz.base:=R_ESI;
                                             p^.location.referenz.offset:=pvarsym(p^.symtableentry)^.adresse;
                                          end;
                         {!!!!!!!!!!! andere Symboltabellenarten bearbeiten}
                      end;

                   { falls Call by Referenz, dann berechenen: }
                   if (pvarsym(p^.symtableentry)^.varspez=vs_var) or
                      ((pvarsym(p^.symtableentry)^.varspez=vs_const) and
                       (
                         (pvarsym(p^.symtableentry)^.definition^.deftype=stringdef) or
                         (pvarsym(p^.symtableentry)^.definition^.deftype=arraydef) or
                         (pvarsym(p^.symtableentry)^.definition^.deftype=recorddef) or
                         (pvarsym(p^.symtableentry)^.definition^.deftype=classdef)
                       )
                      ) then
                     begin
                        simple_loadn:=false;
                        if hregister=R_NO then
                          hregister:=getregister32;
                        emit(MOV,S_L,getreferenzstring(p^.location.referenz)+','+regid2str[hregister]);
                        clear_referenz(p^.location.referenz);
                        p^.location.referenz.base:=hregister;
                     end;
                 end;
              procsym:
                 begin
                    {!!!!! Vorsicht, auch virtuelle Methoden bearbeiten}
                    p^.location.referenz.symbol:=
                      stringdup(pprocsym(p^.symtableentry)^.definition^.mangledname);
                 end;
              typedconstsym :
                 begin
                    p^.location.referenz.symbol:=stringdup(
                      'TC_'+ptypedconstsym(p^.symtableentry)^.prefix^+'_'+
                      +p^.symtableentry^.name);
                 end;
              else internalerror(4);
         end;
      end;

    procedure firstadd(var p : ptree);far;

      var
         lt,rt : ttreetyp;
         t : ptree;
         rv,lv : longint;
         rvd,lvd : double;
         rd,ld : pdef;
         s1,s2 : string;
         concatstrings : boolean;

      begin
         { erst die beiden ste bearbeiten }
         firstpass(p^.left);
         firstpass(p^.right);

         if codegenerror then
           exit;

         { Konstanten zusammemfassen }
         lt:=p^.left^.treetype;
         rt:=p^.right^.treetype;
         if is_constintnode(p^.left) and
           is_constintnode(p^.right) then
           begin
              lv:=p^.left^.value;
              rv:=p^.right^.value;
              case p^.treetype of
                 addn : t:=genordinalconstnode(lv+rv,s32bitdef);
                 subn : t:=genordinalconstnode(lv-rv,s32bitdef);
                 muln : t:=genordinalconstnode(lv*rv,s32bitdef);
                 xorn : t:=genordinalconstnode(lv xor rv,s32bitdef);
                 orn : t:=genordinalconstnode(lv or rv,s32bitdef);
                 andn : t:=genordinalconstnode(lv and rv,s32bitdef);
                 ltn : t:=genordinalconstnode(ord(lv<rv),booldef);
                 lten : t:=genordinalconstnode(ord(lv<=rv),booldef);
                 gtn : t:=genordinalconstnode(ord(lv>rv),booldef);
                 gten : t:=genordinalconstnode(ord(lv>=rv),booldef);
                 equaln : t:=genordinalconstnode(ord(lv=rv),booldef);
                 unequaln : t:=genordinalconstnode(ord(lv<>rv),booldef);
                 else
                   error(type_mismatch);
              end;
              disposetree(p);
              p:=t;
              exit;
           end
         else
           { Realkonstanten: }
           if (lt=realconstn) and (rt=realconstn) then
           begin
              lvd:=p^.left^.value;
              rvd:=p^.right^.value;
              case p^.treetype of
                 addn : t:=genrealconstnode(lvd+rvd);
                 subn : t:=genrealconstnode(lvd-rvd);
                 muln : t:=genrealconstnode(lvd*rvd);
                 ltn : t:=genordinalconstnode(ord(lvd<rvd),booldef);
                 lten : t:=genordinalconstnode(ord(lvd<=rvd),booldef);
                 gtn : t:=genordinalconstnode(ord(lvd>rvd),booldef);
                 gten : t:=genordinalconstnode(ord(lvd>=rvd),booldef);
                 equaln : t:=genordinalconstnode(ord(lvd=rvd),booldef);
                 unequaln : t:=genordinalconstnode(ord(lvd<>rvd),booldef);
                 else
                   error(type_mismatch);
              end;
              disposetree(p);
              p:=t;
              exit;
           end;
         concatstrings:=false;
         if (lt=ordconstn) and (rt=ordconstn) and
           (p^.left^.resulttype^.deftype=grunddef) and
           (pgrunddef(p^.left^.resulttype)^.typ=uchar) and
           (p^.right^.resulttype^.deftype=grunddef) and
           (pgrunddef(p^.right^.resulttype)^.typ=uchar) then
           begin
              s1:=char(p^.left^.value);
              s2:=char(p^.right^.value);
              concatstrings:=true;
           end
         else if (lt=stringconstn) and (rt=ordconstn) and
           (p^.right^.resulttype^.deftype=grunddef) and
           (pgrunddef(p^.right^.resulttype)^.typ=uchar) then
           begin
              s1:=pstring(p^.left^.value)^;
              s2:=char(p^.right^.value);
              concatstrings:=true;
           end
         else if (lt=ordconstn) and (rt=stringconstn) and
           (p^.left^.resulttype^.deftype=grunddef) and
           (pgrunddef(p^.left^.resulttype)^.typ=uchar) then
           begin
              s1:=char(p^.left^.value);
              s2:=pstring(p^.right^.value)^;
              concatstrings:=true;
           end
         else if (lt=stringconstn) and (rt=stringconstn) then
           begin
              s1:=pstring(p^.left^.value)^;
              s2:=pstring(p^.right^.value)^;
              concatstrings:=true;
           end;

         if concatstrings then
           begin
              case p^.treetype of
                 addn : t:=genstringconstnode(s1+s2);
                 ltn : t:=genordinalconstnode(ord(s1<s2),booldef);
                 lten : t:=genordinalconstnode(ord(s1<=s2),booldef);
                 gtn : t:=genordinalconstnode(ord(s1>s2),booldef);
                 gten : t:=genordinalconstnode(ord(s1>=s2),booldef);
                 equaln : t:=genordinalconstnode(ord(s1=s2),booldef);
                 unequaln : t:=genordinalconstnode(ord(s1<>s2),booldef);
              end;
              disposetree(p);
              p:=t;
              exit;
           end;
         rd:=p^.right^.resulttype;
         ld:=p^.left^.resulttype;

         { wenn beides Boolean: }
         if ((ld^.deftype=grunddef) and
            (pgrunddef(ld)^.typ=bool8bit)) and
            ((rd^.deftype=grunddef) and
            (pgrunddef(rd)^.typ=bool8bit)) then
           begin
              if (p^.treetype=andn) or (p^.treetype=orn) then
                begin
                   calcregisters(p,0,0);
                   p^.location.loc:=LOC_JUMP;
                end
              else if (p^.treetype=unequaln) or (p^.treetype=equaln) then
                begin
                   calcregisters(p,1,0);
                   p^.location.loc:=LOC_FLAGS;
                   p^.resulttype:=booldef;
                end
              else error(type_mismatch);
           end
         { wenn beides vom Char dann keine Konvertiereung einfuegen }
         else if ((rd^.deftype=grunddef) and (pgrunddef(rd)^.typ=uchar)) and
            ((ld^.deftype=grunddef) and (pgrunddef(ld)^.typ=uchar)) then
            begin
               calcregisters(p,1,0);
            end

         { wenn links String und rechts Char, dann Char konvertieren }
         else if ((rd^.deftype=stringdef) and
                 ((ld^.deftype=grunddef) and (pgrunddef(ld)^.typ=uchar))) or
                 ((ld^.deftype=stringdef) and
                 ((rd^.deftype=grunddef) and (pgrunddef(rd)^.typ=uchar))) then
           begin
              if ((rd^.deftype=grunddef) and (pgrunddef(rd)^.typ=uchar)) then
                p^.right:=gentypeconvnode(p^.right,cstringdef)
                else p^.left:=gentypeconvnode(p^.left,cstringdef);
              firstpass(p^.left);
              firstpass(p^.right);
              calcregisters(p,0,0);
              p^.location.loc:=LOC_MEM;
           end
         else
           if ((rd^.deftype=stringdef) and (ld^.deftype=stringdef)) then
           { nichts tun, ist aber OK }
         { sollte eine Realzahl dabei sein, beide nach s64bitreal konvertieren }
         else
            if ((rd^.deftype=grunddef) and (pgrunddef(rd)^.typ=s64real)) or
            ((ld^.deftype=grunddef) and (pgrunddef(ld)^.typ=s64real)) then
           begin
              p^.right:=gentypeconvnode(p^.right,cs64realdef);
              p^.left:=gentypeconvnode(p^.left,cs64realdef);
              firstpass(p^.left);
              firstpass(p^.right);
              calcregisters(p,0,1);
              p^.location.loc:=LOC_FPUSTACK;
           end
         { Pointervergleiche und Subtraktion: }
         else if (rd^.deftype=pointerdef) and (ld^.deftype=pointerdef) and
           is_equal(rd,ld) then
           begin
              calcregisters(p,1,0);
              p^.location.loc:=LOC_REGISTER;
              case p^.treetype of
                 equaln,unequaln : ;
                 ltn,lten,gtn,gten : if exprlevel<1 then
                                       error(type_mismatch);
                 subn : begin
                           if exprlevel<1 then
                             error(type_mismatch);
                           p^.resulttype:=s32bitdef;
                           exit;
                        end;
                 else error(type_mismatch);
              end;
           end
         else if (rd^.deftype=pointerdef) then
           begin
              p^.location.loc:=LOC_REGISTER;
              p^.left:=gentypeconvnode(p^.left,s32bitdef);
              firstpass(p^.left);
              calcregisters(p,1,0);
              if p^.treetype=addn then
                begin
                   if exprlevel<1 then
                     error(type_mismatch);
                end
              else error(type_mismatch);
           end
         else if (ld^.deftype=aufzaehldef) and (rd^.deftype=aufzaehldef)
            and (is_equal(ld,rd)) then
           begin
              calcregisters(p,1,0);
              case p^.treetype of
                 equaln,unequaln,
                 ltn,lten,gtn,gten : ;
                 else error(type_mismatch);
              end;
           end
         else if (ld^.deftype=pointerdef) then
           begin
              p^.location.loc:=LOC_REGISTER;
              p^.right:=gentypeconvnode(p^.right,s32bitdef);
              firstpass(p^.right);
              calcregisters(p,1,0);
              case p^.treetype of
                 addn,subn : if exprlevel<1 then
                               error(type_mismatch);
                 else error(type_mismatch);
              end;
           end
         { sonst immer nach 32 Bit-Int konvertieren }
         else
           begin
              p^.right:=gentypeconvnode(p^.right,s32bitdef);
              p^.left:=gentypeconvnode(p^.left,s32bitdef);
              firstpass(p^.left);
              firstpass(p^.right);
              calcregisters(p,1,0);
              p^.location.loc:=LOC_REGISTER;
           end;

         if codegenerror then
           exit;

         { Resultatyp feststellen; bei Vergleich Boolean }
         case p^.treetype of
            ltn,lten,gtn,gten,equaln,unequaln : begin
                                                   p^.resulttype:=booldef;
                                                   p^.location.loc:=LOC_FLAGS;
                                                end;
            else p^.resulttype:=p^.left^.resulttype;
         end;
      end;

    procedure secondadd(var p : ptree);far;

    { (wird auch fr xor,and,"mul","sub",or und Vergleichsoperatoren }
    { verwendet)                                                     }

      var
         swapp : ptree;
         hregister : tregister;
         pushed,mboverflow,cmpop : boolean;
         op : tasmop;
         swapl : tlocation;
         pushedregs : tpushed;
         flags : tresflags;
         otl,ofl,power : longint;
         href : treferenz;
         opsize : topsize;

      label
         do_normal;

      begin
         opsize:=S_L;
         { den komplexeren Operanden berechnen }
         firstcomplex(p);
         { bool'sche Ausdruecke extra behandeln: }
         if ((p^.left^.resulttype^.deftype=grunddef) and
            (pgrunddef(p^.left^.resulttype)^.typ=bool8bit)) then
{            ((p^.right^.resulttype^.deftype=grunddef) and
            (pgrunddef(p^.right^.resulttype)^.typ=bool8bit)) then }
           begin
              if (p^.treetype=andn) or (p^.treetype=orn) then
                begin
                   p^.location.loc:=LOC_JUMP;
                   cmpop:=false;
                   case p^.treetype of
                     andn : begin
                               otl:=truelabel;
                               truelabel:=getlabel;
                               secondpass(p^.left);
                               maketojumpbool(p^.left);
                               emitl(A_LABEL,truelabel);
                               truelabel:=otl;
                            end;
                     orn : begin
                              ofl:=falselabel;
                              falselabel:=getlabel;
                              secondpass(p^.left);
                              maketojumpbool(p^.left);
                              emitl(A_LABEL,falselabel);
                              falselabel:=ofl;
                           end;
                     else error(type_mismatch);
                   end;
                  secondpass(p^.right);
                  maketojumpbool(p^.right);
                end
{              else if (p^.treetype=unequaln) or (p^.treetype=equaln) then
                begin
                   opsize:=S_B;
                   if p^.left^.treetype=boolconstn then
                     begin
                        swapp:=p^.right;
                        p^.right:=p^.left;
                        p^.left:=swapp;
                        p^.swaped:=not(p^.swaped);
                     end;
                   secondpass(p^.left);
                   p^.location:=p^.left^.location;
                   { sind zuwenig Register frei?
                   pushed:=maybe_push(p^.right^.registers32,p);
                   secondpass(p^.right);
                   if pushed then restore(p);
                   goto do_normal;
                end }
              else error(type_mismatch);
           end
         else
           begin
              { falls eine Konstante dabei, so kommt diese nach rechts }
              if p^.left^.treetype=ordconstn then
                begin
                   swapp:=p^.right;
                   p^.right:=p^.left;
                   p^.left:=swapp;
                   p^.swaped:=not(p^.swaped);
                end;
              secondpass(p^.left);
              p^.location:=p^.left^.location;
              { sind zuwenig Register frei? }
              pushed:=maybe_push(p^.right^.registers32,p);
              secondpass(p^.right);
              if pushed then restore(p);
              if (p^.left^.resulttype^.deftype=pointerdef) or
                 (p^.right^.resulttype^.deftype=pointerdef) or
                 (p^.left^.resulttype^.deftype=aufzaehldef) or
                 ((p^.left^.resulttype^.deftype=grunddef) and
                 (pgrunddef(p^.left^.resulttype)^.typ=s32bit)) or
                 ((p^.right^.resulttype^.deftype=grunddef) and
                 (pgrunddef(p^.right^.resulttype)^.typ=s32bit)) then
                begin
           do_normal:
                   mboverflow:=false;
                   cmpop:=false;
                   case p^.treetype of
                      addn : begin
                                op:=ADD;
                                mboverflow:=true;
                             end;
                      muln : begin
                                op:=IMUL;
                                mboverflow:=true;
                             end;
                      subn : begin
                                op:=SUB;
                                mboverflow:=true;
                             end;
                      ltn,lten,gtn,gten,
                      equaln,unequaln :
                             begin
                                op:=CMP;
                                cmpop:=true;
                             end;
                      xorn : op:=A_XOR;
                      orn : op:=A_OR;
                      andn : op:=A_AND;
                      else error(type_mismatch);
                   end;
                   { links und rechts kein Register ? }
                   { dann mu eins angefordert werden }
                   if (p^.location.loc<>LOC_REGISTER) and
                     (p^.right^.location.loc<>LOC_REGISTER) then
                     begin
                        del_referenz(p^.location.referenz);

                        { erst freigeben, dann neues Register anfordern }
                        case opsize of
                           S_L : hregister:=getregister32;
                           S_B : hregister:=reg32toreg8(getregister32);
                        end;
                        emit(MOV,opsize,getreferenzstring(p^.location.referenz)+','
                          +regid2str[hregister]);
                        p^.location.loc:=LOC_REGISTER;
                        p^.location.register:=hregister;
                     end
                   else
                     { wenn rechts das Register, dann vertauschen }
                     if (p^.right^.location.loc=LOC_REGISTER) then
                       begin
                          swapl:=p^.location;
                          p^.location:=p^.right^.location;
                          p^.right^.location:=swapl;

                          { erneut vertauscht, also swaped-Flag neu setzten }
                          p^.swaped:=not(p^.swaped);
                       end;
                   if p^.right^.location.loc<>LOC_REGISTER then
                     begin
                        if (p^.treetype=subn) and p^.swaped then
                          begin
                             emit(MOV,opsize,getreferenzstring(p^.right^.location.referenz)+',%edi');
                             emit(SUB,opsize,regid2str[p^.location.register]+',%edi');
                             emit(MOV,opsize,'%edi,'+regid2str[p^.location.register]);
                          end
                        else
                          begin
                             if (p^.right^.treetype=ordconstn) and
                                (op=CMP) and
                                (p^.right^.value=0) then
                               begin
                                  emit(TEST,opsize,
                                    regid2str[p^.location.register]+','
                                    +regid2str[p^.location.register]);
                               end
                             else if (p^.right^.treetype=ordconstn) and
                                (op=ADD) and
                                (p^.right^.value=1) then
                               begin
                                  emit(A_INC,opsize,
                                    regid2str[p^.location.register]);
                               end
                             else if (p^.right^.treetype=ordconstn) and
                                (op=SUB) and
                                (p^.right^.value=1) then
                               begin
                                  emit(A_DEC,opsize,
                                    regid2str[p^.location.register]);
                               end
                             else if (p^.right^.treetype=ordconstn) and
                                (op=IMUL) and
                                (ispowerof2(p^.right^.value,power)) then
                               begin
                                  emit(A_SHL,opsize,'$'+tostr(power)+','+
                                    +regid2str[p^.location.register]);
                               end
                             else
                               emit(op,opsize,getreferenzstring(
                                 p^.right^.location.referenz)+','
                                 +regid2str[p^.location.register]);
                          end;
                        del_referenz(p^.right^.location.referenz);
                     end
                   else
                     begin
                        { wenn vertauscht, anderes Ergebnis-Register }
                        if (p^.treetype=subn) and p^.swaped then
                          begin
                             emit(SUB,opsize,
                               regid2str[p^.location.register]+','+
                               regid2str[p^.right^.location.register]);

                               swapl:=p^.location;
                               p^.location:=p^.right^.location;
                               p^.right^.location:=swapl;

                               { erneut vertauscht, also swaped-Flag }
                               { neu setzten (nur der Ordnung halber) }
                               p^.swaped:=not(p^.swaped);
                          end
                        else
                          begin
                             emit(op,opsize,
                               regid2str[p^.right^.location.register]+','
                               +regid2str[p^.location.register]);
                          end;
                        case opsize of
                           S_L : ungetregister32(p^.right^.location.register);
                           S_B : ungetregister32(reg8toreg32(p^.right^.location.register));
                        end;
                     end;

                   if cmpop then
                     case opsize of
                        S_L : ungetregister32(p^.location.register);
                        S_B : ungetregister32(reg8toreg32(p^.location.register));
                     end;

                   { Nur wenn Overflow verdchtige Operation }
                   { Overflow-Code erzeugen }
                   if mboverflow then
                     emitoverflowcheck;
                end
              else if ((p^.left^.resulttype^.deftype=grunddef) and
                 (pgrunddef(p^.left^.resulttype)^.typ=uchar)) then
                begin
                   case p^.treetype of
                      ltn,lten,gtn,gten,
                      equaln,unequaln :
                                cmpop:=true;
                      else error(type_mismatch);
                   end;
                   { links und rechts kein Register ? }
                   { dann mu eins angefordert werden }
                   if (p^.location.loc<>LOC_REGISTER) and
                     (p^.right^.location.loc<>LOC_REGISTER) then
                     begin
                        del_referenz(p^.location.referenz);

                        { erst freigeben, dann neues Register anfordern }
                        hregister:=reg32toreg8(getregister32);
                        emit(MOV,S_B,getreferenzstring(p^.location.referenz)+','
                          +regid2str[hregister]);
                        p^.location.loc:=LOC_REGISTER;
                        p^.location.register:=hregister;
                     end;

                   { nun p nun auf jeden Fall Register }
                   if (p^.right^.location.loc=LOC_REGISTER) and
                      (p^.location.loc<>LOC_REGISTER) then
                     begin
                        swapl:=p^.location;
                        p^.location:=p^.right^.location;
                        p^.right^.location:=swapl;

                        { erneut vertauscht, also swaped-Flag neu setzten }
                        p^.swaped:=not(p^.swaped);
                     end;
                   if p^.right^.location.loc<>LOC_REGISTER then
                     begin
                        emit(CMP,S_B,getreferenzstring(
                          p^.right^.location.referenz)+','
                           +regid2str[p^.location.register]);
                        del_referenz(p^.right^.location.referenz);
                     end
                   else
                     begin
                        emit(CMP,S_L,
                          regid2str[p^.right^.location.register]+','
                          +regid2str[p^.location.register]);
                        ungetregister32(reg8toreg32(p^.right^.location.register));
                     end;
                   ungetregister32(reg8toreg32(p^.location.register));
                end
              else if ((p^.left^.resulttype^.deftype=grunddef) and
                  (pgrunddef(p^.left^.resulttype)^.typ=s64real)) then
                 begin
                    { Realkonstante nach rechts }
                    if p^.left^.treetype=realconstn then
                      begin
                         swapp:=p^.right;
                         p^.right:=p^.left;
                         p^.left:=swapp;
                         p^.swaped:=not(p^.swaped);
                      end;
                    cmpop:=false;
                    case p^.treetype of
                       addn : op:=FADDP;
                       muln : op:=FMULP;
                       subn : op:=FSUBP;
                       slashn : op:=FDIVP;
                       ltn,lten,gtn,gten,
                       equaln,unequaln :
                              begin
                                 op:=FCOMPP;
                                 cmpop:=true;
                              end;
                       else error(type_mismatch);
                    end;
                    { links nicht auf dem Stack, dann laden }
                    if (p^.left^.location.loc<>LOC_FPUSTACK) then
                      emit(FLD,S_L,getreferenzstring(p^.left^.location.referenz));
                    if (p^.right^.location.loc<>LOC_FPUSTACK) then
                      emit(FLD,S_L,getreferenzstring(p^.right^.location.referenz));
                    if p^.swaped and
                       ((p^.treetype=slashn) or (p^.treetype=subn)) then
                      emit(FXCH,S_L,'');
                    emit(op,S_NO,'%st,%st(1)');
                    { bei Vergleich Flags laden }
                    if cmpop then
                      begin
                         if not(R_EAX in unused) then
                           emit(MOV,S_L,'%eax,%edi');
                         emit(FNSTS,S_W,'%ax');
                         emit(SAHF,S_NO,'');
                         if not(R_EAX in unused) then
                           emit(MOV,S_L,'%edi,%eax');
                      end;
                    p^.location.loc:=LOC_FPUSTACK;
                 end
              else if (p^.left^.resulttype^.deftype=stringdef) then
                begin
                   { String-Operationen ist nicht kommutativ }
                   if p^.swaped then
                     begin
                        swapp:=p^.left;
                        p^.left:=p^.right;
                        p^.right:=swapp;
                        { wegen der Sprungerzeugung bei Vergleichen unten: }
                        p^.swaped:=not(p^.swaped);
                     end;
                   case p^.treetype of
                      addn : begin
                                cmpop:=false;
                                if (p^.left^.location.loc<>LOC_MEM) or
                                   (p^.left^.treetype=stringconstn) then
                                  begin
                                     { kann nur Referenz sein }
                                     { String in Register waere komisch }
                                     { also einen temporaeren String erzeugen }

                                     del_referenz(p^.left^.location.referenz);
                                     pushusedregisters(pushedregs,$ff);
                                     clear_referenz(href);
                                     href.offset:=gettempofsize(256);
                                     href.base:=R_EBP;
                                     emitpushreferenzaddr(href);
                                     emitpushreferenzaddr(p^.left^.location.referenz);
                                     emit(PUSH,S_L,'$255');
                                     emit(CALL,S_NO,'STRCOPY');
                                     maybe_loadesi;
                                     p^.left^.location.referenz:=href;
                                     popusedregisters(pushedregs);
                                  end;

                                { Rechts brauchen wir die Register auch nicht mehr }
                                del_referenz(p^.right^.location.referenz);
                                pushusedregisters(pushedregs,$ff);
                                p^.location:=p^.left^.location;
                                emitpushreferenzaddr(p^.location.referenz);
                                emitpushreferenzaddr(p^.right^.location.referenz);
                                emit(CALL,S_NO,'STRCONCAT');
                                maybe_loadesi;
                                popusedregisters(pushedregs);
                             end;
                   ltn,lten,gtn,gten,
                     equaln,unequaln :
                             begin
                                cmpop:=true;
                                del_referenz(p^.right^.location.referenz);
                                del_referenz(p^.left^.location.referenz);
                                pushusedregisters(pushedregs,$ff);
                                emitpushreferenzaddr(p^.left^.location.referenz);
                                emitpushreferenzaddr(p^.right^.location.referenz);
                                emit(CALL,S_NO,'STRCMP');
                                maybe_loadesi;
                                popusedregisters(pushedregs);
                             end;
                     else error(type_mismatch);
                   end;
                end
              else error(type_mismatch);
           end;
         { Sollte es ein Vergleichsoperator sein, dann Ergebnis in den Flags }
         if cmpop then
           begin
              if p^.swaped then
                case p^.treetype of
                   equaln : flags:=F_E;
                   unequaln : flags:=F_NE;
                   ltn : flags:=F_G;
                   lten : flags:=F_GE;
                   gtn : flags:=F_L;
                   gten : flags:=F_LE;
                end
              else
                case p^.treetype of
                   equaln : flags:=F_E;
                   unequaln : flags:=F_NE;
                   ltn : flags:=F_L;
                   lten : flags:=F_LE;
                   gtn : flags:=F_G;
                   gten : flags:=F_GE;
                end;
              p^.location.loc:=LOC_FLAGS;
              p^.location.resflags:=flags;
           end;
      end;

    procedure firstmoddiv(var p : ptree);far;

      var
         t : ptree;
         power : longint;

      begin
         firstpass(p^.left);
         firstpass(p^.right);

         if codegenerror then
           exit;

         if is_constintnode(p^.left) and is_constintnode(p^.right) then
           begin
              case p^.treetype of
                 modn : t:=genordinalconstnode(p^.left^.value mod p^.right^.value,s32bitdef);
                 divn : t:=genordinalconstnode(p^.left^.value div p^.right^.value,s32bitdef);
              end;
              disposetree(p);
              p:=t;
              exit;
           end;

         p^.right:=gentypeconvnode(p^.right,s32bitdef);
         p^.left:=gentypeconvnode(p^.left,s32bitdef);
         firstpass(p^.left);
         firstpass(p^.right);

         if codegenerror then
           exit;

         p^.registers32:=p^.left^.registers32;

         if p^.registers32<p^.right^.registers32 then
           p^.registers32:=p^.right^.registers32;
         if p^.registers32<2 then p^.registers32:=2;

         p^.resulttype:=s32bitdef;
         p^.location.loc:=LOC_REGISTER;
      end;

    procedure secondmoddiv(var p : ptree);far;

      var
         hreg1 : tregister;
         pushed,popeax,popedx : boolean;
         power,hl : longint;

      begin
         secondpass(p^.left);
         p^.location:=p^.left^.location;
         pushed:=maybe_push(p^.right^.registers32,p);
         secondpass(p^.right);
         if pushed then restore(p);

         { Zaehler in Register bringen }
         if p^.left^.location.loc<>LOC_REGISTER then
           begin
              del_referenz(p^.left^.location.referenz);
              p^.left^.location.loc:=LOC_REGISTER;
              hreg1:=getregister32;
              emit(MOV,S_L,getreferenzstring(p^.left^.location.referenz)+','+
                regid2str[hreg1]);
              p^.left^.location.register:=hreg1;
           end
         else hreg1:=p^.left^.location.register;

         if (p^.treetype=divn) and (p^.right^.treetype=ordconstn) and
            ispowerof2(p^.right^.value,power) then
           begin
              emit(A_OR,S_L,regid2str[hreg1]+','+
                regid2str[hreg1]);
              hl:=getlabel;
              emitl(JNS,hl);
              if power=1 then
                emit(A_INC,S_L,regid2str[hreg1])
              else emit(ADD,S_L,'$'+tostr(p^.right^.value-1)+','+regid2str[hreg1]);
              emitl(A_LABEL,hl);
              emit(SAR,S_L,'$'+tostr(power)+','+regid2str[hreg1]);
           end
         else
           begin
              { Nenner nach EDI bringen }
              if p^.right^.location.loc<>LOC_REGISTER then
                begin
                   del_referenz(p^.right^.location.referenz);
                   p^.left^.location.loc:=LOC_REGISTER;
                   emit(MOV,S_L,getreferenzstring(p^.right^.location.referenz)+',%edi');
                end
              else
                begin
                   ungetregister32(p^.right^.location.register);
                   emit(MOV,S_L,regid2str[p^.right^.location.register]+',%edi');
                end;
              popedx:=false;
              popeax:=false;
              if hreg1=R_EDX then
                begin
                   if not(R_EAX in unused) then
                     begin
                        emit(PUSH,S_L,'%eax');
                        popeax:=true;
                     end;
                   emit(MOV,S_L,'%edx,%eax');
                end
              else
                begin
                   if not(R_EDX in unused) then
                     begin
                        emit(PUSH,S_L,'%edx');
                        popedx:=true;
                     end;
                   if hreg1<>R_EAX then
                     begin
                        if not(R_EAX in unused) then
                          begin
                             emit(PUSH,S_L,'%eax');
                             popeax:=true;
                          end;
                        emit(MOV,S_L,regid2str[hreg1]+',%eax');
                     end;
                end;
              emit(CLTD,S_NO,'');
              emit(IDIV,S_L,'%edi');
              if p^.treetype=divn then
                begin
                   { wenn das Ergebnisregister belegt ist, dann umkopieren }
                   if (hreg1<>R_EAX) and not popeax then
                     emit(MOV,S_L,'%eax,'+regid2str[hreg1])
                   else
                     hreg1:=R_EAX;
                end
              else
                begin
                   { dito. }
                   if (hreg1<>R_EDX) and not popedx then
                     emit(MOV,S_L,'%edx,'+regid2str[hreg1])
                   else
                     hreg1:=R_EDX;
                end;
              if popeax then
                emit(POP,S_L,'%eax');
              if popedx then
                emit(POP,S_L,'%edx');
           end;
         p^.location.loc:=LOC_REGISTER;
         p^.location.register:=hreg1;
      end;

    procedure firstshlshr(var p : ptree);far;

      var
         t : ptree;

      begin
         firstpass(p^.left);
         firstpass(p^.right);

         if codegenerror then
           exit;

         if is_constintnode(p^.left) and is_constintnode(p^.right) then
           begin
              case p^.treetype of
                 shrn : t:=genordinalconstnode(p^.left^.value shr p^.right^.value,s32bitdef);
                 shln : t:=genordinalconstnode(p^.left^.value shl p^.right^.value,s32bitdef);
              end;
              disposetree(p);
              p:=t;
              exit;
           end;
         p^.right:=gentypeconvnode(p^.right,s32bitdef);
         p^.left:=gentypeconvnode(p^.left,s32bitdef);
         firstpass(p^.left);
         firstpass(p^.right);

         if codegenerror then
           exit;

         p^.registers32:=p^.left^.registers32;

         if p^.registers32<p^.right^.registers32 then
           p^.registers32:=p^.right^.registers32;
         if p^.registers32<1 then p^.registers32:=1;

         p^.resulttype:=s32bitdef;
         p^.location.loc:=LOC_REGISTER;
      end;

    procedure secondshlshr(var p : ptree);far;

      var
         hregister : tregister;
         pushed,popecx : boolean;
         op : tasmop;

      begin
         secondpass(p^.left);
         pushed:=maybe_push(p^.right^.registers32,p);
         secondpass(p^.right);
         if pushed then restore(p);

         { Verschiebung mit einer Konstante direkt auscodieren: }
         if (p^.right^.treetype=ordconstn) then
           begin
              { linken Operanden in ein Register laden }
              if p^.left^.location.loc<>LOC_REGISTER then
                begin
                   del_referenz(p^.left^.location.referenz);
                   hregister:=getregister32;
                   emit(MOV,S_L,getreferenzstring(p^.left^.location.referenz)+
                      ','+regid2str[hregister]);
                end
              else hregister:=p^.left^.location.register;
              { Operator feststellen }
              if p^.treetype=shln then
                op:=A_SHL
              else
                op:=A_SHR;
              emit(op,S_L,'$'+tostr(p^.right^.location.referenz.offset mod 32)
                +','+regid2str[hregister]);
              p^.location.loc:=LOC_REGISTER;
              p^.location.register:=hregister;
           end
         else
           begin
              internalerror(5);
              { falls linker Operand in ECX, dann nach EDI }
              { sonst falls noetig ECX retten }
              { ...und belegen }

              { jetzt linken Operander in Register laden, }
              { rechten Operand in ECX laden }
              { ...und Erzeugen }
              { falls noetig ECX zurueck }
           end;
      end;

    procedure firstrealconst(var p : ptree);far;

      begin
         p^.location.loc:=LOC_MEM;
      end;

    procedure secondrealconst(var p : ptree);far;

      begin
         clear_referenz(p^.location.referenz);
         if p^.labnumber=-1 then
           begin
              p^.labnumber:=getlabel;
              constsegment.insert(gennasmrec(A_DOUBLE,S_NO,double2str(p^.valued)));
              constsegment.insert(genlasmrec(A_LABEL,p^.labnumber));
           end;
         p^.location.referenz.symbol:=stringdup(tolabel(p^.labnumber));
      end;

    procedure firstordconst(var p : ptree);far;

      begin
         p^.location.loc:=LOC_MEM;
      end;

    procedure secondordconst(var p : ptree);far;

      begin
         { eine Intconst. tut so, als ob sie eine Memoryreferenz wre }
         p^.location.loc:=LOC_MEM;
         p^.location.referenz.isintvalue:=true;
         p^.location.referenz.offset:=p^.value;
      end;

    procedure firstniln(var p : ptree);far;

      begin
         p^.resulttype:=voidpointerdef;
         p^.location.loc:=LOC_MEM;
      end;

    procedure secondniln(var p : ptree);far;

      begin
         p^.location.loc:=LOC_MEM;
         p^.location.referenz.isintvalue:=true;
         p^.location.referenz.offset:=0;
      end;

    procedure firststringconst(var p : ptree);far;

      begin
         p^.resulttype:=new(pstringdef,init(length(p^.values^)));
         p^.location.loc:=LOC_MEM;
      end;

    procedure secondstringconst(var p : ptree);far;

      var
         l : longint;

      begin
         clear_referenz(p^.location.referenz);
         l:=getconstlabel;
         p^.location.referenz.symbol:=stringdup('L'+tostr(l));
         constsegment.concat(gennasmrec(DIRECT,S_NO,'L'+tostr(l)+':'));
         constsegment.concat(gennasmrec(ASCII,S_NO,'"'+ibm2ascii(char(length(p^.values^))+p^.values^)+'\0"'));
      end;

    procedure firstumminus(var p : ptree);far;

      var
         t : ptree;

      begin
         firstpass(p^.left);

         if codegenerror then
           exit;
         if is_constintnode(p^.left) then
           begin
              t:=genordinalconstnode(-p^.left^.value,s32bitdef);
              disposetree(p);
              p:=t;
              exit;
           end;
         p^.left:=gentypeconvnode(p^.left,s32bitdef);
         firstpass(p^.left);

         if codegenerror then
           exit;

         p^.registers32:=p^.left^.registers32;
         if (p^.left^.location.loc<>LOC_REGISTER) and
            (p^.registers32<1) then
            p^.registers32:=1;
         p^.location.loc:=LOC_REGISTER;
         p^.resulttype:=p^.left^.resulttype;
      end;

    procedure secondumminus(var p : ptree);far;

      begin
         secondpass(p^.left);
         p^.location.loc:=LOC_REGISTER;
         case p^.left^.location.loc of
            LOC_REGISTER : begin
                              p^.location.register:=p^.left^.location.register;
                              emit(NEG,S_L,regid2str[p^.location.register]);
                           end;
            LOC_REFERENZ,LOC_MEM :
                           begin
                              del_referenz(p^.left^.location.referenz);
                              p^.location.register:=getregister32;
                              emit(MOV,S_L,getreferenzstring(p^.left^.location.referenz)+
                                ','+regid2str[p^.location.register]);
                              emit(NEG,S_L,regid2str[p^.location.register]);
                           end;
         end;
         emitoverflowcheck;
      end;

    procedure firstaddr(var p : ptree);far;

      var
         hp : ptree;

      begin
         if p^.left^.treetype=calln then
           begin
              hp:=genloadnode(pvarsym(p^.left^.symtableprocentry),p^.left^.symtableproc);
              disposetree(p^.left);
              p^.left:=hp;
           end;

         firstpass(p^.left);

         if codegenerror then
           exit;

         if (p^.left^.location.loc<>LOC_REFERENZ) then
           error(error_in_expression);

         { falls Adresse von einer Callnode bestimmt werden soll, }
         { die Callnode in eine Loadnode umwandeln }

         p^.registers32:=p^.left^.registers32;
         if p^.registers32<1 then
           p^.registers32:=1;
         p^.location.loc:=LOC_REGISTER;
         if exprlevel<2 then
           p^.resulttype:=voidpointerdef
         else p^.resulttype:=new(ppointerdef,init(p^.left^.resulttype));
      end;

    procedure secondaddr(var p : ptree);far;

      begin
         secondpass(p^.left);
         p^.location.loc:=LOC_REGISTER;
         del_referenz(p^.left^.location.referenz);
         p^.location.register:=getregister32;
         emit(LEA,S_L,getreferenzstring(p^.left^.location.referenz)+','+
           regid2str[p^.location.register]);
      end;

    procedure firstnot(var p : ptree);far;

      var
         t : ptree;

      begin
         firstpass(p^.left);

         if codegenerror then
           exit;

         if (p^.left^.treetype=ordconstn) then
           begin
              t:=genordinalconstnode(not(p^.left^.value),p^.left^.resulttype);
              disposetree(p);
              p:=t;
              exit;
           end;
         p^.resulttype:=p^.left^.resulttype;
         p^.location.loc:=p^.left^.location.loc;
         if (p^.resulttype^.deftype=grunddef) and
            (pgrunddef(p^.resulttype)^.typ=bool8bit) then
              begin
                 p^.registers32:=p^.left^.registers32;
                 if (p^.location.loc=LOC_REFERENZ) and
                    (p^.registers32<1) then
                   p^.registers32:=1;
              end
           else
             begin
                p^.left:=gentypeconvnode(p^.left,s32bitdef);
                firstpass(p^.left);

                if codegenerror then
                  exit;

                p^.resulttype:=p^.left^.resulttype;
                p^.registers32:=p^.left^.registers32;
                if (p^.left^.location.loc<>LOC_REGISTER) and
                   (p^.registers32<1) then
                  p^.registers32:=1;
                p^.location.loc:=LOC_REGISTER;
             end;
      end;

    procedure secondnot(var p : ptree);far;

      const
         flagsinvers : array[F_E..F_LE] of tresflags =
            (F_NE,F_E,F_LE,F_GE,F_L,F_G);

      var
         hl : longint;

      begin
         if (p^.resulttype^.deftype=grunddef) and
            (pgrunddef(p^.resulttype)^.typ=bool8bit) then
              begin
                 case p^.location.loc of
                    LOC_JUMP : begin
                                  hl:=truelabel;
                                  truelabel:=falselabel;
                                  falselabel:=hl;
                                  secondpass(p^.left);
                                  maketojumpbool(p^.left);
                                  hl:=truelabel;
                                  truelabel:=falselabel;
                                  falselabel:=hl;
                               end;
                    LOC_FLAGS : begin
                                   secondpass(p^.left);
                                   p^.location.resflags:=flagsinvers[p^.left^.location.resflags];
                                end;
                    LOC_REGISTER : begin
                                      secondpass(p^.left);
                                      p^.location.register:=p^.left^.location.register;
                                      emit(A_XOR,S_B,'$1,'+regid2str[p^.location.register]);
                                   end;
                    LOC_REFERENZ,LOC_MEM : begin
                                              secondpass(p^.left);
                                              del_referenz(p^.left^.location.referenz);
                                              p^.location.loc:=LOC_REGISTER;
                                              p^.location.register:=reg32toreg8(getregister32);
                                              emit(MOV,S_B,getreferenzstring(p^.left^.location.referenz)+
                                                ','+regid2str[p^.location.register]);
                                                emit(A_XOR,S_B,'$1,'+regid2str[p^.location.register]);
                                           end;
                 end;
              end
            else
              begin
                secondpass(p^.left);
                p^.location.loc:=LOC_REGISTER;
                case p^.left^.location.loc of
                   LOC_REGISTER : begin
                                     p^.location.register:=p^.left^.location.register;
                                     emit(A_NOT,S_L,regid2str[p^.location.register]);
                                  end;
                   LOC_REFERENZ,LOC_MEM :
                                  begin
                                     del_referenz(p^.left^.location.referenz);
                                     p^.location.register:=getregister32;
                                     emit(MOV,S_L,getreferenzstring(p^.left^.location.referenz)+
                                       ','+regid2str[p^.location.register]);
                                     emit(A_NOT,S_L,regid2str[p^.location.register]);
                                  end;
                end;
             end;
      end;

    procedure firstnothing(var p : ptree);far;

      begin
      end;

    procedure secondnothing(var p : ptree);far;

      begin
      end;

    procedure firstassignment(var p : ptree);far;

      begin
         firstpass(p^.left);
         firstpass(p^.right);

         if codegenerror then
           exit;

         { sollte rechts und links ein String stehen, muss nicht konvertiert }
         { werden, da STRCOPY mit den richtigen Parametern aufgerufen wird   }
         if not((p^.right^.resulttype^.deftype=stringdef)
            and (p^.left^.resulttype^.deftype=stringdef)) then
           begin
              p^.right:=gentypeconvnode(p^.right,p^.left^.resulttype);

              { nochmal firstpass wegen der Typkonvertierung aufrufen }
              firstpass(p^.right);

              if codegenerror then
                exit;

           end;
         if (exprlevel<4) then p^.resulttype:=voiddef
           else p^.resulttype:=p^.right^.resulttype;
         p^.registers32:=max(p^.left^.registers32,p^.right^.registers32);
      end;

    procedure secondassignment(var p : ptree);far;

      var
         opsize : topsize;
         withresult : boolean;
         pushed : tpushed;
         otlabel,hlabel,oflabel : longint;

      begin
         otlabel:=truelabel;
         oflabel:=falselabel;
         truelabel:=getlabel;
         falselabel:=getlabel;

         withresult:=not(exprlevel<4);
         if (p^.left^.registers32<p^.right^.registers32) then
           begin
              secondpass(p^.right);
              secondpass(p^.left);
           end
         else
           begin
              secondpass(p^.left);
              secondpass(p^.right);
           end;
         if (p^.left^.location.loc<>LOC_REFERENZ) then
            error(error_in_expression);

         if p^.right^.resulttype^.deftype=stringdef then
           begin
              { das Ziel brauchen wir sicher nicht mehr }
              del_referenz(p^.left^.location.referenz);
              { die Quelle nur, wenn withresult gesetzt: }
              if not(withresult) then
                del_referenz(p^.right^.location.referenz);
              copystring(p^.left^.location.referenz,p^.right^.location.referenz,
                min(pstringdef(p^.right^.resulttype)^.len,pstringdef(p^.left^.resulttype)^.len));
           end
         else case p^.right^.location.loc of
            LOC_REFERENZ,
            LOC_MEM : begin
                         { ordinale Konstanten getrennt bearbeiten }
                         if p^.right^.treetype=ordconstn then
                           begin
                              case p^.left^.resulttype^.size of
                                 1 : opsize:=S_B;
                                 2 : opsize:=S_W;
                                 4 : opsize:=S_L;
                              end;
                              emit(MOV,opsize,
                                getreferenzstring(p^.right^.location.referenz)+','+
                                getreferenzstring(p^.left^.location.referenz));
                           end
                         else concatcopy(p^.right^.location.referenz,
                            p^.left^.location.referenz,p^.left^.resulttype^.size,
                            withresult);
                      end;
            LOC_REGISTER : begin
                              case p^.right^.resulttype^.size of
                                 1 : opsize:=S_B;
                                 2 : opsize:=S_W;
                                 4 : opsize:=S_L;
                              end;
                              emit(MOV,opsize,
                                regid2str[p^.right^.location.register]+','+
                                getreferenzstring(p^.left^.location.referenz));
                           end;
            LOC_FPUSTACK : begin
                              emit(FSTP,S_L,
                                getreferenzstring(p^.left^.location.referenz));
                           end;
            LOC_JUMP     : begin
                              hlabel:=getlabel;
                              emitl(A_LABEL,truelabel);
                              emit(MOV,S_B,'$1,'+
                                getreferenzstring(p^.left^.location.referenz));
                              emitl(JMP,hlabel);
                              emitl(A_LABEL,falselabel);
                              emit(MOV,S_B,'$0,'+
                                getreferenzstring(p^.left^.location.referenz));
                              emitl(A_LABEL,hlabel);
                           end;
            LOC_FLAGS    : begin
                              emit(flag_2_set[p^.right^.location.resflags],S_NO,
                                getreferenzstring(p^.left^.location.referenz));
                           end;
         end;
         truelabel:=otlabel;
         falselabel:=oflabel;
      end;

    { erster Pass fr links und rechts }

    procedure firstlr(var p : ptree);far;

      begin
         firstpass(p^.left);
         firstpass(p^.right);

         if codegenerror then
           exit;

      end;

    procedure firstderef(var p : ptree);far;

      begin
         firstpass(p^.left);
         if codegenerror then
           exit;

         if p^.left^.location.loc=LOC_REGISTER then
           p^.registers32:=p^.left^.registers32
         else p^.registers32:=p^.left^.registers32+1;
         if p^.left^.resulttype^.deftype<>pointerdef then
           error(invalid_qualifizier);
         p^.resulttype:=ppointerdef(p^.left^.resulttype)^.definition;
         p^.location.loc:=LOC_REFERENZ;
      end;

    procedure secondderef(var p : ptree);far;

      var
         hr : tregister;

      begin
         secondpass(p^.left);
         clear_referenz(p^.location.referenz);
         if p^.left^.location.loc=LOC_REGISTER then
           begin
              p^.location.referenz.base:=p^.left^.location.register;
           end
         else
           begin
              { Register freigeben }
              del_referenz(p^.left^.location.referenz);

              { ...und eines fr den Pointer reservieren }
              hr:=getregister32;
              emit(
              MOV,S_L,getreferenzstring(p^.left^.location.referenz)+','+
                regid2str[hr]);
              p^.location.referenz.base:=hr;
           end;
      end;

    procedure firstrange(var p : ptree);far;

      var
         ct : tconverttype;

      begin
         firstpass(p^.left);
         firstpass(p^.right);
         if codegenerror then
           exit;
         { nur ordinale Konstanten zulassen }
         if not((p^.left^.treetype=ordconstn) and
                 (p^.right^.treetype=ordconstn)) then
           error(error_in_expression);
         { Obergrenze mu grer oder gleich Untergrenze sein }
         if (p^.left^.value>p^.right^.value) then
           error(upper_l_lower);
         { beide Typen mssen kompatibel sein }
         if not(isconvertable(p^.left^.resulttype,p^.right^.resulttype,
           ct,ordconstn { nur Dummy} )) and
           not(is_equal(p^.left^.resulttype,p^.right^.resulttype)) then
           error(type_mismatch);
      end;

    procedure firstvecn(var p : ptree);far;

      var
         harr : pdef;

      begin
         firstpass(p^.left);
         firstpass(p^.right);
         if codegenerror then
           exit;

         { Aufzhltypen drfen nur Index fr Arrays sein }
         if (p^.left^.resulttype^.deftype=arraydef) and
           (p^.right^.resulttype^.deftype=aufzaehldef) then
           begin
              if not is_equal(parraydef(p^.left^.resulttype)^.rangedef,
                p^.right^.resulttype) then
                error(type_mismatch);
           end
         else
           p^.right:=gentypeconvnode(p^.right,s32bitdef);

         { nochmal firstpass }
         firstpass(p^.right);

         if codegenerror then
           exit;

         { Returntyp berechnen }
         if p^.left^.resulttype^.deftype=arraydef then
           p^.resulttype:=parraydef(p^.left^.resulttype)^.definition
         else if (p^.left^.resulttype^.deftype=pointerdef) then
           begin
              { Pointer in Array umwandeln }
              harr:=new(parraydef,init(0,$7fffffff,s32bitdef));
              parraydef(harr)^.definition:=ppointerdef(p^.left^.resulttype)^.definition;
              p^.left:=gentypeconvnode(p^.left,harr);
              firstpass(p^.left);

              if codegenerror then
                exit;
              p^.resulttype:=parraydef(harr)^.definition
           end
         else
         { indizierter Zugriff auf String }
           p^.resulttype:=cchardef;

         p^.registers32:=max(p^.left^.registers32,p^.right^.registers32);

         if (p^.right^.location.loc<>LOC_REGISTER) and
            (p^.registers32<3) and (p^.right^.registers32=0) then
            inc(p^.registers32);

         { es wird derselbe Speichertyp wie links zurckgegeben }
         p^.location.loc:=p^.left^.location.loc;
      end;

    procedure secondvecn(var p : ptree);far;

      var
         pushed : boolean;
         ind : tregister;
         _p : ptree;

      procedure calc_emit_mul;

        var
           l1,l2 : longint;

        begin
           l1:=p^.resulttype^.size;
           case l1 of
              1,2,4,8 : p^.location.referenz.scalefactor:=l1;
           else
                begin
                   if ispowerof2(l1,l2) then
                     emit(A_SHL,S_L,'$'+tostr(l2)+','+regid2str[ind])
                   else
                     emit(IMUL,S_L,'$'+tostr(l1)+','+regid2str[ind]);
                end;
           end;
        end;

      begin
         secondpass(p^.left);
         p^.location:=p^.left^.location;
         { Offset kann nur ungleich 0, wenn arraydef }
         if p^.left^.resulttype^.deftype=arraydef then
           dec(p^.location.referenz.offset,
             p^.resulttype^.size*parraydef(p^.left^.resulttype)^.lowrange);
         if p^.right^.treetype=ordconstn then
           begin

              { Offset kann nur ungleich 0, wenn arraydef }
              if p^.left^.resulttype^.deftype=arraydef then
                begin
                   if (p^.right^.value>parraydef(p^.left^.resulttype)^.highrange) or
                      (p^.right^.value<parraydef(p^.left^.resulttype)^.lowrange) then
                     error(range_check_error);

                   dec(p^.left^.location.referenz.offset,
                      p^.resulttype^.size*parraydef(p^.left^.resulttype)^.lowrange);
                end;
              inc(p^.left^.location.referenz.offset,
                 p^.right^.value*p^.resulttype^.size);
              p^.left^.resulttype:=p^.resulttype;
              disposetree(p^.right);
              _p:=p^.left;
              putnode(p);
              p:=_p;
           end
         else
           begin
              { von links nach rechts rechnen }
              if (p^.location.loc<>LOC_REFERENZ) and
                 (p^.location.loc<>LOC_MEM) then
                error(error_in_expression);
              pushed:=maybe_push(p^.right^.registers32,p);
              secondpass(p^.right);
              if pushed then restore(p);
              if p^.right^.location.loc<>LOC_REGISTER then
                begin
                   del_referenz(p^.right^.location.referenz);
                   ind:=getregister32;
                   emit(MOV,S_L,getreferenzstring(p^.right^.location.referenz)+
                       ','+regid2str[ind]);
                end
              else ind:=p^.right^.location.register;

              { moeglichen Rangecheckcode erzeugen: }
              if rangechecking then
                begin
                   if p^.left^.resulttype^.deftype=arraydef then
                     begin
                        parraydef(p^.left^.resulttype)^.genrangecheck;
                        emit(A_BOUND,S_L,regid2str[ind]+
                          ',R_'+tostr(parraydef(p^.left^.resulttype)^.rangenr));
                     end;
                end;
              if p^.location.referenz.index=R_NO then
                begin
                   p^.location.referenz.index:=ind;
                   calc_emit_mul;
                end
              else
                begin
                   if p^.location.referenz.base=R_NO then
                     begin
                        calc_emit_mul;
                        p^.location.referenz.base:=p^.location.referenz.index;
                        p^.location.referenz.index:=ind;
                     end
                   else
                     begin
                        emit(
                           LEA,S_L,+getreferenzstring(p^.location.referenz)+','+
                             regid2str[p^.location.referenz.base]);
                        ungetregister32(p^.location.referenz.index);
                        stringdispose(p^.location.referenz.symbol);
                        p^.location.referenz.offset:=0;
                        calc_emit_mul;
                        p^.location.referenz.index:=ind;
                     end;
                end;
           end;
      end;

    { *************** Typekonvertierungen **************** }

    { erzeugt, falls noetig Rangecheckcode }

    procedure maybe_rangechecking(p : ptree;p2,p1 : pdef);

      var
         hregister : tregister;

      begin
         if rangechecking and
           ((pgrunddef(p1)^.von>pgrunddef(p2)^.von) or
           (pgrunddef(p1)^.bis<pgrunddef(p2)^.bis)) then
           begin
              pgrunddef(p1)^.genrangecheck;
              if pgrunddef(p1)^.typ=u8bit then
                begin
                   if p^.location.loc=LOC_REGISTER then
                     emit(MOVZX,S_BL,regid2str[p^.location.register]+',%edi')
                   else
                     emit(MOVZX,S_BL,getreferenzstring(p^.location.referenz)+',%edi');
                   hregister:=R_EDI;
                end
              else if pgrunddef(p1)^.typ=s8bit then
                begin
                   if p^.location.loc=LOC_REGISTER then
                     emit(MOVSX,S_BL,regid2str[p^.location.register]+',%edi')
                   else
                     emit(MOVSX,S_BL,getreferenzstring(p^.location.referenz)+',%edi');
                   hregister:=R_EDI;
                end
              else if pgrunddef(p1)^.typ=s32bit then
                begin
                   if p^.location.loc=LOC_REGISTER then
                     hregister:=p^.location.register
                   else
                     begin
                        emit(MOV,S_L,getreferenzstring(p^.location.referenz)+',%edi');
                        hregister:=R_EDI;
                     end;
                end
              else if pgrunddef(p1)^.typ=u16bit then
                begin
                   if p^.location.loc=LOC_REGISTER then
                     emit(MOVZX,S_WL,regid2str[p^.location.register]+',%edi')
                   else
                     emit(MOVZX,S_WL,getreferenzstring(p^.location.referenz)+',%edi');
                   hregister:=R_EDI;
                end
              else if pgrunddef(p1)^.typ=s16bit then
                begin
                   if p^.location.loc=LOC_REGISTER then
                     emit(MOVSX,S_WL,regid2str[p^.location.register]+',%edi')
                   else
                     emit(MOVSX,S_WL,getreferenzstring(p^.location.referenz)+',%edi');
                   hregister:=R_EDI;
                end
              else internalerror(6);
              emit(A_BOUND,S_L,regid2str[hregister]+',R_'+tostr(pgrunddef(p1)^.rangenr))
           end;
      end;

    type
       tfirstconvproc = procedure(p : ptree);

    procedure first_bigger_smaller(p : ptree);far;

      begin
         if (p^.left^.location.loc<>LOC_REGISTER) and (p^.registers32=0) then
           p^.registers32:=1;
         p^.location.loc:=LOC_REGISTER;
      end;

    procedure first_cstring_charpointer(p : ptree);far;

      begin
         p^.registers32:=1;
         p^.location.loc:=LOC_REGISTER;
      end;

    procedure first_cstring_chararray(p : ptree);far;

      begin
         p^.location.loc:=LOC_MEM;
      end;

    procedure first_string_string(p : ptree);far;

      begin
         p^.location.loc:=LOC_MEM;
      end;

    procedure first_char_to_string(p : ptree);far;

      begin
         p^.location.loc:=LOC_MEM;
      end;

    procedure first_nothing(p : ptree);far;

      begin
      end;

    procedure first_array_to_pointer(p : ptree);far;

      begin
         if p^.registers32<1 then
           p^.registers32:=1;
         p^.location.loc:=LOC_REGISTER;
      end;

    procedure first_int_real(p : ptree);far;

      begin
         if p^.left^.treetype=ordconstn then
           begin
              { Konstanten direkt umwandeln }
              p^.treetype:=realconstn;
              p^.labnumber:=-1;
              p^.valued:=p^.left^.value;
              p^.disposetyp:=dt_nothing;
              disposetree(p^.left);
              p^.location.loc:=LOC_MEM;
           end
         else
           begin
              inc(p^.registersfpu);
              p^.location.loc:=LOC_FPUSTACK;
           end;
      end;

    procedure first_pointer_to_array(p : ptree);far;

      begin
         if p^.registers32<1 then
           p^.registers32:=1;
         p^.location.loc:=LOC_REFERENZ;
      end;

    { Achtung:  *** kein ***  rekursiver Aufruf von firstpass }

    procedure firsttypeconv(var p : ptree);far;

      var
         hp : ptree;

      const
         firstconvert : array[tc_u8bit_2_s32bit..tc_int_2_real] of
           tfirstconvproc = (first_bigger_smaller,first_nothing,first_bigger_smaller,
                             first_bigger_smaller,first_bigger_smaller,
                             first_bigger_smaller,first_bigger_smaller,
                             first_bigger_smaller,first_string_string,
                             first_cstring_charpointer,first_cstring_chararray,
                             first_array_to_pointer,first_pointer_to_array,
                             first_char_to_string,first_bigger_smaller,
                             first_bigger_smaller,first_bigger_smaller,
                             first_bigger_smaller,first_bigger_smaller,
                             first_bigger_smaller,first_bigger_smaller,
                             first_bigger_smaller,first_bigger_smaller,
                             first_int_real);

      begin
         { bei expliziten Typkonvertierungen firstpass ausfhren }
         if p^.explizit then
           firstpass(p^.left);

         if codegenerror then
           exit;

         { Typekonvertierungen auf sich selbst entfernen }
         if is_equal(p^.left^.resulttype,p^.resulttype) then
           begin
              hp:=p;
              p:=p^.left;
              p^.resulttype:=hp^.resulttype;
              putnode(hp);
              exit;
           end;
         p^.registers32:=p^.left^.registers32;
         p^.registersfpu:=p^.left^.registersfpu;
         p^.location:=p^.left^.location;
         if not(isconvertable(p^.left^.resulttype,p^.resulttype,p^.convtyp,p^.left^.treetype)) then
           begin
              if p^.explizit then
                begin
                   { wenn Aufzhltyp nach Ordinal konvertiert werden soll }
                   { dann Aufzhltyp=s32bit                               }
                   if (p^.left^.resulttype^.deftype=aufzaehldef) and
                      is_ordinal(p^.resulttype) then
                     begin
                        if p^.left^.treetype=ordconstn then
                          begin
                             hp:=genordinalconstnode(p^.left^.value,p^.resulttype);
                             disposetree(p);
                             p:=hp;
                             exit;
                          end;
                        if not isconvertable(s32bitdef,p^.resulttype,p^.convtyp,ordconstn { nur Dummy} ) then
                          error(ill_type_cast);
                     end
                   {  entsprechend Ordinal nach Aufzhl: }
                   else if (p^.resulttype^.deftype=aufzaehldef) and
                      is_ordinal(p^.left^.resulttype) then
                     begin
                        if p^.left^.treetype=ordconstn then
                          begin
                             hp:=genordinalconstnode(p^.left^.value,p^.resulttype);
                             disposetree(p);
                             p:=hp;
                             exit;
                          end;
                        if not isconvertable(p^.left^.resulttype,s32bitdef,p^.convtyp,ordconstn { nur Dummy} ) then
                          error(ill_type_cast);
                     end
                   { nur wenn gleiche Gre }
                   else if (p^.left^.resulttype^.deftype<>formaldef) and
                     (p^.left^.resulttype^.size<>p^.resulttype^.size) then
                     error(ill_type_cast);
                   { und nach strukturierten Typen nur,     }
                   { wenn die Quelle nicht ein Register ist }
                   case p^.resulttype^.deftype of
                      recorddef,stringdef,arraydef,classdef :
                        if p^.left^.location.loc=LOC_REGISTER then
                          error(ill_type_cast);
                   end;
                end
              else
                error(type_mismatch);
           end
         else
           begin
              p^.explizit:=false;
              { ordinale Konstanten direkt konvertieren }
              if (p^.left^.treetype=ordconstn) and is_ordinal(p^.resulttype) then
                begin
                   hp:=genordinalconstnode(p^.left^.value,p^.resulttype);
                   disposetree(p);
                   p:=hp;
                   exit;
                end;
              if p^.convtyp<>tc_equal then
                firstconvert[p^.convtyp](p);
           end;
      end;

    type
       tsecondconvproc = procedure(p,hp : ptree;convtyp : tconverttype);

    procedure second_only_rangecheck(p,hp : ptree;convtyp : tconverttype);far;

      begin
         maybe_rangechecking(p,hp^.resulttype,p^.resulttype);
      end;

    procedure second_bigger(p,hp : ptree;convtyp : tconverttype);far;

      var
         hregister : tregister;
         opsize : topsize;
         op : tasmop;
         is_register : boolean;

      begin
         is_register:=p^.left^.location.loc=LOC_REGISTER;
         if not(is_register) then
           del_referenz(p^.left^.location.referenz);
         case convtyp of
            tc_u8bit_2_s32bit :
              begin
                 if is_register then
                   hregister:=reg8toreg32(p^.left^.location.register)
                 else hregister:=getregister32;
                 op:=MOVZX;
                 opsize:=S_BL;
              end;
            tc_s8bit_2_s32bit :
              begin
                 if is_register then
                   hregister:=reg8toreg32(p^.left^.location.register)
                 else hregister:=getregister32;
                 op:=MOVSX;
                 opsize:=S_BL;
              end;
            tc_u16bit_2_s32bit :
              begin
                 if is_register then
                   hregister:=tregister(byte(p^.left^.location.register)-byte(R_ESP))
                 else hregister:=getregister32;
                 op:=MOVZX;
                 opsize:=S_WL;
              end;
            tc_s16bit_2_s32bit :
              begin
                 if is_register then
                   hregister:=tregister(byte(p^.left^.location.register)-byte(R_ESP))
                 else hregister:=getregister32;
                 op:=MOVSX;
                 opsize:=S_WL;
              end;
            tc_s8bit_2_u16bit,
            tc_u8bit_2_s16bit,
            tc_u8bit_2_u16bit :
              begin
                 if is_register then
                   hregister:=reg8toreg16(p^.left^.location.register)
                 else hregister:=reg32toreg16(getregister32);
                 op:=MOVZX;
                 opsize:=S_BW;
              end;
            tc_s8bit_2_s16bit :
              begin
                 if is_register then
                   hregister:=reg8toreg16(p^.left^.location.register)
                 else hregister:=reg32toreg16(getregister32);
                 op:=MOVSX;
                 opsize:=S_BW;
              end;
         end;
         if is_register then
           begin
              emit(op,opsize,
                regid2str[p^.left^.location.register]+','+regid2str[hregister]);
           end
         else
           begin
              emit(op,opsize,
                getreferenzstring(p^.left^.location.referenz)+','+regid2str[hregister]);
           end;
         maybe_rangechecking(p,p^.left^.resulttype,p^.resulttype);
         p^.location.loc:=LOC_REGISTER;
         p^.location.register:=hregister;
      end;

    procedure second_string_string(p,hp : ptree;convtyp : tconverttype);far;

      var
         pushedregs : tpushed;

      begin
         clear_referenz(p^.location.referenz);
         p^.location.referenz.base:=R_EBP;
         p^.location.referenz.offset:=gettempofsize(p^.resulttype^.size);
         del_referenz(p^.left^.location.referenz);
         pushusedregisters(pushedregs,$ff);
         emitpushreferenzaddr(p^.location.referenz);
         emitpushreferenzaddr(p^.left^.location.referenz);
         emit(PUSH,S_L,'$'+tostr(pstringdef(p^.resulttype)^.len));
         emit(CALL,S_NO,'STRCOPY');
         maybe_loadesi;
         popusedregisters(pushedregs);
      end;

    procedure second_cstring_charpointer(p,hp : ptree;convtyp : tconverttype);far;

      begin
         p^.location.loc:=LOC_REGISTER;
         p^.location.register:=getregister32;
         inc(p^.left^.location.referenz.offset);
         emit(LEA,S_L,getreferenzstring(p^.left^.location.referenz)+
           ','+regid2str[p^.location.register]);
      end;

    procedure second_cchar_charpointer(p,hp : ptree;convtyp : tconverttype);far;

      begin
         {!!!!}
         p^.location.loc:=LOC_REGISTER;
         p^.location.register:=getregister32;
         inc(p^.left^.location.referenz.offset);
         emit(LEA,S_L,getreferenzstring(p^.left^.location.referenz)+
           ','+regid2str[p^.location.register]);
      end;

    procedure second_cstring_chararray(p,hp : ptree;convtyp : tconverttype);far;

      begin
         p^.location.referenz:=p^.left^.location.referenz;
         inc(p^.location.referenz.offset);
      end;

    procedure second_array_to_pointer(p,hp : ptree;convtyp : tconverttype);far;

      begin
         del_referenz(p^.left^.location.referenz);
         p^.location.loc:=LOC_REGISTER;
         p^.location.register:=getregister32;
         emit(LEA,S_L,getreferenzstring(p^.left^.location.referenz)+','+
           regid2str[p^.location.register]);
      end;

    procedure second_pointer_to_array(p,hp : ptree;convtyp : tconverttype);far;

      begin
         p^.location.loc:=LOC_REFERENZ;
         clear_referenz(p^.location.referenz);
         if p^.left^.location.loc=LOC_REGISTER then
           p^.location.referenz.base:=p^.left^.location.register
         else
           begin
              del_referenz(p^.left^.location.referenz);
              p^.location.referenz.base:=getregister32;
              emit(MOV,S_L,getreferenzstring(p^.left^.location.referenz)+','+
                regid2str[p^.location.referenz.base]);
           end;
      end;

    procedure second_char_to_string(p,hp : ptree;convtyp : tconverttype);far;

      begin
         clear_referenz(p^.location.referenz);
         p^.location.referenz.offset:=gettempofsize(256);
         p^.location.referenz.base:=R_EBP;
         if p^.left^.treetype=ordconstn { eigentlich charconst } then
           emit(MOV,S_W,'$'+tostr(p^.left^.value*256+1)+','+getreferenzstring(p^.location.referenz))
         else
           begin
              { nicht so elegant (ging besser mit extra Register }
              if p^.left^.location.loc=LOC_REGISTER then
                begin
                   emit(MOVZX,S_BW,regid2str[p^.left^.location.register]+',%di');
                   ungetregister32(reg8toreg32(p^.left^.location.register));
                end
              else
                begin
                   emit(MOVZX,S_BW,getreferenzstring(p^.left^.location.referenz)+',%di');
                   del_referenz(p^.left^.location.referenz);
                end;
              emit(A_SHL,S_W,'$8,%di');
              emit(A_OR,S_W,'$1,%di');
              emit(MOV,S_W,'%di,'+getreferenzstring(p^.location.referenz));
           end;
      end;

    procedure second_int_real(p,hp : ptree;convtyp : tconverttype);far;

      var
         hs : string;

      begin
         if p^.left^.location.loc=LOC_REGISTER then
           begin
              hs:=regid2str[p^.left^.location.register];
              ungetregister(p^.left^.location.register);
           end
         else
           begin
              hs:=getreferenzstring(p^.left^.location.referenz);
              del_referenz(p^.left^.location.referenz);
           end;
         hs:=hs+',%edi';
         case pgrunddef(p^.left^.resulttype)^.typ of
            s8bit : emit(MOVSX,S_BL,hs);
            u8bit : emit(MOVZX,S_BL,hs);
            s16bit : emit(MOVSX,S_WL,hs);
            u16bit : emit(MOVZX,S_WL,hs);
            s32bit : emit(MOV,S_L,hs);
         end;
         emit(FILD,S_L,'%edi');
         p^.location.loc:=LOC_FPUSTACK;
      end;

    procedure second_smaller(p,hp : ptree;convtyp : tconverttype);far;

      var
         destregister : tregister;
         opsize : topsize;
         ref : boolean;

      begin
         { !!!!!!!! Rangechecking }
         ref:=false;
         if p^.left^.location.loc=LOC_REGISTER then
           begin
              destregister:=p^.location.register;
              case convtyp of
                 tc_s32bit_2_s8bit,
                 tc_s32bit_2_u8bit : destregister:=reg32toreg8(destregister);
                 tc_s32bit_2_s16bit,
                 tc_s32bit_2_u16bit : destregister:=reg32toreg16(destregister);
                 tc_s16bit_2_s8bit,
                 tc_s16bit_2_u8bit,
                 tc_u16bit_2_s8bit,
                 tc_u16bit_2_u8bit : destregister:=reg16toreg8(destregister);
              end;
              p^.location.register:=destregister;
           end;
      end;

    procedure secondtypeconv(var p : ptree);far;

      const
         secondconvert : array[tc_u8bit_2_s32bit..tc_int_2_real] of
           tsecondconvproc = (second_bigger,second_only_rangecheck,
                              second_bigger,second_bigger,second_bigger,
                              second_smaller,second_smaller,
                              second_smaller,second_string_string,
                              second_cstring_charpointer,second_cstring_chararray,
                              second_array_to_pointer,second_pointer_to_array,
                              second_char_to_string,second_bigger,
                              second_bigger,second_bigger,
                              second_smaller,second_smaller,
                              second_smaller,second_smaller,
                              second_bigger,second_smaller,
                              second_int_real); {,second_cchar_charpointer); }

      begin
         secondpass(p^.left);
         p^.location:=p^.left^.location;
         if not(p^.explizit) then
           begin
              if p^.convtyp<>tc_equal then
                secondconvert[p^.convtyp](p,p,p^.convtyp)
              else
                p^.resulttype:=p^.left^.resulttype;
           end;
      end;

    { *************** Funktionshandling **************** }

    { dreht die Parameterliste um }

    function reversparameter(p : ptree) : ptree;

      var
         hp1,hp2 : ptree;

      begin
         hp1:=nil;
         while assigned(p) do
           begin
              { Aushaengen }
              hp2:=p;
              p:=p^.right;

              { Neu Einhaengen }
              hp2^.right:=hp1;
              hp1:=hp2;
           end;
         reversparameter:=hp1;
      end;

    procedure firstcallparan(var p : ptree;defcoll : pdefcoll);

      begin
         if assigned(p^.right) then
           begin
              if defcoll=nil then
                firstcallparan(p^.right,nil)
                else firstcallparan(p^.right,defcoll^.next);
              p^.registers32:=p^.right^.registers32;
           end;
         if defcoll=nil then
           begin
              firstpass(p^.left);

              if codegenerror then
                exit;

              p^.resulttype:=p^.left^.resulttype;
           end
         { falls aufzurufendes Unterprogramm schon bekannt, dann }
         { Typkonvertierungen einfgen                           }
         else
           begin
              if not((p^.left^.resulttype^.deftype=stringdef) and
                     (defcoll^.data^.deftype=stringdef)) and
                     (defcoll^.data^.deftype<>formaldef) then
                begin
                   if (defcoll^.paratyp=vs_var) and
                     not(is_equal(p^.left^.resulttype,defcoll^.data)) then
                     begin
                        error(call_by_ref_without_typeconv);
                        exit;
                     end;
                   p^.left:=gentypeconvnode(p^.left,defcoll^.data);
                   firstpass(p^.left);

                   if codegenerror then
                     exit;

                end;
              p^.resulttype:=defcoll^.data;
           end;
         if p^.left^.registers32>p^.registers32 then
           p^.registers32:=p^.left^.registers32;
      end;

    { speichert die Gre der gepushten Parameter }
    var
       pushedparasize : longint;

    procedure secondcallparan(var p : ptree;defcoll : pdefcoll);

      var
         regstr : string[3];
         size : longint;
         stackref : treferenz;
         otlabel,hlabel,oflabel : longint;

         { Temporre Variablen: }
         tempdeftype : tdeftype;
         tempreferenz : treferenz;

      begin
         otlabel:=truelabel;
         oflabel:=falselabel;
         truelabel:=getlabel;
         falselabel:=getlabel;
         secondpass(p^.left);
         if (defcoll^.data^.deftype=formaldef) then
           begin
              if (p^.left^.location.loc<>LOC_REFERENZ) and
                 (p^.left^.location.loc<>LOC_MEM) then
                error(type_mismatch);
              emitpushreferenzaddr(p^.left^.location.referenz);
              inc(pushedparasize,4);
           end
         else if (defcoll^.paratyp=vs_var) then
           begin
              if (p^.left^.location.loc<>LOC_REFERENZ) then
                error(var_must_be_referenz);
              emitpushreferenzaddr(p^.left^.location.referenz);
              inc(pushedparasize,4);
           end
         else
           begin
              tempdeftype:=p^.resulttype^.deftype;
              if tempdeftype=filedef then
                error(file_must_call_by_referenz);
              if (defcoll^.paratyp=vs_const) and
                 (
                   (tempdeftype=stringdef) or
                   (tempdeftype=arraydef) or
                   (tempdeftype=recorddef) or
                   (tempdeftype=classdef)
                 ) then
                begin
                   emitpushreferenzaddr(p^.left^.location.referenz);
                   inc(pushedparasize,4);
                end
              else
                case p^.left^.location.loc of
                   LOC_REGISTER : begin
                                     case p^.left^.location.register of
                                        R_EAX,R_EBX,R_ECX,R_EDX,R_ESI,
                                        R_EDI,R_ESP,R_EBP :
                                          begin
                                             emit(PUSH,S_L,regid2str[p^.left^.location.register]);
                                             inc(pushedparasize,4);
                                             ungetregister32(p^.left^.location.register);
                                          end;
                                        R_AX,R_BX,R_CX,R_DX,R_SI,R_DI : begin
                                              emit(PUSH,S_W,regid2str[p^.left^.location.register]);
                                              inc(pushedparasize,2);
                                              ungetregister32(reg16toreg32(p^.left^.location.register));
                                           end;
                                        R_AL,R_BL,R_CL,R_DL:
                                          begin
                                             { hier etwas Schnipseln um die 8-Bit }
                                             { Register auf den Stack zu bringen: }
                                             regstr:=regid2str[p^.left^.location.register];
                                             emit(PUSH,S_W,'%'+regstr[2]+'x');
                                             inc(pushedparasize,2);
                                             ungetregister32(reg8toreg32(p^.left^.location.register));
                                          end;
                                     end;
                                  end;
                   LOC_REFERENZ,LOC_MEM :
                               begin
                                  del_referenz(p^.left^.location.referenz);
                                  tempreferenz:=p^.left^.location.referenz;
                                  case p^.resulttype^.deftype of
                                     grunddef : begin
                                                   case pgrunddef(p^.resulttype)^.typ of
                                                      s32bit :
                                                        begin
                                                           emit(PUSH,S_L,getreferenzstring(tempreferenz));
                                                           inc(pushedparasize,4);
                                                        end;
                                                      s8bit,u8bit,uchar,bool8bit,s16bit,u16bit : begin
                                                          emit(PUSH,S_W,getreferenzstring(tempreferenz));
                                                          inc(pushedparasize,2);
                                                       end;
                                                      s64real : begin
                                                                   inc(tempreferenz.offset,4);
                                                                   emit(PUSH,S_L,getreferenzstring(tempreferenz));
                                                                   dec(tempreferenz.offset,4);
                                                                   emit(PUSH,S_L,getreferenzstring(tempreferenz));
                                                                   inc(pushedparasize,8);
                                                                end;
                                                   end;
                                                end;
                                     aufzaehldef : begin
                                                      emit(PUSH,S_L,getreferenzstring(tempreferenz));
                                                      inc(pushedparasize,4);
                                                   end;
                                     arraydef,recorddef,stringdef :
                                                begin
                                                   size:=p^.resulttype^.size;

                                                   { Alignment }
                                                   if (size>=4) and ((size and 3)<>0) then
                                                     inc(size,4-(size and 3))
                                                   else if (size>=2) and ((size and 1)<>0) then
                                                     inc(size,2-(size and 1))
                                                   else if size=1 then size:=2;

                                                   { Stackraum schaffen }
                                                   emit(SUB,S_L,'$'+tostr(size)+',%esp');
                                                   inc(pushedparasize,size);
                                                   { Stackreferenz erstellen }
                                                   clear_referenz(stackref);
                                                   stackref.base:=R_ESP;
                                                   { Kopie erzeugen }
                                                   if p^.resulttype^.deftype=stringdef then
                                                     begin
                                                        copystring(stackref,p^.left^.location.referenz,
                                                          pstringdef(p^.resulttype)^.len);
                                                     end
                                                   else
                                                     begin
                                                        concatcopy(p^.left^.location.referenz,
                                                          stackref,p^.resulttype^.size,true);
                                                     end;
                                                end;
                                     pointerdef : begin
                                                     emit(PUSH,S_L,getreferenzstring(p^.left^.location.referenz));
                                                     inc(pushedparasize,4);
                                                  end;
                                     {!!!!!!!!!}
                                     else error(error_in_expression);
                                  end;
                               end;
                 LOC_JUMP     : begin
                                   hlabel:=getlabel;
                                   inc(pushedparasize,2);
                                   emitl(A_LABEL,truelabel);
                                   emit(PUSH,S_W,'$1');
                                   emitl(JMP,hlabel);
                                   emitl(A_LABEL,falselabel);
                                   emit(PUSH,S_W,'$0');
                                   emitl(A_LABEL,hlabel);
                                end;
                 LOC_FLAGS    : begin
                                   emit(A_XOR,S_W,'%ax,%ax');
                                   emit(flag_2_set[p^.right^.location.resflags],S_NO,
                                     '%al');
                                   inc(pushedparasize,2);
                                   emit(PUSH,S_W,'%ax');
                                end;
                end;
           end;
         truelabel:=otlabel;
         falselabel:=oflabel;
         { von rechts nach links pushen }
         if assigned(p^.right) then
           secondcallparan(p^.right,defcoll^.next);
      end;

    procedure firstcalln(var p : ptree);far;

      type
         pprocdefcoll = ^tprocdefcoll;

         tprocdefcoll = record
            data : pprocdef;
            nextpara : pdefcoll;
            next : pprocdefcoll;
         end;

      var
         hp,procs,hp2 : pprocdefcoll;
         pd : pprocdef;
         pt : ptree;
         exactmatch : boolean;
         paralength,l : longint;
         pdc : pdefcoll;

         { nur ein Dummy }
         hcvt : tconverttype;

      { types.is_euqal darf keine formaldef's behandeln !}

      function is_equal(def1,def2 : pdef) : boolean;

        begin
           { alle Typen knnen an formaldef bergeben werden }
           is_equal:=(def1^.deftype=formaldef) or
             types.is_equal(def1,def2);
        end;

      begin
         { Register freigeben! }
         { falls procdefinition<>nil, dann wurde schon firstpass }
         { aufgerufen                                            }
         { scheint nicht so gut wegen der Register }
         { if assigned(p^.procdefinition) then
           exit; }

         if assigned(p^.left) then
           begin
              firstcallparan(p^.left,nil);
              if codegenerror then
                exit;
           end;

         { Lnge der Parameterliste feststellen }
         pt:=p^.left;
         paralength:=0;
         while assigned(pt) do
           begin
              inc(paralength);
              pt:=pt^.right;
           end;

         { alle in Frage kommenden Prozeduren in eine }
         { verkettete Liste einfgen                  }
         pd:=p^.symtableprocentry^.definition;
         procs:=nil;
         while assigned(pd) do
           begin
              { Laenge der deklarierten Parameterliste feststellen: }
              pdc:=pd^.para1;
              l:=0;
              while assigned(pdc) do
                begin
                   inc(l);
                   pdc:=pdc^.next;
                end;
              { nur wenn die Parameterlnge pat, dann Einfgen }
              if l=paralength then
                begin
                   new(hp);
                   hp^.data:=pd;
                   hp^.next:=procs;
                   hp^.nextpara:=pd^.para1;
                   procs:=hp;
                end;
              pd:=pd^.nextoverloaded;
           end;

         { nun alle Parameter nacheinander vergleichen }
         pt:=p^.left;
         while assigned(pt) do
           begin

              { pat der Parameter irgendwo exakt? }
              exactmatch:=false;
              hp:=procs;
              while assigned(hp) do
                begin
                   if is_equal(hp^.nextpara^.data,pt^.resulttype) then
                     exactmatch:=true;
                   hp:=hp^.next;
                end;
              { ja, dann alle anderen Prozeduren entfernen }
              if exactmatch then
                begin
                   { erst am Anfang }
                   while (assigned(procs)) and not(is_equal(procs^.nextpara^.data,pt^.resulttype)) do
                     begin
                        hp:=procs^.next;
                        dispose(procs);
                        procs:=hp;
                     end;
                   { und jetzt aus der Mitte }
                   hp:=procs;
                   while (assigned(hp)) and assigned(hp^.next) do
                     begin
                        if not(is_equal(hp^.next^.nextpara^.data,pt^.resulttype)) then
                          begin
                             hp2:=hp^.next^.next;
                             dispose(hp^.next);
                             hp^.next:=hp2;
                          end
                        else
                          hp:=hp^.next;
                     end;
                end
              { sollte nirgendwo ein Parameter exakt passen, }
              { so alle Prozeduren entfernen, bei denen      }
              { der Parameter auch nach einer impliziten     }
              { Typkonvertierung nicht passt                 }
              else
                begin
                   { erst am Anfang }
                   while (assigned(procs)) and
                     not(isconvertable(pt^.resulttype,procs^.nextpara^.data,hcvt,pt^.left^.treetype)) do
                     begin
                        hp:=procs^.next;
                        dispose(procs);
                        procs:=hp;
                     end;
                   { und jetzt aus der Mitte }
                   hp:=procs;
                   while (assigned(hp)) and assigned(hp^.next) do
                     begin
                        if not(isconvertable(pt^.resulttype,hp^.next^.nextpara^.data,
                          hcvt,pt^.left^.treetype)) then
                          begin
                             hp2:=hp^.next^.next;
                             dispose(hp^.next);
                             hp^.next:=hp2;
                          end
                        else
                          hp:=hp^.next;
                     end;
                end;
              { nun bei denn Prozeduren den nextpara-Zeiger auf den }
              { naechsten Parameter setzten                         }
              hp:=procs;
              while assigned(hp) do
                begin
                   hp^.nextpara:=hp^.nextpara^.next;
                   hp:=hp^.next;
                end;
              pt:=pt^.right;
           end;

         if procs=nil then
           begin
              error(no_para_match);
              exit;
           end;

         if assigned(procs^.next) then
           error(too_much_matches);
         p^.procdefinition:=procs^.data;
         p^.resulttype:=procs^.data^.retdef;
         p^.location.loc:=LOC_MEM;
         { nochmal die Parameter beackern, um die Typkonvertiereungen }
         { einbauen zu koennen                                        }
         if assigned(p^.left) then
           firstcallparan(p^.left,p^.procdefinition^.para1);

         { interne Proceduren bearbeiten }
         if (p^.procdefinition^.options and pointernproc)<>0 then
           begin
              pt:=geninlinenode(p^.procdefinition^.extnumber,p^.left^.left);
              if assigned(p^.left^.right) then
                disposetree(p^.left^.right);
              putnode(p^.left);
              putnode(p);
              firstpass(pt);

              if codegenerror then
                exit;

              p:=pt;
              exit;
           end;
         { und falls noetig ein Register fuer den Returnwert bereitstellen }
         if (p^.resulttype<>pdef(voiddef)) then
           begin
              p^.location.loc:=LOC_REGISTER;
              if ((p^.resulttype^.deftype=procdef) or
                 (p^.resulttype^.deftype=aufzaehldef) or
                 (p^.resulttype^.deftype=pointerdef)) then
                 p^.registers32:=1
              else if (p^.resulttype^.deftype=grunddef) then
                begin
                   if pgrunddef(p^.resulttype)^.typ=s64real then
                     begin
                        p^.registersfpu:=1;
                        p^.location.loc:=LOC_FPUSTACK;
                     end
                   else p^.registers32:=1
                end;
           end;
         { noch falls ntig Classpointer berechen }
         if p^.methodpointer<>nil then
           begin
              firstpass(p^.methodpointer);
           end;
      end;

    procedure pushexceptlabel;

      begin
         if not(procinfo.exceptions) then
           error(proc_must_handleexceptions);
         emit(PUSH,S_L,'$'+tolabel(aktexceptlabel));
      end;

    procedure secondcalln(var p : ptree);far;

      var
         unusedregisters : tregisterset;
         asmrec : pasmrec;
         pushed : tpushed;
         funcretref : treferenz;
         hregister : tregister;
         oldpushedparasize : longint;
         loadesi : boolean;

      begin
         loadesi:=true;
         unusedregisters:=unused;
         { zuerst alle benutzen Register pushen: }
         pushusedregisters(pushed,p^.procdefinition^.usedregisters);

         { benutzte Register durchreichen: }
         usedinproc:=usedinproc or p^.procdefinition^.usedregisters;

         { Parameter berechnen und pushen }
         oldpushedparasize:=pushedparasize;
         if assigned(p^.left) then
           begin
              pushedparasize:=0;
              secondcallparan(p^.left,p^.procdefinition^.para1);
           end;
         if (p^.resulttype<>pdef(voiddef)) and
            ((p^.resulttype^.deftype=arraydef) or
            (p^.resulttype^.deftype=stringdef) or
            (p^.resulttype^.deftype=classdef) or
            (p^.resulttype^.deftype=recorddef)) then
           begin
              clear_referenz(funcretref);
              funcretref.base:=R_EBP;
              funcretref.offset:=gettempofsize(p^.procdefinition^.retdef^.size);
              emitpushreferenzaddr(funcretref);
              inc(pushedparasize,4);
           end;

         { self pushen }

         if (p^.symtable^.symtabletype and $3fff)=objectsymtable then
           begin
              if assigned(p^.methodpointer) then
                begin
                   { irgendein Methodenaufruf }
                   secondpass(p^.methodpointer);
                   case p^.methodpointer^.location.loc of
                      LOC_REGISTER : begin
                                        ungetregister32(p^.methodpointer^.location.register);
                                        emit(PUSH,S_L,regid2str[p^.methodpointer^.location.register]);
                                     end;
                      else begin
                              emitpushreferenzaddr(p^.methodpointer^.location.referenz);
                              del_referenz(p^.methodpointer^.location.referenz);
                           end;
                   end;
                end
              else
                begin
                   { Memberaufruf, also ESI nicht verndern }
                   loadesi:=false;
                   emit(PUSH,S_L,'%esi');
                end;
           end;

         { Basepointer ?}
         if (lexlevel>0) and assigned(p^.procdefinition^.parast) and
             ((p^.procdefinition^.parast^.symtabletype and $3fff)>1) then
           begin
              if (lexlevel=(p^.procdefinition^.parast^.symtabletype and $3fff)) then
                emit(PUSH,S_L,tostr(procinfo.framepointer)+'(%ebp)')
              else if lexlevel<(p^.procdefinition^.parast^.symtabletype and $3fff) then
                emit(PUSH,S_L,'%ebp');
           end;

         { Exceptionadresse pushen }
         if ((p^.procdefinition^.options and poexceptions)<>0) then
           begin
              pushexceptlabel;
              inc(pushedparasize,4);
           end;

         if (p^.procdefinition^.options and poexports)<>0 then
           error(dont_call_exported_direct);

         if (p^.procdefinition^.options and povirtualmethod)<>0 then
           begin
              emit(MOV,S_L,'(%esi),%edi');
              emit(CALL,S_NO,tostr(p^.procdefinition^.extnumber*4+12)+'(%edi)');
           end
         else
            emit(CALL,S_NO,p^.procdefinition^.mangledname);
         if ((p^.procdefinition^.options and poclearstack)<>0) then
           emit(ADD,S_L,'$'+tostr(pushedparasize)+',%esp');
         pushedparasize:=oldpushedparasize;
         unused:=unusedregisters;
         { Register restaurieren }
         if p^.resulttype<>pdef(voiddef) then
           begin
              if ((p^.resulttype^.deftype=arraydef) or
                  (p^.resulttype^.deftype=stringdef) or
                  (p^.resulttype^.deftype=classdef) or
                  (p^.resulttype^.deftype=recorddef)) then
                begin
                   p^.location.loc:=LOC_MEM;
                   p^.location.referenz:=funcretref;
                end
              else
                begin
                   p^.location.loc:=LOC_REGISTER;
                   hregister:=getregister32;
                   if (p^.resulttype^.deftype=grunddef) then
                     begin
                        case pgrunddef(p^.resulttype)^.typ of
                           s32bit : begin
                                       if hregister<>R_EAX then
                                         emit(MOV,S_L,'%eax,'+regid2str[hregister]);
                                       p^.location.register:=hregister;
                                    end;
                           uchar,u8bit,bool8bit,s8bit : begin
                                       if hregister<>R_EAX then
                                         emit(MOV,S_B,'%al,'+regid2str[reg32toreg8(hregister)]);
                                       p^.location.register:=reg32toreg8(hregister);
                                    end;
                           s16bit,u16bit : begin
                                       if hregister<>R_EAX then
                                         emit(MOV,S_W,'%ax,'+regid2str[reg32toreg16(hregister)]);
                                       p^.location.register:=reg32toreg16(hregister);
                                    end;
                           s64real : p^.location.loc:=LOC_FPUSTACK;
                           else internalerror(7);
                        end
                     end
                   else
                     begin
                        if hregister<>R_EAX then
                          emit(MOV,S_L,'%eax,'+regid2str[hregister]);
                        p^.location.register:=hregister;
                     end;
                end;
           end;
         popusedregisters(pushed);
         if loadesi then
           maybe_loadesi;
      end;

    procedure firstfuncret(var p : ptree);far;

      begin
         p^.resulttype:=procinfo.retdef;
         p^.location.loc:=LOC_REFERENZ;
         if (procinfo.retdef^.deftype=arraydef) or
            (procinfo.retdef^.deftype=stringdef) or
            (procinfo.retdef^.deftype=recorddef) or
            (procinfo.retdef^.deftype=classdef) then
            p^.registers32:=1;
      end;

    procedure secondfuncret(var p : ptree);far;

      var
         hregister : tregister;

      begin
         clear_referenz(p^.location.referenz);
         p^.location.referenz.base:=R_EBP;
         p^.location.referenz.offset:=procinfo.retoffset;
         if (procinfo.retdef^.deftype=arraydef) or
            (procinfo.retdef^.deftype=stringdef) or
            (procinfo.retdef^.deftype=classdef) or
            (procinfo.retdef^.deftype=recorddef) then
           begin
              hregister:=getregister32;
              emit(MOV,S_L,getreferenzstring(p^.location.referenz)+','+regid2str[hregister]);
              p^.location.referenz.base:=hregister;
              p^.location.referenz.offset:=0;
           end;
      end;

    { interne Inlineprozeduren }

    procedure firstinline(var p : ptree);far;

      var
         hp : ptree;

      begin
         p^.registers32:=p^.left^.registers32;
         p^.registersfpu:=p^.left^.registersfpu;
         p^.location:=p^.left^.location;
         case p^.inlinenumber of
            in_lo_word,in_hi_word : begin
                       if p^.registers32<1 then
                         p^.registers32:=1;
                       p^.resulttype:=u8bitdef;
                       p^.location.loc:=LOC_REGISTER;
                    end;
            in_lo_long,in_hi_long : begin
                       if p^.registers32<1 then
                         p^.registers32:=1;
                       p^.resulttype:=u16bitdef;
                       p^.location.loc:=LOC_REGISTER;
                    end;
            in_ord_char : begin
                       p^.resulttype:=u8bitdef;
                    end;
            in_chr_byte : begin
                       p^.resulttype:=cchardef;
                    end;
            in_length_string : begin
                       p^.resulttype:=u8bitdef;
                       { String nach Stringkonvertierungen brauchen wir hier nicht }
                       if (p^.left^.treetype=typeconvn) and
                          (p^.left^.left^.resulttype^.deftype=stringdef) then
                         begin
                            hp:=p^.left^.left;
                            putnode(p^.left);
                            p^.left:=hp;
                         end;
                    end;
            in_assigned_x : begin
                               p^.resulttype:=booldef;
                               p^.location.loc:=LOC_FLAGS;
                            end;
            in_dec_dword,
            in_dec_word,
            in_dec_byte,
            in_inc_dword,
            in_inc_word,
            in_inc_byte : begin
                             p^.resulttype:=voiddef;
                             if p^.left^.location.loc<>LOC_REFERENZ then
                               error(error_in_expression);
                          end;
            in_read_x,
            in_readln_x,
            in_write_x,
            in_writeln_x : begin
                              p^.resulttype:=voiddef;
                              if assigned(p^.left) then
                                begin
                                   firstcallparan(p^.left,nil);
                                   { Typkonvertierungen fr write(ln) einfgen }
                                   if (p^.inlinenumber=in_write_x) or (p^.inlinenumber=in_writeln_x) then
                                     begin
                                        hp:=p^.left;
                                        while assigned(hp) do
                                          begin
                                             if hp^.left^.resulttype^.deftype=grunddef then
                                               case pgrunddef(hp^.left^.resulttype)^.typ of
                                                  u8bit,s8bit,
                                                  u16bit,s16bit :
                                                        hp^.left:=gentypeconvnode(hp^.left,s32bitdef);
                                               end;
                                             hp:=hp^.right;
                                          end;
                                     end;
                                   { nochmals alle Parameter bearbeiten }
                                   firstcallparan(p^.left,nil);
                                end;
                           end;
            else internalerror(8);
         end;
      end;

    procedure secondinline(var p : ptree);far;

      var
         aktfile : treferenz;
         ft : tfiletyp;

      { Behandelt die Codeerzeugung fr READ(LN) und WRITE(LN) }

      procedure handlereadwrite(doread,callwriteln : boolean);

        var
           node,hp : ptree;
           dummycoll : tdefcoll;
           typedtyp : pdef;
           doflush : boolean;

        begin
           { kein  automatischer Aufruf von Flush }
           doflush:=false;
           { temporren Zeiger auf Dateivariable reservieren }
           gettempofsizereferenz(4,aktfile);
           { erst einmal Textdatei angeben }
           ft:=ft_text;
           { auch ein Parameter angegeben ? }
           if p^.left=nil then
             begin
                { Bildschirmausgabe, also gleich ausgeben }
                doflush:=true;
                { die folgenden Anweisungen sind fr "writeln;" }
                if target=target_DOS then
                  begin
                     if doread then
                       emit(LEA,S_L,'U_SYSTEM_INPUT,%edi')
                     else emit(LEA,S_L,'U_SYSTEM_OUTPUT,%edi');
                  end
                else if target=target_OS2 then
                  begin
                     if doread then
                       emit(LEA,S_L,'U_SYSOS2_INPUT,%edi')
                     else emit(LEA,S_L,'U_SYSOS2_OUTPUT,%edi');
                  end;
                { @Dateivariable in temp. Var. abspeichern }
                emit(MOV,S_L,'%edi,'+getreferenzstring(aktfile));
                exit;
             end;

           { Parameter umdrehen ! }
           node:=reversparameter(p^.left);

           { Dateivariable berechnen }
           { erster Parameter ein Filetyp ? }
           if node^.left^.resulttype^.deftype=filedef then
             begin
                ft:=pfiledef(node^.left^.resulttype)^.filetyp;
                if ft=ft_typed then
                  typedtyp:=pfiledef(node^.left^.resulttype)^.typed_as;

                secondpass(node^.left);
                if codegenerror then
                  exit;

                { Referenz in temporre Variable retten }
                if node^.left^.location.loc<>LOC_REFERENZ then
                  begin
                     error(error_in_expression);
                     exit;
                  end;

                emit(LEA,S_L,getreferenzstring(node^.left^.location.referenz)+',%edi');

                { erster Parameter ist behandelt, also lschen }
                node:=node^.right;
             end
           else
             begin
                { Bildschirmausgabe, also gleich ausgeben }
                doflush:=true;
                if target=target_DOS then
                  begin
                     if doread then
                       emit(LEA,S_L,'U_SYSTEM_INPUT,%edi')
                     else emit(LEA,S_L,'U_SYSTEM_OUTPUT,%edi');
                  end
                else if target=target_OS2 then
                  begin
                     if doread then
                       emit(LEA,S_L,'U_SYSOS2_INPUT,%edi')
                     else emit(LEA,S_L,'U_SYSOS2_OUTPUT,%edi');
                  end;
             end;

           { @Dateivariable in temp. Var. abspeichern }
           emit(MOV,S_L,'%edi,'+getreferenzstring(aktfile));
           if doread then
             { Paramter an READ werden Call by Referenz bergeben }
             dummycoll.paratyp:=vs_var
             { an WRITE Call by "Const" }
           else dummycoll.paratyp:=vs_const;
           while node<>nil do
             begin
                hp:=node;
                node:=node^.right;
                hp^.right:=nil;

                secondcallparan(hp,@dummycoll);
                if codegenerror then
                  exit;

                emit(PUSH,S_L,getreferenzstring(aktfile));
                if (ft=ft_typed) then
                  begin
                     {!!!!!!!}
                  end
                else
                  begin
                     pushexceptlabel;
                     case hp^.left^.resulttype^.deftype of
                        stringdef : begin
                                       if doread then
                                         emit(CALL,S_NO,'READ_TEXT_STRING')
                                       else emit(CALL,S_NO,'WRITE_TEXT_STRING');
                                    end;
                        pointerdef : begin
                                        if (ppointerdef(hp^.left^.resulttype)^.definition^.deftype=grunddef)
                                          and (pgrunddef(ppointerdef(hp^.left^.resulttype)^.definition)^.typ=uchar) then
                                          begin
                                             if doread then
                                               emit(CALL,S_NO,'READ_TEXT_PCHAR_AS_POINTER')
                                             else emit(CALL,S_NO,'WRITE_TEXT_PCHAR_AS_POINTER');
                                          end
                                        else error(no_para_match);
                                     end;
                        arraydef : begin
                                      if (parraydef(hp^.left^.resulttype)^.lowrange=0)
                                        and (parraydef(hp^.left^.resulttype)^.definition^.deftype=grunddef)
                                        and (pgrunddef(parraydef(hp^.left^.resulttype)^.definition)^.typ=uchar) then
                                        begin
                                           if doread then
                                             emit(CALL,S_NO,'READ_TEXT_PCHAR_AS_ARRAY')
                                           else emit(CALL,S_NO,'WRITE_TEXT_PCHAR_AS_ARRAY');
                                        end
                                      else error(no_para_match);
                                   end;
                        grunddef : begin
                                      case pgrunddef(hp^.left^.resulttype)^.typ of
                                         s32bit : if doread then
                                                    emit(CALL,S_NO,'READ_TEXT_LONGINT')
                                                  else emit(CALL,S_NO,'WRITE_TEXT_LONGINT');
                                         uchar : if doread then
                                                    emit(CALL,S_NO,'READ_TEXT_CHAR')
                                                  else emit(CALL,S_NO,'WRITE_TEXT_CHAR');
                                         else error(no_para_match);
                                      end;
                                   end;
                        else error(no_para_match);
                     end;
                  end;
             end;
           if callwriteln then
             begin
                emit(PUSH,S_L,getreferenzstring(aktfile));
                pushexceptlabel;
                if ft<>ft_text then
                  error(no_para_match);
                emit(CALL,S_NO,'WRITELN_TEXT');
             end;
           if doflush and not(doread) then
             begin
               pushexceptlabel;
               emit(CALL,S_NO,'FLUSH_STDOUT');
             end;
           { in Methoden ESI neu laden }
           maybe_loadesi;
        end;

      begin
         case p^.inlinenumber of
            in_lo_word,
            in_hi_word : begin
                       secondpass(p^.left);
                       p^.location.loc:=LOC_REGISTER;
                       if p^.left^.location.loc<>LOC_REGISTER then
                         begin
                            del_referenz(p^.left^.location.referenz);
                            p^.location.register:=reg32toreg16(getregister32);
                            emit(MOV,S_W,getreferenzstring(p^.left^.location.referenz)
                              +','+regid2str[p^.location.register]);
                         end
                       else p^.location.register:=p^.left^.location.register;
                       if p^.inlinenumber=in_hi_word then
                         emit(A_SHR,S_W,'$8,'+regid2str[p^.location.register]);
                       p^.location.register:=reg16toreg8(p^.location.register);
                    end;
            in_lo_long,
            in_hi_long : begin
                       secondpass(p^.left);
                       p^.location.loc:=LOC_REGISTER;
                       if p^.left^.location.loc<>LOC_REGISTER then
                         begin
                            del_referenz(p^.left^.location.referenz);
                            p^.location.register:=getregister32;
                            emit(MOV,S_L,getreferenzstring(p^.left^.location.referenz)
                              +','+regid2str[p^.location.register]);
                         end
                       else p^.location.register:=p^.left^.location.register;
                       if p^.inlinenumber=in_hi_long then
                         emit(A_SHR,S_L,'$16,'+regid2str[p^.location.register]);
                       p^.location.register:=reg32toreg16(p^.location.register);
                    end;
            in_ord_char,
            in_chr_byte,
            in_length_string : begin
                       secondpass(p^.left);
                       p^.location:=p^.left^.location;
                    end;
            in_inc_dword : begin
                              secondpass(p^.left);
                              emit(A_INC,S_L,getreferenzstring(p^.left^.location.referenz));
                              emitoverflowcheck;
                           end;
            in_inc_word : begin
                              secondpass(p^.left);
                              emit(A_INC,S_W,getreferenzstring(p^.left^.location.referenz));
                              emitoverflowcheck;
                           end;
            in_inc_byte : begin
                              secondpass(p^.left);
                              emit(A_INC,S_B,getreferenzstring(p^.left^.location.referenz));
                              emitoverflowcheck;
                           end;
            in_dec_dword : begin
                              secondpass(p^.left);
                              emit(A_DEC,S_L,getreferenzstring(p^.left^.location.referenz));
                              emitoverflowcheck;
                           end;
            in_dec_word : begin
                              secondpass(p^.left);
                              emit(A_DEC,S_W,getreferenzstring(p^.left^.location.referenz));
                              emitoverflowcheck;
                           end;
            in_dec_byte : begin
                              secondpass(p^.left);
                              emit(A_DEC,S_B,getreferenzstring(p^.left^.location.referenz));
                              emitoverflowcheck;
                           end;
            in_assigned_x : begin
                               secondpass(p^.left^.left);
                               p^.location.loc:=LOC_FLAGS;
                               if p^.left^.location.loc=LOC_REGISTER then
                                 begin
                                    emit(A_OR,S_L,regid2str[p^.left^.left^.location.register]+','+
                                      regid2str[p^.left^.left^.location.register]);
                                    ungetregister32(p^.left^.left^.location.register);
                                 end
                               else
                                 begin
                                    emit(CMP,S_L,'$0,'+getreferenzstring(p^.left^.left^.location.referenz));
                                    del_referenz(p^.left^.left^.location.referenz);
                                 end;
                               p^.location.resflags:=F_NE;
                            end;
            in_write_x : handlereadwrite(false,false);
            in_writeln_x : handlereadwrite(false,true);
            in_read_x : handlereadwrite(true,false);
            in_readln_x : begin
                              handlereadwrite(true,false);
                              emit(PUSH,S_L,getreferenzstring(aktfile));
                              pushexceptlabel;
                              if ft<>ft_text then
                                error(no_para_match);
                              emit(CALL,S_NO,'READLN_TEXT');
                              maybe_loadesi;
                           end;
            else internalerror(9);
         end;
      end;

    { Recordzugriff }

    procedure firstsubscriptn(var p : ptree);far;

      begin
         firstpass(p^.left);

         if codegenerror then
           exit;

         if (p^.left^.location.loc<>LOC_MEM) and
            (p^.left^.location.loc<>LOC_REFERENZ) then
           error(error_in_expression);
         p^.resulttype:=p^.vs^.definition;
         p^.location:=p^.left^.location;
         p^.registers32:=p^.left^.registers32;
      end;

    procedure secondsubscriptn(var p : ptree);far;

      begin
         secondpass(p^.left);
         p^.location:=p^.left^.location;
         inc(p^.location.referenz.offset,p^.vs^.adresse);
      end;

    procedure firstselfn(var p : ptree);far;

      begin
         p^.location.loc:=LOC_REFERENZ;
      end;

    procedure secondselfn(var p : ptree);far;

      begin
         clear_referenz(p^.location.referenz);
         p^.location.referenz.base:=R_ESI;
      end;

    procedure firstpass(var p : ptree);

      const
         procedures : array[addn..errorn] of firstpassproc =
            (firstadd,firstadd,firstadd,firstmoddiv,
             firstmoddiv,firstassignment,firstload,firstrange,
             firstadd,firstadd,firstadd,firstadd,
             firstadd,firstadd,firstnothing,firstadd,
             firstadd,firstshlshr,firstshlshr,firstnothing,
             firstadd,firstsubscriptn,firstderef,firstaddr,
             firstordconst,firsttypeconv,firstcalln,firstnothing,
             firstrealconst,firstumminus,firstnothing,firstvecn,
             firststringconst,firstfuncret,firstselfn,
             firstnot,firstinline,firstniln,firsterror);

      var
         oldcodegenerror : boolean;

      begin
         oldcodegenerror:=codegenerror;
         codegenerror:=false;
         if not(p^.error) then
           begin
              procedures[p^.treetype](p);
              p^.error:=codegenerror;
              codegenerror:=codegenerror or oldcodegenerror;
           end
         else codegenerror:=true;
      end;

    procedure secondpass(var p : ptree);

      const
         procedures : array[addn..errorn] of secondpassproc =
            (secondadd,secondadd,secondadd,secondmoddiv,
             secondmoddiv,secondassignment,secondload,secondnothing,
             secondadd,secondadd,secondadd,secondadd,
             secondadd,secondadd,secondnothing,secondadd,
             secondadd,secondshlshr,secondshlshr,secondnothing,
             secondadd,secondsubscriptn,secondderef,secondaddr,
             secondordconst,secondtypeconv,secondcalln,secondnothing,
             secondrealconst,secondumminus,secondnothing,secondvecn,
             secondstringconst,secondfuncret,secondselfn,
             secondnot,secondinline,secondniln,seconderror);

      var
         oldcodegenerror : boolean;

      begin
         oldcodegenerror:=codegenerror;
         codegenerror:=false;
         if not(p^.error) then
           begin
              procedures[p^.treetype](p);
              p^.error:=codegenerror;
              codegenerror:=codegenerror or oldcodegenerror;
           end
         else codegenerror:=true;
      end;

    { spart einige far-Aufrufe }

    function do_firstpass(var p : ptree) : boolean;

      begin
         codegenerror:=false;
         firstpass(p);
         do_firstpass:=codegenerror;
      end;

    function do_secondpass(var p : ptree) : boolean;

      begin
         codegenerror:=false;
         if not(p^.error) then
           secondpass(p);
         do_secondpass:=codegenerror;
      end;

    procedure generatecode(var p : ptree);

      begin
         cleartempgen;
         if not(do_firstpass(p)) then;
           do_secondpass(p);
         aktproccode.concatlist(@exprasmlist);
      end;

    procedure codegeninit;

      begin
         aktbreaklabel:=0;
         aktexitlabel:=0;
         aktcontinuelabel:=0;
         aktexitlabel:=0;
         aktexceptlabel:=0;
         aktentrycode.init;
         aktexitcode.init;
         aktexceptcode.init;
         aktproccode.init;
         exprasmlist.init;
         datasegment.init;
         constsegment.init;
         vmtasmlist.init;
         debuginfos.init;
      end;

    procedure codegendone;

      begin
         aktentrycode.done;
         aktexitcode.done;
         aktexceptcode.done;
         aktproccode.done;
         exprasmlist.done;
         datasegment.done;
         constsegment.done;
         vmtasmlist.done;
         debuginfos.done;
      end;

end.
