{****************************************************************************

                 Copyright (c) 1993,94 by Florian Klmpfl

 ****************************************************************************}

unit parser;

  interface

    uses
       scanner,globals,symtable,tree,cobjects,asmgen,codegen,
       asmbl,tempad,dos,types,strings,opt;

    var
       compiling_a_unit : boolean;

       { fr die Statistik }
       unitsize : longint;

    { parse ruft den Parser auf und gibt die }
    { Gre des globalen Datensegmentes      }
    { zurck                                 }

    function parse : longint;

    const
       heapsize : longint = 4000000;
       stacksize : longint = 8096;

  implementation

    {$I innr.inc}

    var
       token : ttoken;
       datasize : longint; { Gre des Datensegmentes, wird von proc_unit }
                           { oder proc_program gesetzt                    }

       refsymtable : psymtable; { Symboltabelle in welcher die          }
                                { Unitreferenzen abgelegt werden sollen }

       parse_only : boolean; { wird auf true gesetzt, wenn              }
                             { nur Funktionskpfe geparst werden sollen }
       unitlist1,unitlist2 : tstringcontainer;

    procedure befehlsblock;forward;
    procedure anweisung;forward;
    function typ(const name : stringid) : pdef;forward;
    function expr : ptree;forward;
    procedure block;forward;
    procedure proc_head;forward;
    procedure formal_parameter_list;forward;

    { versucht das Token i zu consumieren, pat }
    { das Token nicht, so wird ein Syntaxfehler }
    { ausgegeben                                }

    procedure consume(i : ttoken);

      { gibt einen Syntaxfehler aus }

      procedure syntaxerror(const s : string);

        begin
           exterror:=strpnew(s+' erwartet. ');
           error(syntax_error);
        end;

      const tokens : array[PLUS..CCHAR] of string[12] = (
                 '+','-','*','/','=','>','<','[',']',
                 '.',',','(',')',':',';','^',
                 '@',':=','<>','>=','<=','..',
                 'Bezeichner','real. Konst.','Dateiende',
                 'ord. Konst.','Stringkonst.','Charkonst.');

      var
         j : integer;

      begin
         if token<>i then
           begin
              if i<_ABSOLUTE then
                syntaxerror(tokens[i])
              else
                begin

                   { um die Programmgre klein zu halten, }
                   { wird fr ein Schlsselwort-Token der  }
                   { "Text" in der Schlsselworttabelle    }
                   { des Scanners nachgeschaut             }

                   for j:=1 to anz_keywords do
                     if keyword_token[j]=i then
                       syntaxerror(keyword[j])
                end;
           end
         else
           token:=yylex;
      end;

    { liest eine Liste durch Komma getrennter Bezeichner }
    { in eine Stringcontainer ein                        }

    function idlist : pstringcontainer;

      var
         sc : pstringcontainer;

      begin
         sc:=new(pstringcontainer,init);
         repeat
           sc^.insert(pattern);
           consume(ID);
           if token=COMMA then consume(COMMA)
             else break
         until false;
         idlist:=sc;
      end;

    procedure label_dec;

      begin
         repeat
           if (token<>ID) and (token<>INTCONST) then consume(ID);
           consume(token);
           if token<>SEMICOLON then consume(COMMA);
         until (token<>ID) and (token<>INTCONST);
         consume(SEMICOLON);
      end;

    { fgt die Symbole aus sc mit def in st ein  }
    { entfernt sc!                               }

    procedure insert_syms(st : psymtable;sc : pstringcontainer;def : pdef);

      var
         s : string;

      begin
         s:=sc^.get;
         while s<>'' do
           begin
              st^.insert(new(pvarsym,init(s,def)));
              s:=sc^.get;
           end;
         dispose(sc,done);
      end;

    { liest einen einzelnen Stringtyp mit optionaler }
    { Lngenangabe und gibt einen Pointer auf die    }
    { Definition zurck                              }

    function stringtyp : pdef;

      var
         p : ptree;
         d : pdef;

      begin
         consume(_STRING);
         if token=LECKKLAMMER then
           begin
              consume(LECKKLAMMER);
              p:=expr;
              do_firstpass(p);
              if not is_constintnode(p) then
                error(error_in_expression);
              if (p^.value<1) or (p^.value>255) then
                begin
                   error(string_too_long);
                   p^.value:=255;
                end;
              consume(RECKKLAMMER);
              d:=new(pstringdef,init(p^.value));
           end
         else d:=new(pstringdef,init(255));
         stringtyp:=d;
      end;


    { Zeiger auf das zuletzt gelesene Typsymbol }
    { (fr "forward"-Typen)                     }

    var
       lasttypesym : ptypesym;

    { liest einen Typbezeichner und gibt einen }
    { Pointer auf die Definition zurck        }
    { s ist der Name des Typs                  }

    function id_type(var s : string) : pdef;

      begin
         getsym(pattern,true);
         s:=pattern;
         consume(ID);
         if srsym^.typ=unitsym then
           begin
              consume(POINT);
              getsymonlyin(punitsym(srsym)^.unitsymtable,pattern);
              s:=pattern;
              consume(ID);
           end;
         if srsym^.typ<>typesym then
           begin
              fatalerror(type_id_expect);
              srsym:=generrorsym;
           end;
         lasttypesym:=ptypesym(srsym);
         id_type:=ptypesym(srsym)^.definition;
      end;

    { liest einen String oder Typbezeichner }
    { und gibt einen Pointer auf die        }
    { Definition zurck                     }

    function single_type(var s : string) : pdef;

      begin
         case token of
            _STRING : begin
                         single_type:=stringtyp;
                         s:='STRING';
                         lasttypesym:=nil;
                      end;
            else single_type:=id_type(s);
         end;
      end;

    { liest die Feldliste eines Records in }
    { symtablestack ein                    }
    { wenn record=false, dann knnen auch  }
    { Klassenfelder eingelesen werden, da  }
    { variante Recordkonstruktionen igno-  }
    { riert werden                         }
    { do_absolute, gibt an, ob ABSOLUTE,   }
    { sowie Dateitypen erlaubt sind        }

    procedure feldliste(is_record : boolean;do_absolute : boolean);

      var
         sc : pstringcontainer;
         s : stringid;
         p,casedef : pdef;
         hs : string;
         { maxsize enthlt maximale Gre eines varianten Astes }
         { startvarrec die Startaddresse des varianten Teiles }
         maxsize,startvarrec : longint;
         pt : ptree;

      begin
         while (token=ID) do
           begin
              sc:=idlist;
              consume(COLON);
              p:=typ('');
              if do_absolute and (token=_ABSOLUTE) then
                begin
                   consume(_ABSOLUTE);
                end
              else
                begin
                   insert_syms(symtablestack,sc,p);
                end;
              consume(SEMICOLON);
           end;
         if (token=_CASE) and is_record then
           begin
              maxsize:=0;
              consume(_CASE);
              s:=pattern;
              consume(ID);
              consume(COLON);
              casedef:=single_type(hs);
              if not is_ordinal(casedef) then
                error(ordinal_expect);
              symtablestack^.insert(new(pvarsym,init(s,casedef)));
              consume(_OF);
              startvarrec:=symtablestack^.datasize;
              repeat
                repeat
                  pt:=expr;
                  do_firstpass(pt);
                  if not(pt^.treetype=ordconstn) then
                    error(error_in_expression);
                  if token=COMMA then consume(COMMA)
                    else break;
                until false;
                consume(COLON);
                consume(LKLAMMER);
                while true do
                  begin
                     sc:=idlist;
                     consume(COLON);
                     p:=typ('');
                     insert_syms(symtablestack,sc,p);
                     if token=SEMICOLON then
                       consume(SEMICOLON)
                     else break;
                  end;
                consume(RKLAMMER);
                consume(SEMICOLON);
                { bisher grter varianter Teil? }
                maxsize:=max(maxsize,symtablestack^.datasize);
                symtablestack^.datasize:=startvarrec;
              until token=_END;
              symtablestack^.datasize:=maxsize;
           end;
      end;

    procedure _proc_head;forward;

    procedure constructor_head;

      begin
         consume(_CONSTRUCTOR);
         _proc_head;
         consume(SEMICOLON);
         { der Rckgabetyp von Konstruktoren ist Boolean }
         aktprocsym^.definition^.retdef:=
           new(pgrunddef,init(bool8bit,0,1));
      end;

    procedure destructor_head;

      begin
         consume(_DESTRUCTOR);
         _proc_head;
         consume(SEMICOLON);
         if assigned(aktprocsym^.definition^.para1) then
           error(no_paras_2_destructor);
         { kein Rckgabetyp }
         aktprocsym^.definition^.retdef:=voiddef;
      end;

    function object_dec(const n : stringid) : pdef;

      type
         tzugriffsmode = (priv,prot,pub);

      var
         aktzugriffsmode : tzugriffsmode;

      procedure object_komponenten;

        var
           oldparse_only : boolean;

        begin
           repeat
             case token of
                ID : feldliste(false,false);
                _PROCEDURE,_FUNCTION : begin
                               oldparse_only:=parse_only;
                               parse_only:=true;
                               proc_head;
                               parse_only:=oldparse_only;
                               if token=_VIRTUAL then
                                 begin
                                    if aktzugriffsmode=priv then
                                      error(priv_meth_not_virtual);
                                    consume(_VIRTUAL);
                                    consume(SEMICOLON);
                                    aktprocsym^.definition^.options:=
                                      aktprocsym^.definition^.options or povirtualmethod;
                                 end;
                            end;
                _CONSTRUCTOR : begin
                                  if aktzugriffsmode<>pub then
                                    error(const_cannot_priv);
                                  oldparse_only:=parse_only;
                                  parse_only:=true;
                                  constructor_head;
                                  parse_only:=oldparse_only;
                                  aktprocsym^.definition^.options:=
                                    aktprocsym^.definition^.options or poconstructor;
                               end;
                _DESTRUCTOR : begin
                               if aktzugriffsmode<>pub then
                                 error(dest_cannot_priv);
                               oldparse_only:=parse_only;
                               parse_only:=true;
                               destructor_head;
                               parse_only:=oldparse_only;
                               aktprocsym^.definition^.options:=
                                 aktprocsym^.definition^.options or podestructor;
                               if token=_VIRTUAL then
                                 begin
                                    consume(_VIRTUAL);
                                    consume(SEMICOLON);
                                    aktprocsym^.definition^.options:=
                                      aktprocsym^.definition^.options or povirtualmethod;
                                 end;
                            end;
                else exit;
             end;
           until false;
        end;

      var
         childof : pclassdef;
         class : pclassdef;

      begin
         aktzugriffsmode:=pub;
         if (symtablestack^.symtabletype and $3fff<>globalsymtable) and
           (symtablestack^.symtabletype and $3fff<>staticsymtable) then
           error(no_local_objects);
         if n='' then error(no_anonym_objects);
         consume(_OBJECT);
         childof:=nil;
         if token=LKLAMMER then
           begin
              consume(LKLAMMER);
              if token<>ID then
                consume(ID);
              getsym(pattern,true);
              if (srsym^.typ<>typesym) and
                 (ptypesym(srsym)^.definition^.deftype<>classdef) then
                 begin
                    error(class_type_expect);
                    childof:=nil;
                 end
              else childof:=pclassdef(ptypesym(srsym)^.definition);
              consume(ID);
              consume(RKLAMMER);
           end;
         class:=new(pclassdef,init(n,childof));

         { Default: public }
         aktzugriffsmode:=pub;
         class^.publicsyms^.next:=symtablestack;
         symtablestack:=class^.publicsyms;
         procinfo.class:=class;
         while token<>_END do
           begin
{              if token=_PRIVATE then
                begin
                   consume(_PRIVATE);
                   symtablestack:=symtablestack^.next;
                   class^.privatesyms^.next:=symtablestack;
                   symtablestack:=class^.privatesyms;
                   aktzugriffsmode:=priv;
                end;
              if token=_PROTECTED then
                begin
                   consume(_PROTECTED);
                   symtablestack:=symtablestack^.next;
                   class^.protectedsyms^.next:=symtablestack;
                   symtablestack:=class^.protectedsyms;
                   aktzugriffsmode:=prot;
                end; }
              if token=_PUBLIC then
                begin
                   consume(_PUBLIC);
                   symtablestack:=symtablestack^.next;
                   class^.publicsyms^.next:=symtablestack;
                   symtablestack:=class^.publicsyms;
                   aktzugriffsmode:=pub;
                end;
              object_komponenten;
           end;
         consume(_END);
         { VMT erzeugen: }
         vmtasmlist.concat(gennasmrec(DIRECT,S_NO,'.globl VMT_'+n));
         vmtasmlist.concat(gennasmrec(DIRECT,S_NO,'VMT_'+n+':'));
         vmtasmlist.concat(gennasmrec(A_LONG,S_NO,tostr(class^.size)));
         vmtasmlist.concat(gennasmrec(A_LONG,S_NO,tostr(-class^.size)));
         if assigned(class^.childof) then
           vmtasmlist.concat(gennasmrec(A_LONG,S_NO,'VMT_'+class^.childof^.name^))
         else
           vmtasmlist.concat(gennasmrec(A_LONG,S_NO,'0'));
         genvmt(class);
         symtablestack:=symtablestack^.next;
         procinfo.class:=nil;
         object_dec:=class;
      end;

    { liest eine Recorddefinition und gibt }
    { einen Pointer darauf zurck          }

    function record_dec : pdef;

      var
         symtable : psymtable;

      begin
         symtable:=new(psymtable,init(recordsymtable));
         symtable^.next:=symtablestack;
         symtablestack:=symtable;
         consume(_RECORD);
         feldliste(true,false);
         consume(_END);
         symtablestack:=symtable^.next;
         record_dec:=new(precdef,init(symtable));
      end;

    { liest eine Typdefinition und gibt eine }
    { Pointer darauf zurck                  }

    function typ(const name : stringid) : pdef;

      var
         hp1,p : pdef;
         pt : ptree;
         aufdef : paufzaehldef;
         ap : parraydef;
         s : stringid;
         l : longint;
         hs : string;

      begin
         case token of
            ID : p:=id_type(hs);
            LKLAMMER :
                  begin
                     consume(LKLAMMER);
                     l:=-1;
                     aufdef:=new(paufzaehldef,init);
                     repeat
                       s:=pattern;
                       consume(ID);
                       if token=ASSIGNMENT then
                         begin
                            consume(ASSIGNMENT);
                            pt:=expr;
                            do_firstpass(pt);
                            if not is_constintnode(pt) then
                              error(error_in_expression)
                            else l:=pt^.value;
                            if pt^.value<=l then
                              error(dup_enum);
                            disposetree(pt);
                         end
                       else
                         inc(l);
                       constsymtable^.insert(new(paufzaehlsym,init(s,aufdef,l)));
                       if token=COMMA then consume(COMMA)
                         else break;
                     until false;
                     aufdef^.max:=l;
                     p:=aufdef;
                     consume(RKLAMMER);
                  end;
            _STRING : p:=stringtyp;
            _ARRAY :
                  begin
                     consume(_ARRAY);
                     consume(LECKKLAMMER);
                     p:=nil;
                     repeat
                       { Ausdruck lesen und prfen }
                       pt:=expr;
                       do_firstpass(pt);

                       if (pt^.treetype<>rangen) or
                          (pt^.left^.treetype<>ordconstn) then
                         error(error_in_type);
                       { Registrierung der Grenzen erzwingen: }
                       if pt^.right^.resulttype=pdef(s32bitdef) then
                         pt^.right^.resulttype:=new(pgrunddef,init(
                           s32bit,$80000000,$7fffffff));
                       if p=nil then
                         begin
                            ap:=new(parraydef,init(pt^.left^.value,pt^.right^.value,pt^.right^.resulttype));
                            p:=ap;
                         end
                       else
                         begin
                            ap^.definition:=new(parraydef,init(pt^.left^.value,pt^.right^.value,pt^.right^.resulttype));
                            ap:=parraydef(ap^.definition);
                         end;
                       if token=COMMA then consume(COMMA)
                         else break;
                     until false;
                     consume(RECKKLAMMER);
                     consume(_OF);
                     ap^.definition:=typ('');
                  end;
            _SET : begin
                      consume(_SET);
                      consume(_OF);
                      consume(ID);
                   end;
{            _FILE : begin
                       consume(_FILE);
                       if token=_OF then
                         begin
                            consume(_OF);
                            p:=typ('');
                         end;
                    end;     }
            CARET : begin
                       consume(CARET);
                       forwardsallowed:=true;
                       hp1:=single_type(hs);
                       p:=new(ppointerdef,init(hp1));
                       if lasttypesym<>nil then
                         save_forward(ppointerdef(p),lasttypesym);
                       forwardsallowed:=false;
                    end;
            _RECORD : begin
                         p:=record_dec;
                      end;
            _OBJECT : begin
                         p:=object_dec(name);
                      end;
            _PROCEDURE : begin
                            proc_head;
                         end;
            _FUNCTION : begin
                           proc_head;
                        end;
            else
               begin
                  { kann dann nur Bereichstyp sein }
                  pt:=expr;
                  do_firstpass(pt);

                  { erlaubter Ausdruck ? }
                  if (pt^.treetype<>rangen) or
                     (pt^.left^.treetype<>ordconstn) then
                    error(error_in_type);
                  p:=new(pgrunddef,init(uauto,pt^.left^.value,pt^.right^.value));
                  disposetree(pt);
               end;
         end;
         typ:=p;
      end;

    { sucht in symtablestack nach zwar daklarierten }
    { aber nicht definierten Typen                  }

    procedure testforward_types(p : psym);far;

      begin
         if (p^.typ=typesym) and (p^.forwarddef) then
           error(type_id_not_defined);
      end;

    { liest den type-Abschnitt in symtablestack ein }

    procedure type_dec;

      var
         typename : stringid;

      begin
         consume(_TYPE);
         repeat
           typename:=pattern;
           consume(ID);
           consume(EQUAL);
           symtablestack^.insert(new(ptypesym,init(typename,typ(typename))));
           consume(SEMICOLON);
         until token<>ID;
         symtablestack^.foreach(testforward_types);
         resolve_forwards;
      end;

    { parst Variablendeklarationen und fgt sie in die }
    { oberste Symboltabelle ein                        }

    procedure var_dec;

      var
         p : pdef;
         sc : pstringcontainer;

      begin
         consume(_VAR);
         feldliste(false,true);
      end;

    procedure readtypedconst(def : pdef);

      var
         p : ptree;
         i,l : longint;
         s : string;
         ca : array[0..512] of char;

      begin
         case def^.deftype of
            grunddef : begin
                          p:=expr;
                          do_firstpass(p);
                          case pgrunddef(def)^.typ of
                             {!!!!!! Rangechecking }
                             s8bit,
                             u8bit : begin
                                        if not is_constintnode(p) then
                                          { Hier einfach Fehler ausgeben, der }
                                          { falsche Wert macht berhaubt nichts }
                                          error(error_in_expression);
                                        constsegment.concat(gennasmrec(A_BYTE,S_NO,tostr(p^.value)));
                                     end;
                             s32bit : begin
                                         if not is_constintnode(p) then
                                           error(error_in_expression);
                                         constsegment.concat(gennasmrec(A_LONG,S_NO,tostr(p^.value)));
                                     end;
                             s64real : begin
                                          if not is_constrealnode(p) then
                                            error(error_in_expression);
                                          constsegment.concat(gennasmrec(A_DOUBLE,S_NO,double2str(p^.value)));
                                       end;
                             bool8bit : begin
                                           if not is_constboolnode(p) then
                                             error(error_in_expression);
                                           constsegment.concat(gennasmrec(A_BYTE,S_NO,tostr(p^.value)));
                                        end;
                             uchar : begin
                                         if not is_constcharnode(p) then
                                           error(error_in_expression);
                                         constsegment.concat(gennasmrec(A_BYTE,S_NO,tostr(p^.value)));
                                     end;
                             u16bit,
                             s16bit : begin
                                         if not is_constintnode(p) then
                                           error(error_in_expression);
                                         constsegment.concat(gennasmrec(A_WORD,S_NO,tostr(p^.value)));
                                     end;
                          end;
                       end;
         pointerdef : begin
                         p:=expr;
                         do_firstpass(p);
                         if p^.treetype=niln then
                           constsegment.concat(gennasmrec(A_LONG,S_NO,'0'))
                         else
                           { kann sonst nur PCHAR sein }
                           if (ppointerdef(def)^.definition^.deftype=grunddef) and
                              (pgrunddef(ppointerdef(def)^.definition)^.typ=uchar) then
                             begin
                                l:=getlabel;
                                { String am Anfang einfgen }
                                if p^.treetype=stringconstn then
                                  constsegment.insert(gennasmrec(ASCII,S_NO,'"'+ibm2ascii(p^.values^)+'\0"'))
                                else if is_constcharnode(p) then
                                  constsegment.insert(gennasmrec(ASCII,S_NO,'"'+ibm2ascii(char(p^.value))+'\0"'))
                                else error(error_in_expression);
                                constsegment.insert(genlasmrec(A_LABEL,l));
                                { Label einfgen }
                                constsegment.concat(gennasmrec(A_LONG,S_NO,tolabel(l)));
                             end
                         else error(error_in_expression);
                     end;
         stringdef : begin
                        p:=expr;
                        do_firstpass(p);
                        if p^.treetype=stringconstn then
                          begin
                             s:=p^.values^;
                             if length(s)+1>def^.size then
                               s[0]:=chr(def^.size-1);
                             constsegment.concat(gennasmrec(ASCII,S_NO,'"'+ibm2ascii(
                               char(length(s))+s)+'"'));
                          end
                        else if is_constcharnode(p) then
                          begin
                             constsegment.concat(gennasmrec(ASCII,S_NO,'"\001'+ibm2ascii(
                              char(p^.value))+'"'));
                             s:=char(p^.value);
                          end
                        else error(error_in_expression);
                        ca[0]:='"';
                        fillchar(ca[1],def^.size-length(s)-1,' ');
                        ca[def^.size-length(s)]:='"';
                        ca[def^.size-length(s)+1]:=#0;
                        constsegment.concat(genpasmrec(ASCII,S_NO,ca));
                     end;
         arraydef : begin
                       consume(LKLAMMER);
                       for l:=parraydef(def)^.lowrange to parraydef(def)^.highrange-1 do
                         begin
                            readtypedconst(parraydef(def)^.definition);
                            consume(COMMA);
                         end;
                       readtypedconst(parraydef(def)^.definition);
                       consume(RKLAMMER);
                    end;
         else error(type_const_not_possible);
         end;
      end;

    procedure const_dec;

      var
         name : stringid;
         p : ptree;
         def : pdef;
         pd : pdouble;

      begin
         consume(_CONST);
         repeat
           name:=pattern;
           consume(ID);
           case token of
              EQUAL : begin
                          consume(EQUAL);
                          p:=expr;
                          do_firstpass(p);
                          case p^.treetype of
                             ordconstn :  begin
                                             if is_constintnode(p) then
                                               symtablestack^.insert(new(pconstsym,init(name,constint,p^.value)))
                                             else if is_constcharnode(p) then
                                               symtablestack^.insert(new(pconstsym,init(name,conststring,p^.value)))
                                             else if is_constboolnode(p) then
                                               symtablestack^.insert(new(pconstsym,init(name,constbool,p^.value)));
                                          end;
                             stringconstn :
                               symtablestack^.insert(new(pconstsym,init(name,conststring,longint(p^.values))));
                             realconstn : begin
                                             new(pd);
                                             pd^:=p^.valued;
                                             symtablestack^.insert(new(pconstsym,init(name,constreal,longint(pd))));
                                          end;
                             else error(error_in_expression);
                          end;
                          consume(SEMICOLON);
                       end;
              COLON : begin
                          consume(COLON);
                          def:=typ('');
                          symtablestack^.insert(new(ptypedconstsym,init(name,def)));
                          consume(EQUAL);
                          readtypedconst(def);
                          consume(SEMICOLON);
                       end;
              else consume(EQUAL);
           end;
         until token<>ID;
      end;

    procedure if_anweisung;

      var
         p : ptree;
         l1,l2,hl : longint;

      begin
         consume(_IF);
         l1:=getlabel;
         l2:=getlabel;
         truelabel:=l1;
         falselabel:=l2;
         p:=expr;
         cleartempgen;
         do_firstpass(p);
         cleartempgen;
         do_secondpass(p);
         maketojumpbool(p);
         aktproccode.concatlist(@exprasmlist);
         consume(_THEN);
         aktproccode.concat(genlasmrec(A_LABEL,l1));
         if token<>_ELSE then
           anweisung;
         if token=_ELSE then
           begin
              consume(_ELSE);
              hl:=getlabel;
              aktproccode.concat(genlasmrec(JMP,hl));
              aktproccode.concat(genlasmrec(A_LABEL,l2));
              anweisung;
              aktproccode.concat(genlasmrec(A_LABEL,hl));
           end
         else
           begin
              aktproccode.concat(genlasmrec(A_LABEL,l2));
           end;
      end;

    procedure case_anweisung;

      type
         pcaserecord = ^tcaserecord;

         tcaserecord = record
            low,high : longint;
            at : longint;
            anweisung : longint;
            less,greater : pcaserecord;
         end;

      var
         oldproccode,casecode,jmpcode : tasmlist;
         aktcaselabel,endlabel,elselabel : longint;
         hregister : tregister;
         opsize : topsize;
         wurzel : pcaserecord;
         ranges : boolean;

      procedure gentreejmp(p : pcaserecord);

        var
           lesslabel,greaterlabel : longint;

        begin
           jmpcode.concat(genlasmrec(A_LABEL,p^.at));

           { Labels fr links und rechts berechnen }
           if (p^.less=nil) then
             lesslabel:=elselabel
           else
             lesslabel:=p^.less^.at;
           if (p^.greater=nil) then
             greaterlabel:=elselabel
           else
             greaterlabel:=p^.greater^.at;

           { keine Bereichslabel :}
           if p^.low=p^.high then
             begin
                jmpcode.concat(gennasmrec(CMP,opsize,'$'+tostr(p^.low)+','+regid2str[hregister]));
                if greaterlabel=lesslabel then
                  begin
                     jmpcode.concat(genlasmrec(JNE,lesslabel));
                  end
                else
                  begin
                     jmpcode.concat(genlasmrec(JL,lesslabel));
                     jmpcode.concat(genlasmrec(JG,greaterlabel));
                  end;
                jmpcode.concat(genlasmrec(JMP,p^.anweisung));
             end
           else
           { Bereichslabel: }
             begin
                jmpcode.concat(gennasmrec(CMP,opsize,'$'+tostr(p^.low)+','+regid2str[hregister]));
                jmpcode.concat(genlasmrec(JL,lesslabel));
                jmpcode.concat(gennasmrec(CMP,opsize,'$'+tostr(p^.high)+','+regid2str[hregister]));
                jmpcode.concat(genlasmrec(JG,greaterlabel));
                jmpcode.concat(genlasmrec(JMP,p^.anweisung));
             end;
           if assigned(p^.less) then
             gentreejmp(p^.less);
           if assigned(p^.greater) then
             gentreejmp(p^.greater);
        end;

      procedure genlinearlist(p : pcaserecord);

        var
           first : boolean;
           last : longint;
           helplabel : longint;

        procedure genitem(t : pcaserecord);

          begin
             if assigned(t^.less) then
               genitem(t^.less);
             if t^.low-last=1 then
               jmpcode.concat(gennasmrec(A_DEC,opsize,regid2str[hregister]))
             else if t^.low-last=0 then
               jmpcode.concat(gennasmrec(A_OR,opsize,regid2str[hregister]+','+regid2str[hregister]))
             else
               jmpcode.concat(gennasmrec(SUB,opsize,'$'+tostr(t^.low-last)+','+regid2str[hregister]));
             last:=t^.low;

             jmpcode.concat(genlasmrec(JZ,t^.anweisung)); {!!!!}
             {
             if t^.low=t^.high then
               jmpcode.concat(genlasmrec(JZ,t^.anweisung))
             else
               begin
                  helplabel:=getlabel;
                  jmpcode.concat(genlasmrec(JG,helplabel));
                  if t^.high-last=1 then
                    jmpcode.concat(gennasmrec(A_DEC,opsize,regid2str[hregister]))
                  else
                    jmpcode.concat(gennasmrec(SUB,opsize,'$'+tostr(t^.high-last)+','+regid2str[hregister]));
                  jmpcode.concat(genlasmrec(JLE,t^.anweisung));
                  jmpcode.concat(genlasmrec(A_LABEL,helplabel));
                  last:=t^.high;
               end; }
             if assigned(t^.greater) then
               genitem(t^.greater);
          end;

        begin
           last:=0;
           genitem(wurzel);
           jmpcode.concat(genlasmrec(JMP,elselabel));
        end;

      procedure newcaselabel(l,h : longint);

        var
           hcaselabel : pcaserecord;

        procedure insertlabel(var p : pcaserecord);

          begin
             if p=nil then p:=hcaselabel
             else
                if (p^.low>hcaselabel^.low) and
                   (p^.low>hcaselabel^.high) then
                  insertlabel(p^.less)
                else if (p^.high<hcaselabel^.low) and
                   (p^.high<hcaselabel^.high) then
                  insertlabel(p^.greater)
                else error(double_caselabel);
          end;

        begin
           new(hcaselabel);
           hcaselabel^.less:=nil;
           hcaselabel^.greater:=nil;
           hcaselabel^.anweisung:=aktcaselabel;
           hcaselabel^.at:=getlabel;
           hcaselabel^.low:=l;
           hcaselabel^.high:=h;
           insertlabel(wurzel);
        end;

      procedure deletecaselabel(p : pcaserecord);

        begin
           if assigned(p^.greater) then
             deletecaselabel(p^.greater);
           if assigned(p^.less) then
             deletecaselabel(p^.less);
           dispose(p);
        end;

      var
         p : ptree;
         hl,hl2 : longint;
         casedef : pdef;

      begin
         endlabel:=getlabel;
         elselabel:=getlabel;
         wurzel:=nil;
         consume(_CASE);
         p:=expr;
         do_firstpass(p);
         if not(is_ordinal(p^.resulttype)) then
           error(ordinal_expect);
         casedef:=p^.resulttype;
         do_secondpass(p);
         case p^.resulttype^.size of
            1 : begin
                   opsize:=S_B;
                   hregister:=R_AL;
                end;
            2 : begin
                   opsize:=S_W;
                   hregister:=R_AX;
                end;
            4 : begin
                   opsize:=S_L;
                   hregister:=R_EAX;
                end;
            else internalerror(2);
         end;
         aktproccode.concatlist(@exprasmlist);
         case p^.location.loc of
           LOC_REGISTER :
             hregister:=p^.location.register;
           LOC_MEM,LOC_REFERENZ :
                aktproccode.concat(gennasmrec(MOV,opsize,getreferenzstring(
                  p^.location.referenz)+','+regid2str[hregister]));
         end;
         consume(_OF);
         disposetree(p);
         oldproccode.init;
         casecode.init;
         jmpcode.init;
         oldproccode.concatlist(@aktproccode);
         ranges:=false;
         repeat
           aktcaselabel:=getlabel;
           repeat
             p:=expr;
             do_firstpass(p);
             if (p^.treetype=rangen) then
               begin
                  hl:=get_ordinal_value(p^.left);
                  hl2:=get_ordinal_value(p^.right);
                  testrange(casedef,hl);
                  testrange(casedef,hl2);
                  newcaselabel(hl,hl2);
                  ranges:=true;
               end
             else
               begin
                  hl:=get_ordinal_value(p);
                  testrange(casedef,hl);
                  newcaselabel(hl,hl);
               end;
             disposetree(p);
             if token=COMMA then consume(COMMA)
               else break;
           until false;
           consume(COLON);

           { Anweisung parsen, und Spruenge und Labels anhaengen }
           casecode.concat(genlasmrec(A_LABEL,aktcaselabel));
           anweisung;
           casecode.concatlist(@aktproccode);
           casecode.concat(genlasmrec(JMP,endlabel));
           if not((token=_ELSE) or (token=_OTHERWISE) or (token=_END)) then
             consume(SEMICOLON);
         until (token=_ELSE) or (token=_OTHERWISE) or (token=_END);

         { Sprnge erzeugen }
         if ranges then
           gentreejmp(wurzel)
         else
           genlinearlist(wurzel);

         deletecaselabel(wurzel);

         { Procedurecode wieder zusammembauen }
         aktproccode.concatlist(@oldproccode);
         aktproccode.concatlist(@jmpcode);
         aktproccode.concatlist(@casecode);

         { ELSE-Zweig bearbeiten }
         aktproccode.concat(genlasmrec(A_LABEL,elselabel));
         if (token=_ELSE) or (token=_OTHERWISE) then
           begin
              if token=_ELSE then consume(_ELSE)
                else consume(_OTHERWISE);
              anweisung;
              if token=SEMICOLON then consume(SEMICOLON);
           end;
         consume(_END);
         aktproccode.concat(genlasmrec(A_LABEL,endlabel));
      end;

    procedure repeat_anweisung;

      var
         p : ptree;
         l1,l2,l3,oldclabel,oldblabel : longint;

      begin
         l1:=getlabel;
         l2:=getlabel;

         { Continue und Breaklabels einrichten: }
         oldclabel:=aktcontinuelabel;
         oldblabel:=aktbreaklabel;
         aktcontinuelabel:=l1;
         aktbreaklabel:=l2;
         aktproccode.concat(genlasmrec(A_LABEL,l1));
         consume(_REPEAT);
         while token<>_UNTIL do
           begin
              anweisung;
              if token<>SEMICOLON then
                break
              else
                consume(SEMICOLON);
              while token=SEMICOLON do
                consume(SEMICOLON);
           end;
         consume(_UNTIL);
         p:=expr;
         truelabel:=l2;
         falselabel:=l1;
         cleartempgen;
         do_firstpass(p);
         cleartempgen;
         do_secondpass(p);
         maketojumpbool(p);
         aktproccode.concatlist(@exprasmlist);

         aktproccode.concat(genlasmrec(A_LABEL,l2));
         aktcontinuelabel:=oldclabel;
         aktbreaklabel:=oldblabel;
      end;

    procedure while_anweisung;

      var
         p : ptree;
         l1,l2,l3,oldclabel,oldblabel : longint;
         hl : tasmlist;

      begin
         consume(_WHILE);
         l1:=getlabel;
         l2:=getlabel;
         l3:=getlabel;

         { Continue und Breaklabels einrichten: }
         oldclabel:=aktcontinuelabel;
         oldblabel:=aktbreaklabel;
         aktcontinuelabel:=l1;
         aktbreaklabel:=l3;

         p:=expr;
         cleartempgen;
         falselabel:=l3;
         truelabel:=l1;
         do_firstpass(p);
         cleartempgen;
         do_secondpass(p);
         maketojumpbool(p);
         hl.init;
         hl.concatlist(@exprasmlist);
         consume(_DO);
         aktproccode.concat(genlasmrec(JMP,l2));
         aktproccode.concat(genlasmrec(A_LABEL,l1));
         if token<>SEMICOLON then
           anweisung;
         aktproccode.concat(genlasmrec(A_LABEL,l2));
         { Bedingungscode am Ende ist mindestens gleich effizient, }
         { oft aber effizienter }
         aktproccode.concatlist(@hl);
         aktproccode.concat(genlasmrec(A_LABEL,l3));
         aktcontinuelabel:=oldclabel;
         aktbreaklabel:=oldblabel;
      end;

    procedure for_anweisung;

      var
         p1,laufvar,tovalue : ptree;
         temp1 : treferenz;
         backward : boolean;

         { bei konstanten Grenzen kann der erste Vergleich oft entfallen }
         omitfirstcmp : boolean;

         l1,l2,oldclabel,oldblabel : longint;
         hs : byte;
         temptovalue : boolean;
         opsize : topsize;
         hop : tasmop;
         cmpreg : tregister;

      begin
         { Anweisung parsen }
         consume(_FOR);
         p1:=expr;
         if token=_DOWNTO then
           begin
              consume(_DOWNTO);
              backward:=true;
           end
         else
           begin
              consume(_TO);
              backward:=false;
           end;
         tovalue:=expr;
         consume(_DO);

         { Labels laden }
         oldclabel:=aktcontinuelabel;
         oldblabel:=aktbreaklabel;
         l1:=getlabel;
         l2:=getlabel;
         aktbreaklabel:=l2;
         aktcontinuelabel:=l1;

         { Startzuweisung erzeugen }
         cleartempgen;
         if p1^.treetype<>assignn then
           fatalerror(error_in_expression);

           { Laufvariable retten }
           laufvar:=getcopy(p1^.left);

           { Grenzen berechnen }
           cleartempgen;
           do_firstpass(tovalue);
           cleartempgen;
           do_firstpass(laufvar);

           { ... knnen wir den ersten Vergleich sparen? }
           if tovalue^.treetype=ordconstn then
             if laufvar^.right^.treetype=ordconstn then
               if (backward and (laufvar^.right^.value>=tovalue^.value)) or
                  (not(backward) and (laufvar^.right^.value<=tovalue^.value)) then
                 omitfirstcmp:=true;

           { laufvar auf Gltigkeit prfen: }
           if (laufvar^.treetype<>loadn) then
             fatalerror(invalid_for_var);
           if (not(is_ordinal(laufvar^.resulttype))) then
             fatalerror(ordinal_expect);
{           if pgrunddef(laufvar^.resulttype)^.typ<>s32bit then
             fatalerror(for_var_must_be_32bit); }

           cleartempgen;
           { berechnet nur Referenz }
           do_secondpass(laufvar);
           if not(simple_loadn) then
             error(invalid_for_var);
           aktproccode.concatlist(@exprasmlist);

         generatecode(p1);

         hs:=laufvar^.resulttype^.size;
         case hs of
            1 : begin
                   opsize:=S_B;
                   cmpreg:=R_AL;
                end;
            2 : begin
                   opsize:=S_W;
                   cmpreg:=R_AX;
                end;
            4 : begin
                   opsize:=S_L;
                   cmpreg:=R_EAX;
                end;
         end;

         { Zielwert berechnen und prfen ob vielleicht vernderlich }
         { falls dem so ist, in temporre Variable laden            }
         if tovalue^.treetype<>ordconstn then
           begin
              tovalue:=gentypeconvnode(tovalue,laufvar^.resulttype);
              generatecode(tovalue);
              gettempofsizereferenz(hs,temp1);
              temptovalue:=true;
              if tovalue^.location.loc=LOC_REGISTER then
                begin
                   aktproccode.concat(
                     gennasmrec(MOV,opsize,regid2str[tovalue^.location.register]
                     +','+getreferenzstring(temp1)));
                 end
              else
                 begin
                    concatcopy(tovalue^.location.referenz,temp1,hs,false);
                    aktproccode.concatlist(@exprasmlist);
                 end;
           end
         else temptovalue:=false;
          if temptovalue then
            begin
               aktproccode.concat(
                 gennasmrec(MOV,opsize,
                 getreferenzstring(laufvar^.location.referenz)+','+
                 regid2str[cmpreg]));
               aktproccode.concat(
                 gennasmrec(CMP,opsize,getreferenzstring(temp1)+','+
                  regid2str[cmpreg]));
            end
          else
            begin
               if not(omitfirstcmp) then
               aktproccode.concat(
                 gennasmrec(CMP,opsize,'$'+tostr(tovalue^.value)+','+
                 getreferenzstring(laufvar^.location.referenz)));
            end;
         if backward then
           hop:=JL
         else hop:=JG;

         if not(omitfirstcmp) or temptovalue then
          aktproccode.concat(
            genlasmrec(hop,aktbreaklabel));

         { "Continuelabel" }
         aktproccode.concat(
           genlasmrec(A_LABEL,aktcontinuelabel));

         { ...nun die Anweisungen: }
         if token<>SEMICOLON then
           anweisung;

         { je nach Zhlrichtung DEC oder INC...}
         if backward then
           hop:=A_DEC
         else hop:=A_INC;

         aktproccode.concat(gennasmrec(hop,opsize,getreferenzstring(laufvar^.location.referenz)));

         { Vergleich erzeugen und den entsprechenden }
         { Sprung                                    }
         if temptovalue then
           begin
               aktproccode.concat(
                 gennasmrec(MOV,opsize,
                 getreferenzstring(laufvar^.location.referenz)+','+
                 regid2str[cmpreg]));
               aktproccode.concat(
                 gennasmrec(CMP,opsize,getreferenzstring(temp1)+','+
                  regid2str[cmpreg]));
           end
         else
           aktproccode.concat(
             gennasmrec(CMP,opsize,'$'+tostr(tovalue^.value)+','+
             getreferenzstring(laufvar^.location.referenz)));
         if backward then
           hop:=JGE
         else hop:=JLE;
         aktproccode.concat(
           genlasmrec(hop,aktcontinuelabel));
         { break-Label: }
         aktproccode.concat(
           genlasmrec(A_LABEL,aktbreaklabel));

         { und aufrumen: }
         aktcontinuelabel:=oldclabel;
         aktbreaklabel:=oldblabel;
      end;

    procedure with_anweisung;

      begin
         consume(_WITH);
         repeat
           expr;
           if token=COMMA then consume(COMMA)
             else break;
         until false;
         consume(_DO);
         if token<>SEMICOLON then
           anweisung;
      end;

    procedure throw_anweisung;

      begin
         if not(procinfo.exceptions) then
           fatalerror(exceptions_not_allowed);
         consume(_THROW);
         consume(ID);
      end;

    procedure try_anweisung;

      begin
         if not(procinfo.exceptions) then
           fatalerror(exceptions_not_allowed);
         consume(_TRY);
         anweisung;
         repeat
           consume(_CATCH);
           consume(ID);
           consume(COLON);
           anweisung;
         until token<>_CATCH;
      end;

    procedure exit_anweisung;

      begin
         consume(_EXIT);
         if token=LKLAMMER then
           begin
              {!!!!!!!! exit(7+8) bearbeiten }
           end;
         aktproccode.concat(genlasmrec(JMP,aktexitlabel));
      end;

    procedure _asm_anweisung;

      begin
         assemble;

         { Erst am Ende _ASM konsumieren, da der Scanner sonst die }
         { erste Assembleranweisung zu lesen versucht! }
         consume(_ASM);

         { (END ist gelesen) }
         if token=LECKKLAMMER then
           begin
              { explizite Angabe der modifizierten Register moeglich }
              consume(LECKKLAMMER);
              if token<>RECKKLAMMER then
                repeat
                  if upper(pattern)='EAX' then
                    usedinproc:=usedinproc or ($80 shr byte(R_EAX))
                  else if upper(pattern)='EBX' then
                    usedinproc:=usedinproc or ($80 shr byte(R_EBX))
                  else if upper(pattern)='ECX' then
                    usedinproc:=usedinproc or ($80 shr byte(R_ECX))
                  else if upper(pattern)='EDX' then
                    usedinproc:=usedinproc or ($80 shr byte(R_EDX))
                  else if upper(pattern)='ESI' then
                    usedinproc:=usedinproc or ($80 shr byte(R_ESI))
                  else if upper(pattern)='EDI' then
                    usedinproc:=usedinproc or ($80 shr byte(R_EDI))
                  else consume(RECKKLAMMER);
                  consume(CSTRING);
                  if token=COMMA then consume(COMMA)
                    else break;
                until false;
              consume(RECKKLAMMER);
           end
         else usedinproc:=$ff;
      end;

    function anweisung_sizeof(var p : pdef) : ptree;far;

      begin
{         consume(LKLAMMER);
         consume(RKLAMMER);
         p:=s32bitdef;
         anweisung_sizeof:=geninlinenode(in_sizeof_x,expr); }
      end;

    function parse_paras : ptree;

      var
         p1,p2 : ptree;

      begin
         p2:=nil;
         while true do
           begin
              p1:=expr;
              p2:=gencallparanode(p1,p2);
              if token=COMMA then
                consume(COMMA)
              else
                break;
           end;
         parse_paras:=p2;
      end;

    function anweisung_syssym(l : longint;var pd : pdef) : ptree;

      var
         p1,p2 : ptree;
         paras : ptree;


      begin
         case l of
            in_assigned_x : begin
                               consume(LKLAMMER);
                               p1:=expr;
                               do_firstpass(p1);
                               if p1^.resulttype^.deftype<>pointerdef then
                                 error(no_para_match);
                               p2:=gencallparanode(p1,nil);
                               p2:=geninlinenode(l,p2);
                               consume(RKLAMMER);
                               pd:=booldef;
                               anweisung_syssym:=p2;
                            end;
            in_concat_x : begin
                             consume(LKLAMMER);
                             p2:=nil;
                             while true do
                               begin
                                  p1:=expr;
                                  do_firstpass(p1);
                                  if not((p1^.resulttype^.deftype=stringdef) or
                                         ((p1^.resulttype^.deftype=grunddef) and
                                          (pgrunddef(p1^.resulttype)^.typ=uchar)
                                         )
                                    ) then error(no_para_match);
                                  if p2<>nil then
                                    p2:=gennode(addn,p2,p1)
                                  else p2:=p1;
                                  if token=COMMA then
                                    consume(COMMA)
                                  else break;
                               end;
                             consume(RKLAMMER);
                             pd:=cstringdef;
                             anweisung_syssym:=p2;
                          end;
            in_write_x,
            in_read_x,
            in_writeln_x,
            in_readln_x : begin
                             if token=LKLAMMER then
                               begin
                                  consume(LKLAMMER);
                                  paras:=parse_paras;
                                  consume(RKLAMMER);
                               end
                             else
                               paras:=nil;
                             pd:=voiddef;
                             anweisung_syssym:=geninlinenode(l,paras);
                          end;
            else internalerror(101);
         end;
      end;

    procedure new_dispose_anweisung;

      var
         p : ptree;
         ht : ttoken;
         asmrec : pasmrec;

      begin
         ht:=token;
         if token=_NEW then consume(_NEW)
           else consume(_DISPOSE);
         consume(LKLAMMER);
         p:=expr;
         do_firstpass(p);
         if p^.resulttype^.deftype<>pointerdef then
           begin
              error(pointer_expect);
              consume(RKLAMMER);
              exit;
           end;
         if p^.location.loc<>LOC_REFERENZ then
           begin
              error(error_in_expression);
              consume(RKLAMMER);
              exit;
           end;
         if ppointerdef(p^.resulttype)^.definition^.deftype=classdef then
           warning(take_extended_syntax);
         do_secondpass(p);
         aktproccode.concatlist(@exprasmlist);
         aktproccode.concat(gennasmrec(LEA,S_L,
           getreferenzstring(p^.location.referenz)+',%edi'));
         aktproccode.concat(gennasmrec(PUSH,S_L,'%edi'));
         aktproccode.concat(gennasmrec(PUSH,S_L,'$'+tostr(ppointerdef(p^.resulttype)^.definition^.size)));

         { Exceptionlabel pushen }
         if not(procinfo.exceptions) then
           error(proc_must_handleexceptions);
         asmrec:=gennasmrec(PUSH,S_L,'$'+tolabel(aktexceptlabel));
{         if jumpoptimize then
           registerused(asmrec,aktexceptlabel); }
         aktproccode.concat(asmrec);
         if ht=_NEW then
           aktproccode.concat(gennasmrec(CALL,S_NO,'GETMEM'))
           else aktproccode.concat(gennasmrec(CALL,S_NO,'FREEMEM'));
         maybe_loadesi;
         consume(RKLAMMER);
      end;

    procedure anweisung;

      var
         p : ptree;

      begin
         aktlinenr:=inputstack^.line_no;
         case token of
            _GOTO : begin
                       consume(_GOTO);
                    end;
            _BEGIN : befehlsblock;
            _IF    : if_anweisung;
            _CASE  : case_anweisung;
            _REPEAT : repeat_anweisung;
            _WHILE : while_anweisung;
            _FOR : for_anweisung;
            _NEW,_DISPOSE : new_dispose_anweisung;
            _WITH : with_anweisung;
            _TRY : try_anweisung;
            _THROW : throw_anweisung;
            SEMICOLON : ;
            _CONTINUE : begin
                           consume(_CONTINUE);
                           if aktcontinuelabel<>0 then
                             aktproccode.concat(genlasmrec(JMP,aktcontinuelabel))
                           else error(continue_not_allowed);
                        end;
            _FAIL : begin
                       if (aktprocsym^.definition^.options and poconstructor)=0 then
                         error(fail_only_in_constructor);
                       consume(_FAIL);
                       { !!!! }
                    end;
            _BREAK : begin
                           consume(_BREAK);
                           if aktbreaklabel<>0 then
                             aktproccode.concat(genlasmrec(JMP,aktbreaklabel))
                           else error(break_not_allowed);
                        end;
            _EXIT : exit_anweisung;
            _ASM : _asm_anweisung;
            else
               begin
                  p:=expr;
                  if (exprlevel<9) and (p^.treetype<>calln)
                     and (p^.treetype<>assignn) and (p^.treetype<>inlinen) then
                    error(error_in_expression);
                  cleartempgen;
                  do_firstpass(p);
                  if (exprlevel<1) and (p^.resulttype=pdef(voiddef)) then
                    fatalerror(error_in_expression);
                  do_secondpass(p);
                  aktproccode.concatlist(@exprasmlist);
                  disposetree(p);
               end;
         end;
         aktlinenr:=0;
      end;

    procedure befehlsblock;

      begin
         consume(_BEGIN);

         {!!!!!!! Labels}
         while token<>_END do
           begin
              anweisung;
              if token<>SEMICOLON then
                break
              else
                consume(SEMICOLON);
              while token=SEMICOLON do
                consume(SEMICOLON);
           end;
         consume(_END);
      end;

    procedure formal_parameter_list;

      var
         sc : pstringcontainer;
         s : string;
         p : pdef;
         ref : boolean;
         vs : pvarsym;
         hs1,hs2 : string;
         varspez : tvarspez;

      begin
         consume(LKLAMMER);
         repeat
           if token=_VAR then
             begin
                consume(_VAR);
                varspez:=vs_var;
             end
           else if token=_CONST then
             begin
                consume(_CONST);
                varspez:=vs_const;
             end
           else varspez:=vs_value;
           sc:=idlist;
           if token=COLON then
             begin
                consume(COLON);
                p:=single_type(hs1);
             end
           else
             begin
                hs1:='$$$';
                p:=new(pformaldef,init);
             end;
           s:=sc^.get;
           hs2:=aktprocsym^.definition^.mangledname;
           while s<>'' do
             begin
                aktprocsym^.definition^.concatdef(p,varspez);
                hs2:=hs2+'$'+hs1;
                vs:=new(pvarsym,init(s,p));
                vs^.varspez:=varspez;
                aktprocsym^.definition^.parast^.insert(vs);
                s:=sc^.get;
             end;
           dispose(sc,done);
           aktprocsym^.definition^.setmangledname(hs2);
           if token=SEMICOLON then consume(SEMICOLON)
             else break;
         until false;
         consume(RKLAMMER);
      end;

    { enthlt den richtigen Namen der Procedure, der nicht }
    { ge"upcased" wurde                                     }
    var
       realname : stringid;

    procedure _proc_head;

      var
         sp : stringid;
         pd : pprocdef;
         paramoffset : longint;
         hsymtab : psymtable;
         sym : psym;
         hs : string;

      begin
         sp:=pattern;
         realname:=orgpattern;
         consume(ID);
         if (token=POINT) and not(parse_only) then
           begin
              consume(POINT);
              getsym(sp,true);
              sym:=srsym;
              if (sym^.typ<>typesym) or
                 (ptypesym(sym)^.definition^.deftype<>classdef) then
                fatalerror(object_type_expect);
              sp:=pattern;
              realname:=orgpattern;
              consume(ID);
              procinfo.class:=pclassdef(ptypesym(sym)^.definition);
              aktprocsym:=pprocsym(procinfo.class^.publicsyms^.search(sp));

              { wird unten provisorisch behoben }
              if aktprocsym=nil then
                error(method_id_expect);
           end
         else
           begin
              aktprocsym:=pprocsym(symtablestack^.search(sp));
              hs:=procprefix+'_'+sp;
              if not(parse_only) then
                begin
                   { hier haben wir nicht nur einen Header }
                   procinfo.class:=nil;
                   hsymtab:=symtablestack;
                   if (aktprocsym=nil) then
                     begin
                        while (assigned(hsymtab)) and (hsymtab^.symtabletype<>globalsymtable) do
                          hsymtab:=hsymtab^.next;
                        if assigned(hsymtab) and (hsymtab^.symtabletype=globalsymtable) then
                        aktprocsym:=pprocsym(hsymtab^.search(sp));
                     end;
                end;
           end;
         if procinfo.class<>nil then
           hs:=procprefix+'_$$_'+procinfo.class^.name^+'_'+sp;
         if aktprocsym=nil then
           begin
              aktprocsym:=new(pprocsym,init(sp));
              symtablestack^.insert(aktprocsym);
           end;
         if aktprocsym^.typ<>procsym then fatalerror(overloaded_no_proc);
         pd:=new(pprocdef,init);

         { Offset der Parameter berechnen }
         paramoffset:=8;

         { Sollten Exceptions eingeschaltet sein, dann Exceptionflag setzten }
         if genexceptcode then
           begin
              pd^.options:=pd^.options or poexceptions;
              inc(paramoffset,4);
           end;

         procinfo.exceptions:=genexceptcode;

         { Framepointeroffset berechnen }

         if lexlevel>0 then
           begin
              procinfo.framepointer:=paramoffset;
              inc(paramoffset,4);
           end;

         { Selfpointeroffset }

         if assigned(procinfo.class) then
           begin
              procinfo.ESI_offset:=paramoffset;
              inc(paramoffset,4);
           end;

         procinfo.call_offset:=paramoffset;

         pd^.parast^.datasize:=0;

         pd^.nextoverloaded:=aktprocsym^.definition;
         aktprocsym^.definition:=pd;
         aktprocsym^.definition^.setmangledname(hs);
         if not(parse_only) then
           procprefix:=hs;
         if token=LKLAMMER then formal_parameter_list;
      end;

    procedure proc_head;

      var
         { Nur ein Hilfsstring, der den Namen des Rckgabetypes einer }
         { Funktion aufnimmt                                          }
         hs : string;

      begin
         if token=_FUNCTION then
           begin
              consume(_FUNCTION);
              _proc_head;
              consume(COLON);
              aktprocsym^.definition^.retdef:=single_type(hs);
           end
         else if token=_PROCEDURE then
           begin
              consume(_PROCEDURE);
              _proc_head;
              aktprocsym^.definition^.retdef:=voiddef;
           end
         else if token=_CONSTRUCTOR then
           begin
              consume(_CONSTRUCTOR);
              _proc_head;
              { kann auch als bool'sche Funktion betrachtet werden }
              aktprocsym^.definition^.retdef:=
                new(pgrunddef,init(bool8bit,0,1));

              { auerdem Contruktorflag setzten }
              aktprocsym^.definition^.options:=
                aktprocsym^.definition^.options or poconstructor;
           end
         else if token=_DESTRUCTOR then
           begin
              consume(_DESTRUCTOR);
              _proc_head;
              aktprocsym^.definition^.retdef:=voiddef;

              { auerdem Destruktorflag setzten }
              aktprocsym^.definition^.options:=
                aktprocsym^.definition^.options or podestructor;
           end;
         consume(SEMICOLON);
      end;

    procedure unter_dec;

      var
         oldprocsym : pprocsym;
         oldexceptlabel,oldexitlabel : longint;
         class : pclassdef;
         oldprocinfo : tprocinfo;

         oldconstsymtable : psymtable;

         { fr geschachtelte Unterprogramme eindeutige Namen erzeugen }
         oldprefix,hs : string;

         { Gre des lokalen Stackframes }
         stackframe : longint;

         { Anzahl der Bytes die mit RET entfernt werden mssen }
         parasize : longint;

         { true wenn kein Stackframe erforderlich ist }
         nostackframe : boolean;

         hd,pd : pprocdef;
         names : tstringcontainer;

         { wird auf true gesetzt, wenn Symbole exportiert werden sollen }
         make_global : boolean;

         { wird auf true gesetzt, wenn ein Unterprogramm schon          }
         { "forward" deklariert wurde                                   }
         was_forward : boolean;

         p : ptree;

      label
         restore;

      begin
         oldprocsym:=aktprocsym;
         oldprefix:=procprefix;
         oldconstsymtable:=constsymtable;
         oldprocinfo:=procinfo;
         proc_head;
         { Returntyp setzen }
         procinfo.retdef:=aktprocsym^.definition^.retdef;

         { Vielleicht ein Zeiger fr einen Returntyp }
         if (procinfo.retdef^.deftype=arraydef) or
            (procinfo.retdef^.deftype=stringdef) or
            (procinfo.retdef^.deftype=classdef) or
            (procinfo.retdef^.deftype=recorddef) then
           begin
              procinfo.retoffset:=procinfo.call_offset;
              inc(procinfo.call_offset,4);
           end;
         { spart Speicherplatz: }
         { param_offset mu gespeichert werden, da geschachtelte Unterprogramme }
         { procinfo ndern }
         aktprocsym^.definition^.parast^.name:=pstring(procinfo.call_offset);

         { Header ist geparst }
         if parse_only then
           goto restore;

         { wird nur wegen EXPORT hier schon durchgefhrt: }
         names.init;
         make_global:=false;
         procinfo.exported:=false;
         case token of
            _FAR : begin
                      consume(_FAR);
                      warning(far_ignored);
                      consume(SEMICOLON);
                   end;
            _NEAR : begin
                       consume(_NEAR);
                       warning(near_ignored);
                       consume(SEMICOLON);
                    end;
            _INTERRUPT : begin
                            consume(_INTERRUPT);
                            warning(interrupt_ignored);
                            consume(SEMICOLON);
                         end;
            _EXPORT : begin
                            consume(_EXPORT);
                            names.insert(realname);
                            make_global:=true;
                            procinfo.exported:=true;
                            consume(SEMICOLON);
                            if gendeffile then
                              writeln(defdatei,#9+aktprocsym^.definition^.mangledname);
                            aktprocsym^.definition^.options:=
                              aktprocsym^.definition^.options or poexports;
                            if procinfo.class<>nil then
                              error(methods_dont_be_export);
                            if lexlevel<>0 then
                              error(dont_nest_export);
                            { zwar drfen Exceptions in exportierten Unterprogrammen }
                            { eingeschaltet sein, jedoch darf natrlich keine        }
                            { Exceptionhandlingadresse auf dem Stack erwartet werden }
                            if procinfo.exceptions then
                              begin
                                 { also call_offset erniedrigen: }
                                 dec(procinfo.call_offset,4);

                                 { call_offset neu zuweisen (s.o.): }
                                 aktprocsym^.definition^.parast^.name:=pstring(procinfo.call_offset);

                                 { auch retoffset, nur falls exportierte Funktionen }
                                 { jemals strukturierte Typen zurckgeben, unwahrscheinlich }
                                 dec(procinfo.retoffset,4);
                              end;
                         end;
            _INLINE : error(inline_not_supported);
         end;
         case token of
            _FORWARD : begin
                          consume(_FORWARD);
                          consume(SEMICOLON);
                       end;
            _EXTERNAL : begin
                           consume(_EXTERNAL);
                           consume(SEMICOLON);
                        end;
            else
               begin

                  { jetzt nach identischen Definitionen suchen     }
                  { ist ein forward vorhanden, dann dieses loeschen }
                  was_forward:=false;
                  pd:=aktprocsym^.definition;
                  while (assigned(pd)) and (assigned(pd^.nextoverloaded)) do
                    begin
                       if equal_paras(aktprocsym^.definition^.para1,pd^.nextoverloaded^.para1) then
                         begin
                            if pd^.nextoverloaded^.forwarddef then
                              { sollte das eine "forward"-Defnition sein, }
                              { dann eine halbe Leiche daraus machen      }
                              { da die Defnition noch in einer Symbol-    }
                              { tabelle zum Loeschen eingetragen ist,     }
                              { kein dispose!                             }
                              begin
                                 { and not(virtual), weil virtual nicht angegeben werden mu }
                                 if ((pd^.nextoverloaded^.options and not(povirtualmethod))
                                    <>aktprocsym^.definition^.options) or
                                   not(is_equal(pd^.nextoverloaded^.retdef,aktprocsym^.definition^.retdef)) then
                                   error(header_dont_match);
                                 hd:=pd^.nextoverloaded;
                                 { erst Name ndern: }

                                 hd^.setmangledname(aktprocsym^.definition^.mangledname);

                                 { pd^.nextoverloaded aus der Liste an den Anfang }
                                 { und aktprocsym^.definition Aushaengen }

                                 pd^.nextoverloaded:=pd^.nextoverloaded^.nextoverloaded;
                                 hd^.nextoverloaded:=aktprocsym^.definition^.nextoverloaded;
                                 aktprocsym^.definition:=hd;
                                 was_forward:=true;
                              end
                            else error(same_parameters);
                         end;
                       pd:=pd^.nextoverloaded;
                    end;
                  if assigned(procinfo.class) and not(was_forward) then
                    error(header_dont_match_any_member);
                  { Funktionsattribute lesen }
                  if token=LECKKLAMMER then
                    begin
                       consume(LECKKLAMMER);
                       repeat
                         if token=_PUBLIC then
                           begin
                              consume(_PUBLIC);
                              make_global:=true;
                           end
                         else if token=ID then
                           begin
                              if pattern='ALIAS' then
                                begin
                                   consume(ID);
                                   consume(COLON);
                                   names.insert(pattern);
                                   if token=CCHAR then consume(CCHAR)
                                     else consume(CSTRING);
                                end
                              else if pattern='INTERNPROC' then
                                begin
                                   consume(ID);
                                   consume(COLON);
                                   p:=expr;
                                   do_firstpass(p);
                                   if p^.treetype<>ordconstn then
                                     fatalerror(error_in_expression);
                                   aktprocsym^.definition^.extnumber:=p^.value;
                                   aktprocsym^.definition^.options:=
                                     aktprocsym^.definition^.options or pointernproc;
                                   disposetree(p);
                                   consume(RECKKLAMMER);
                                   consume(SEMICOLON);
                                   { definiert }
                                   aktprocsym^.definition^.forwarddef:=false;
                                   goto restore;
                                end
                              else if pattern='EXTERNAL' then
                                begin
                                   consume(ID);
                                   consume(COLON);
                                   aktprocsym^.definition^.setmangledname(pattern);
                                   if token=CCHAR then consume(CCHAR)
                                     else consume(CSTRING);
                                   aktprocsym^.definition^.forwarddef:=false;
                                end
                              else if pattern='SYSTEM' then
                                begin
                                   consume(ID);
                                   aktprocsym^.definition^.options:=poclearstack;

                                   { Vermutlich keine Register retten            }
                                   { aktprocsym^.definition^.usedregisters:=$ff; }

                                   aktprocsym^.definition^.forwarddef:=false;
                                   aktprocsym^.definition^.setmangledname(realname);
                                   consume(RECKKLAMMER);
                                   consume(SEMICOLON);
                                   goto restore;
                                end;
                           end;
                         if token=COMMA then consume(COMMA)
                           else break;
                       until false;
                       consume(RECKKLAMMER);
                       consume(SEMICOLON);
                    end;
                  if procinfo.exported then
                    aktprocsym^.definition^.options:=aktprocsym^.definition^.options or
                      poclearstack;
                  oldexitlabel:=aktexitlabel;
                  oldexceptlabel:=aktexceptlabel;
                  aktexitlabel:=getlabel;
                  aktexceptlabel:=getlabel;

                  { lex. Level berechnen }
                  inc(lexlevel);
                  if lexlevel>31 then
                    error(too_much_lexlevel);

                  { break-und continuelabel zurcksetzen, }
                  { mssen aber nicht gerettet werden!    }
                  aktbreaklabel:=0;
                  aktcontinuelabel:=0;

                  { falls Objekt, Membersymboltabelle eintragen }
                  if assigned(procinfo.class) then
                    begin
                       class:=procinfo.class;
                       while assigned(class) do
                         begin
                            class^.publicsyms^.next:=symtablestack;
                            symtablestack:=class^.publicsyms;
                            class:=class^.childof;
                         end;
                    end;

                  { Symboltabellen einfgen }
                  { und die lex. Level eintragen }
                  aktprocsym^.definition^.parast^.next:=symtablestack;
                  symtablestack:=aktprocsym^.definition^.parast;
                  inc(symtablestack^.symtabletype,lexlevel);

                  aktprocsym^.definition^.localst^.next:=symtablestack;
                  symtablestack:=aktprocsym^.definition^.localst;
                  inc(symtablestack^.symtabletype,lexlevel);

                  { hier werden die Aufzaehltypen eingefuegt }
                  constsymtable:=symtablestack;

                  { temporre Variablen zurcksetzten }
                  cleartempgen;

                  pushgened:=false;

                  { keine Register bisher benutzt }
                  usedinproc:=0;
                  block;
                  dec(lexlevel);

                  { jetzt definiert }
                  aktprocsym^.definition^.forwarddef:=false;

                  aktprocsym^.definition^.usedregisters:=usedinproc;

                  stackframe:=gettempsize;

                  if procinfo.class<>nil then
                    aktentrycode.insert(gennasmrec(MOV,S_NO,tostr(procinfo.ESI_offset)+'(%ebp),%esi'));

                  if stackframe<>0 then
                    begin
                       if littlesize then
                         aktentrycode.insert(gennasmrec(ENTER,S_L,'$'+tostr(stackframe)+',$0'))
                       else
                         begin
                            aktentrycode.insert(gennasmrec(SUB,S_L,'$'+tostr(stackframe)+',%esp'));
                            aktentrycode.insert(gennasmrec(MOV,S_L,'%esp,%ebp'));
                            aktentrycode.insert(gennasmrec(PUSH,S_L,'%ebp'));
                         end;
                    end
                  else
                    begin
                       aktentrycode.insert(gennasmrec(MOV,S_L,'%esp,%ebp'));
                       aktentrycode.insert(gennasmrec(PUSH,S_L,'%ebp'));
                    end;
                  names.insert(aktprocsym^.definition^.mangledname);

                  if (aktprocsym^.definition^.owner^.symtabletype=globalsymtable) then
                    make_global:=true;
                  hs:=names.get;
                  while hs<>'' do
                    begin
                       aktentrycode.insert(gennasmrec(DIRECT,S_NO,hs+':'));
                       if make_global then
                         aktentrycode.insert(gennasmrec(DIRECT,S_NO,'.globl '+hs));
                       hs:=names.get;
                    end;

                  parasize:=aktprocsym^.definition^.parast^.datasize+procinfo.call_offset-8;

                  aktlinenr:=inputstack^.line_no;

                  if procinfo.exceptions then
                    begin
                       aktexceptcode.insert(genlasmrec(A_LABEL,aktexceptlabel));
                       if not(nostackframe) then
                         aktexceptcode.concat(gennasmrec(LEAVE,S_NO,''));
                       { exportierte Routinen enden immer nur mit RET }
                       if procinfo.exported then
                         aktexceptcode.concat(gennasmrec(RET,S_NO,''))
                       else
                          begin
                             aktexceptcode.concat(gennasmrec(ADD,S_L,'$4,%esp'));
                             aktexceptcode.concat(gennasmrec(RET,S_NO,'$'+tostr(parasize-4)));
                          end;
                    end;
                  aktexitcode.insert(genlasmrec(A_LABEL,aktexitlabel));
                  if (aktprocsym^.definition^.options and poconstructor)=0 then
                    begin
                       if procinfo.retdef<>pdef(voiddef) then
                         begin
                            if (procinfo.retdef^.deftype=grunddef) then
                              begin
                                 case pgrunddef(procinfo.retdef)^.typ of
                                    s32bit : aktexitcode.concat(gennasmrec(MOV,S_L,tostr(procinfo.retoffset)+'(%ebp),%eax'));
                                    u8bit,s8bit,uchar,bool8bit :
                                      aktexitcode.concat(gennasmrec(MOV,S_B,tostr(procinfo.retoffset)+'(%ebp),%al'));
                                    s16bit,u16bit :
                                      aktexitcode.concat(gennasmrec(MOV,S_W,tostr(procinfo.retoffset)+'(%ebp),%ax'));
                                    s64real :
                                      aktexitcode.concat(gennasmrec(FLD,S_L,tostr(procinfo.retoffset)+'(%ebp)'));
                                 end;
                              end
                             else
                               if (procinfo.retdef^.deftype=pointerdef) or
                                  (procinfo.retdef^.deftype=aufzaehldef) or
                                  (procinfo.retdef^.deftype=procdef) then
                                  aktexitcode.concat(gennasmrec(MOV,S_L,tostr(procinfo.retoffset)+'(%ebp),%eax'));
                         end
                    end
                  else
                    begin
                       { erfolgreicher Konstruktor gibt 0 in EAX zurck }
                       aktexitcode.concat(gennasmrec(A_XOR,S_L,'%eax,%eax'));
                    end;

                  aktexitcode.concat(gennasmrec(LEAVE,S_NO,''));

                  { max. 65535 Bytes Parameter wegen RET imm16 }
                  if parasize>65535 then
                    error(para_too_big);

                  { jetzt noch das RET erzeugen, mit Entfernen der }
                  { mglichen Catchadresse und der Parameter       }

                  { exportierte Routinen enden immer nur mit RET }
                  if (parasize=0) or procinfo.exported then
                    aktexitcode.concat(gennasmrec(RET,S_NO,''))
                  else aktexitcode.concat(gennasmrec(
                      RET,S_NO,'$'+tostr(parasize)));

                  if debuginfo then
                    aktexitcode.concat(gennasmrec(STABS,S_NO,'"'+
                      aktprocsym^.name+'"'+',36,0,0,'+
                      aktprocsym^.definition^.mangledname));

                  aktproccode.insertlist(@aktentrycode);
                  aktproccode.concatlist(@aktexitcode);
                  aktproccode.concatlist(@aktexceptcode);

                  mainasmlist.concatlist(@aktproccode);

                  { ... Symboltabellen entfernen }
                  symtablestack:=symtablestack^.next^.next;

                  { Klassenmember entfernen }
                  while symtablestack^.symtabletype=objectsymtable do
                    symtablestack:=symtablestack^.next;

                  aktexitlabel:=oldexitlabel;
                  aktexceptlabel:=oldexceptlabel;

                  names.done;

                  consume(SEMICOLON);
                  aktlinenr:=0;
               end;
         end;
      restore:
         constsymtable:=oldconstsymtable;
         aktprocsym:=oldprocsym;
         procprefix:=oldprefix;
         procinfo:=oldprocinfo;
      end;

    procedure block;

      begin
         repeat
           case token of
              _LABEL : label_dec;
              _CONST : const_dec;
              _TYPE : type_dec;
              _VAR : var_dec;
              _CONSTRUCTOR,_DESTRUCTOR,
              _FUNCTION,_PROCEDURE : unter_dec;
              else break;
           end;
         until false;

         { bei BEGIN der temporre Variablen setzten }
         if (symtablestack^.symtabletype and $8000)=localsymtable then
           firsttemp:=-symtablestack^.datasize
         else firsttemp:=0;

         { Platz fuer den Returnwert schaffen }
         if procinfo.retdef<>pdef(voiddef) then
           begin
              if (procinfo.retdef^.deftype=grunddef) or
                 (procinfo.retdef^.deftype=pointerdef) or
                 (procinfo.retdef^.deftype=aufzaehldef) or
                 (procinfo.retdef^.deftype=procdef) then
                begin
                   procinfo.retoffset:=gettempofsize(procinfo.retdef^.size);
                   { EAX wird auch verndert: }
                   usedinproc:=usedinproc or ($80 shr byte(R_EAX))
                end;
           end;

         befehlsblock;
      end;

    procedure loadunits;

      var
         st : psymtable;
         s : stringid;

      begin
         consume(_USES);
         repeat
           s:=pattern;
           consume(ID);
           unitlist2.insert(pattern);
           st:=readunit(s);
           refsymtable^.insert(new(punitsym,init(s,st)));
           st^.next:=symtablestack;
           symtablestack:=st;
           if token=COMMA then consume(COMMA)
             else break;
         until false;
         consume(SEMICOLON);
      end;

    procedure insertinternsyms;

      begin
         symtablestack^.insert(new(psyssym,init('CONCAT',in_concat_x)));
         symtablestack^.insert(new(psyssym,init('WRITE',in_write_x)));
         symtablestack^.insert(new(psyssym,init('WRITELN',in_writeln_x)));
         symtablestack^.insert(new(psyssym,init('ASSIGNED',in_assigned_x)));
         symtablestack^.insert(new(psyssym,init('READ',in_read_x)));
         symtablestack^.insert(new(psyssym,init('READLN',in_readln_x)));
      end;

    procedure proc_unit;

      var
         unitname : stringid;
         p : psymtable;

      begin
         if gendeffile then
           error(def_only_in_program);
         compiling_a_unit:=true;
         consume(_UNIT);
         if (compilesystem and
            not(
              ((pattern='SYSTEM') and (target=target_DOS)) or
              ((pattern='SYSOS2') and (target=target_OS2)) or
              ((pattern='SYSLINUX') and (target=target_LINUX)) or
              ((pattern='SYSWIN32') and (target=target_WIN32))
            )) or
            (length(pattern)>8) or
            (pattern<>upper(inputfile)) then
           error(ill_unit_name);
         unitname:=pattern;
         usedunits.insert(unitname);
         consume(ID);
         consume(SEMICOLON);
         consume(_INTERFACE);

         procprefix:='_'+unitname+'$$';
         parse_only:=true;

         { jetzt die globale Symboltabelle erzeugen }
         p:=new(psymtable,init(globalsymtable));
         p^.name:=stringdup(unitname);
         refsymtable:=p;

         { SYSTEM-Qualifier einfgen }
         if not(compilesystem) then
           refsymtable^.insert(new(punitsym,init('SYSTEM',symtablestack)));
         if token=_USES then loadunits;

         { ... aber hier erst einfgen }
         p^.next:=symtablestack;
         symtablestack:=p;

         constsymtable:=symtablestack;

         { beim Uebersetzen von System einige Typen direkt einfuegen: }
         if compilesystem then
           begin
              voiddef:=new(pgrunddef,init(uvoid,0,0));
              symtablestack^.insert(new(ptypesym,init('REAL',new(pgrunddef,init(s64real,0,0)))));
              symtablestack^.insert(new(ptypesym,init('POINTER',new(ppointerdef,init(voiddef)))));
              symtablestack^.insert(new(ptypesym,init('BOOLEAN',new(pgrunddef,init(bool8bit,0,1)))));
              symtablestack^.insert(new(ptypesym,init('CHAR',new(pgrunddef,init(uchar,0,255)))));
              symtablestack^.insert(new(ptypesym,init('TEXT',new(pfiledef,init(ft_text,nil)))));
              symtablestack^.insert(new(ptypesym,init('FILE',new(pfiledef,init(ft_untyped,nil)))));
              { jetzt wurde voiddef ja geaendert }
              procinfo.retdef:=voiddef;

              insertinternsyms;
           end;

         { ... und die Deklarationen parsen }
         repeat
           case token of
              _CONST : const_dec;
              _TYPE : type_dec;
              _VAR : var_dec;
              _FUNCTION,_PROCEDURE : unter_dec;
              else
                begin
                   consume(_IMPLEMENTATION);
                   break;
                end;
           end;
         until false;
         parse_only:=false;
         { statische Symboltabelle erzeugen }
         p:=new(psymtable,init(staticsymtable));
         p^.name:=stringdup(unitname);
         refsymtable:=p;

         {
           if token=_USES then loadunits;
         }

         { ... hier erst Einfgen }
         p^.next:=symtablestack;
         symtablestack:=p;

         constsymtable:=symtablestack;

         repeat
           case token of
              _CONST : const_dec;
              _TYPE : type_dec;
              _VAR : var_dec;
              _FUNCTION,_PROCEDURE,
              _CONSTRUCTOR,_DESTRUCTOR : unter_dec;
              else break;
           end;
         until false;

         { temporre Variablen zurcksetzten }
         cleartempgen;

         { sonstige Einstellungen: }
         aktexitlabel:=getlabel;
         aktexceptlabel:=getlabel;
         aktbreaklabel:=0;
         aktcontinuelabel:=0;
         pushgened:=false;
         procinfo.retdef:=voiddef;
         procinfo.exceptions:=true;
         procinfo.class:=nil;

         mainasmlist.concat(gennasmrec(DIRECT,S_NO,'.globl INIT$$'+unitname));
         mainasmlist.concat(gennasmrec(DIRECT,S_NO,'INIT$$'+unitname+':'));
         if token=_BEGIN then befehlsblock
           else consume(_END);
         consume(POINT);
         if gettempsize<>0 then
           aktentrycode.insert(gennasmrec(SUB,S_L,'$'+tostr(gettempsize)+',%esp'));
         aktentrycode.insert(gennasmrec(MOV,S_L,'%esp,%ebp'));
         aktentrycode.insert(gennasmrec(PUSH,S_L,'%ebp'));
         aktexitcode.concat(genlasmrec(A_LABEL,aktexitlabel));
         aktexitcode.concat(gennasmrec(LEAVE,S_NO,''));
         aktexitcode.concat(gennasmrec(RET,S_NO,''));
         aktexitcode.concat(genlasmrec(A_LABEL,aktexceptlabel));
         aktexitcode.concat(gennasmrec(PUSH,S_L,'$LINITEXPT'));
         aktexitcode.concat(gennasmrec(CALL,S_NO,'INITEXCEPTION'));
         aktexitcode.concat(gennasmrec(DIRECT,S_NO,'LINITEXPT:'));
         aktexitcode.concat(gennasmrec(LEAVE,S_NO,''));
         aktexitcode.concat(gennasmrec(RET,S_NO,''));

         aktproccode.insertlist(@aktentrycode);
         aktproccode.concatlist(@aktexitcode);
         aktproccode.concatlist(@aktexceptcode);

         mainasmlist.concatlist(@aktproccode);

         { Gre der statischen Daten: }
         datasize:=symtablestack^.datasize;

         { statische Symboltabelle loeschen }
         dellexlevel;

         { alle im Implementationsteil aufgefhrten Units entfernen }
         while symtablestack^.symtabletype<>globalsymtable do
           dellexlevel;

         if codegeneration then
           begin
              { globale Deklarationen schreiben }
              unitsize:=writeunitas(inputdir+unitname+'.PPU',symtablestack);
           end;

         inc(datasize,symtablestack^.datasize);
         dellexlevel;
      end;

    function findfile(const s : string) : string;

      var
         found : boolean;
         dirinfo : searchrec;
         envstring : string;

      begin
         if target=target_DOS then
           findfile:=search(s+objext,getenv('PASLIB'),found)
         else if target=target_OS2 then
           findfile:=search(s+objext,getenv('OS2LIB'),found);
         if found then
           exit;
         findfirst(unitpath+s+objext,anyfile,dirinfo);
         if doserror=0 then
           begin
              findfile:=unitpath;
              exit;
           end;
         findfile:=search(s+objext,getunitenv,found);
         if found then
           exit;
         findfile:='';
      end;

    procedure proc_program;

      var
         st : psymtable;
         programname : stringid;
         s : stringid;
         unitinits : tasmlist;

      begin
         { bei -Us Fehler erzeugen }
         if compilesystem then
           consume(_UNIT);
         compiling_a_unit:=false;
         parse_only:=false;
         programname:='';
         if token=_PROGRAM then
           begin
              consume(_PROGRAM);
              programname:=pattern;
              consume(ID);
              if token=LKLAMMER then
                begin
                   consume(LKLAMMER);
                   idlist;
                   consume(RKLAMMER);
                end;
              consume(SEMICOLON);
           end;
         { Nach den Units Hauptsymboltabelle einfgen }
         st:=new(psymtable,init(staticsymtable));

         refsymtable:=st;
         refsymtable^.insert(new(punitsym,init('SYSTEM',symtablestack)));
         case target of
            target_DOS : unitlist2.insert('SYSTEM');
            target_OS2 : unitlist2.insert('SYSOS2');
            target_LINUX : unitlist2.insert('SYSLINUX');
            target_WIN32 : unitlist2.insert('SYSWIN32');
         end;
         if token=_USES then loadunits;

         if gen_makefile then
           begin
              write(makefile,upper(inputfile)+'.S : '+upper(inputfile)+'.PP');
              s:=unitlist2.get;
              while s<>'' do
                begin
                   writeln(makefile,' \');
                   write(makefile,s+'.PPU');
                   s:=unitlist2.get;
                end;
              writeln(makefile);
              writeln(makefile);
              writeln(makefile,upper(inputfile)+'.o : '+upper(inputfile)+'.s');
              close(makefile);
           end;

         st^.next:=symtablestack;
         symtablestack:=st;
         if programname<>'' then
           symtablestack^.insert(new(pprogramsym,init(programname)));

         { ...und als constsymtable setzen }
         constsymtable:=st;

         { Hauptprogramm parsen }

         procinfo.retdef:=voiddef;
         procinfo.exceptions:=true;
         procinfo.class:=nil;
         procprefix:='';
         aktbreaklabel:=0;
         aktcontinuelabel:=0;

         aktexitlabel:=getlabel;
         aktexceptlabel:=getlabel;

         { temporre Variablen zurcksetzten }
         cleartempgen;

         pushgened:=false;

         if target=target_DOS then
           begin
              assign(linkresponse,inputdir+'LINK.RES');
              rewrite(linkresponse);
              writeln(linkresponse,'-o '+inputfile);
              writeln(linkresponse,findfile('PRT0')+'PRT0.o');
              writeln(linkresponse,inputfile+'.o');
           end;
         if target=target_OS2 then
           begin
              assign(linkresponse,inputdir+'LINK.RES');
              rewrite(linkresponse);
              writeln(linkresponse,'-o '+inputfile);
              writeln(linkresponse,inputfile+'.obj');
              if gendeffile then
                writeln(linkresponse,inputfile+'.def');
           end;
         unitinits.init;
         s:=usedunits.get;
         while s<>'' do
           begin
              unitinits.concat(gennasmrec(CALL,S_NO,'INIT$$'+s));
              if (target=target_DOS) or (target=target_OS2) then
                writeln(linkresponse,findfile(s)+s+objext);
              s:=usedunits.get;
           end;
         if (target=target_DOS) or (target=target_OS2) then
           close(linkresponse);

         block;

         aktentrycode.insertlist(@unitinits);
         if gettempsize<>0 then
           aktentrycode.insert(gennasmrec(SUB,S_L,'$'+tostr(gettempsize)+',%esp'));
         aktentrycode.insert(gennasmrec(MOV,S_L,'%esp,%ebp'));
         aktentrycode.insert(gennasmrec(PUSH,S_L,'%ebp'));
         if target=target_DOS then
           begin
              aktentrycode.insert(gennasmrec(DIRECT,S_NO,'PASCALMAIN:'));
              aktentrycode.insert(gennasmrec(DIRECT,S_NO,'.globl'#9'PASCALMAIN'));
           end;
         aktentrycode.insert(gennasmrec(DIRECT,S_NO,'_main:'));
         aktentrycode.insert(gennasmrec(DIRECT,S_NO,'.globl'#9'_main'));

         aktexceptcode.concat(genlasmrec(A_LABEL,aktexceptlabel));
         aktexceptcode.concat(gennasmrec(CALL,S_NO,'__EXIT'));
         aktexceptcode.concat(gennasmrec(LEAVE,S_NO,''));
         if target=target_DOS then
           begin
              aktexceptcode.concat(gennasmrec(ADD,S_L,'$4,%esp'));
              aktexceptcode.concat(gennasmrec(RET,S_NO,''));
           end
         else if target=target_OS2 then
           begin
              {!!!!!! Hier Exceptions fr OS/2 abfangen }
              aktexceptcode.concat(gennasmrec(RET,S_NO,''));
           end
         else if target=target_LINUX then
           begin
              {!!!!!! Hier Exceptions fr Linux abfangen }
              aktexceptcode.concat(gennasmrec(RET,S_NO,''));
           end
         else if target=target_WIN32 then
           begin
              {!!!!!! Hier Exceptions fr Win32 abfangen }
              aktexceptcode.concat(gennasmrec(RET,S_NO,''));
           end;

         aktexitcode.concat(genlasmrec(A_LABEL,aktexitlabel));
         aktexitcode.concat(gennasmrec(CALL,S_NO,'__EXIT'));
         aktexitcode.concat(gennasmrec(LEAVE,S_NO,''));
         if target=target_DOS then
           aktexitcode.concat(gennasmrec(RET,S_NO,'$4'))
         else if target=target_OS2 then
           aktexitcode.concat(gennasmrec(RET,S_NO,''))
         { !!!!!!!!!!!! Abndern: }
         else if target=target_LINUX then
           aktexitcode.concat(gennasmrec(RET,S_NO,''))
         else if target=target_WIN32 then
           aktexitcode.concat(gennasmrec(RET,S_NO,''));

         aktproccode.insertlist(@aktentrycode);
         aktproccode.concatlist(@aktexitcode);
         aktproccode.concatlist(@aktexceptcode);

         mainasmlist.concatlist(@aktproccode);

         if target=target_DOS then
           begin
              { aber nur unter DOS! }
              { im Datensegment Heap erzeugen }
              datasegment.concat(gennasmrec(
                A_GLOBAL,S_NO,'HEAP,'+tostr(heapsize)));
              constsegment.concat(gennasmrec(
                DIRECT,S_NO,'.globl HEAPSIZE'));
              constsegment.concat(gennasmrec(
                DIRECT,S_NO,'HEAPSIZE:'));
              constsegment.concat(gennasmrec(
                A_LONG,S_NO,tostr(heapsize)));
           end;
         datasize:=symtablestack^.datasize;
         while assigned(symtablestack) do
           dellexlevel;

         consume(POINT);
      end;

    function factor(getaddr : boolean) : ptree;

      var
         l : longint;
         p1,p2,p3 : ptree;
         code : word;
         pd : pdef;
         nochmal : boolean;
         sym : pvarsym;
         classh : pclassdef;
         d : double;

      procedure do_proc_call;

        begin
           { soll nur die Adresse eines }
           { Unterprogramms festgestellt werden ? }
           if not(getaddr) then
             begin
                if token=LKLAMMER then
                  begin
                     consume(LKLAMMER);
                     p1^.left:=parse_paras;
                     consume(RKLAMMER);
                  end
                else p1^.left:=nil;

                { Schon den einen ersten Durchlauf  }
                { durchfhren, da wir den Returntyp }
                { brauchen                          }
                do_firstpass(p1);
                pd:=p1^.resulttype;
             end
           else
             begin
                p1^.left:=nil;

                { vergessen wir pd }
                pd:=nil;

                { keine Postfixoperatoren }
                nochmal:=false;
             end;
        end;

      begin
         case token of
            ID       : begin
                          { Postfixoperatoren sind normal erlaubt }
                          nochmal:=true;
                          getsym(pattern,true);
                          consume(ID);
                          if (aktprocsym<>nil) and
                             (srsym^.name=aktprocsym^.name) and
                              (procinfo.retdef<>pdef(voiddef)) and
                              (token<>LKLAMMER) then
                           begin
                              p1:=genzeronode(funcretn);
                              pd:=procinfo.retdef;
                           end
                          else
                            begin
                               if srsym^.typ=unitsym then
                                 begin
                                    consume(POINT);
                                    getsymonlyin(punitsym(srsym)^.unitsymtable,pattern);
                                    consume(ID);
                                 end;
                               case srsym^.typ of
                                  varsym  : begin
                                               p1:=genloadnode(pvarsym(srsym),srsymtable);
                                               pd:=pvarsym(srsym)^.definition;
                                            end;
                                  typedconstsym :
                                            begin
                                               p1:=gentypedconstloadnode(ptypedconstsym(srsym),srsymtable);
                                               pd:=ptypedconstsym(srsym)^.definition;
                                            end;
                                  syssym :  begin
                                               p1:=anweisung_syssym(psyssym(srsym)^.number,pd);
                                            end;
                                  typesym : begin
                                               pd:=ptypesym(srsym)^.definition;
                                               consume(LKLAMMER);
                                               p1:=expr;
                                               consume(RKLAMMER);
                                               p1:=gentypeconvnode(p1,pd);
                                               p1^.explizit:=true;
                                            end;
                                  aufzaehlsym : begin
                                                  p1:=genaufzaehlnode(paufzaehlsym(srsym));
                                                  pd:=p1^.resulttype;
                                               end;
                                  constsym : begin
                                                case pconstsym(srsym)^.consttype of
                                                   constint : p1:=genordinalconstnode(pconstsym(srsym)^.value,s32bitdef);
                                                   conststring : p1:=genstringconstnode(pstring(pconstsym(srsym)^.value)^);
                                                   constchar : p1:=genordinalconstnode(pconstsym(srsym)^.value,cchardef);
                                                   constreal : p1:=genrealconstnode(pdouble(pconstsym(srsym)^.value)^);
                                                   constbool : p1:=genordinalconstnode(pconstsym(srsym)^.value,booldef);
                                                end;
                                                pd:=p1^.resulttype;
                                             end;
                                  procsym : begin
                                               p1:=gencallnode(pprocsym(srsym),srsymtable);
                                               do_proc_call;
                                            end;
                                  errorsym : begin
                                                p1:=genzeronode(errorn);
                                                pd:=generrordef;
                                                if token=LKLAMMER then
                                                  begin
                                                     consume(LKLAMMER);
                                                     parse_paras;
                                                     consume(RKLAMMER);
                                                  end;
                                             end;
                                  else
                                    begin
                                       p1:=genzeronode(errorn);
                                       pd:=generrordef;
                                       error(error_in_expression);
                                    end;
                               end;
                            end;
                          { Postfixoperatoren bearbeiten }
                          while nochmal do
                            begin
                               case token of
                                  CARET : begin
                                             consume(CARET);
                                             if pd^.deftype<>pointerdef then
                                               begin
                                                  error(invalid_qualifizier);
                                                  disposetree(p1);
                                                  p1:=genzeronode(errorn);
                                               end
                                             else
                                               begin
                                                  p1:=gensinglenode(derefn,p1);
                                                  pd:=ppointerdef(pd)^.definition;
                                               end;
                                          end;
                                  LECKKLAMMER : begin
                                                   consume(LECKKLAMMER);
                                                   repeat
                                                     if (pd^.deftype<>arraydef) and
                                                        (pd^.deftype<>stringdef) and
                                                        (pd^.deftype<>pointerdef) then
                                                       begin
                                                          error(invalid_qualifizier);
                                                          disposetree(p1);
                                                          p1:=genzeronode(errorn);
                                                       end
                                                     else if (pd^.deftype=pointerdef) then
                                                       begin
                                                          p2:=expr;
                                                          p1:=gennode(vecn,p1,p2);
                                                          pd:=ppointerdef(pd)^.definition;
                                                       end
                                                     else
                                                       begin
                                                          p2:=expr;
                                                          p1:=gennode(vecn,p1,p2);
                                                          if pd^.deftype=stringdef then
                                                            pd:=cchardef
                                                          else pd:=parraydef(pd)^.definition;
                                                       end;
                                                     if token=COMMA then consume(COMMA)
                                                       else break;
                                                   until false;
                                                   consume(RECKKLAMMER);
                                                end;
                                  POINT       : begin
                                                   consume(POINT);
                                                   case pd^.deftype of
                                                      recorddef :
                                                            begin
                                                               sym:=pvarsym(precdef(pd)^.symtable^.search(pattern));
                                                               consume(ID);
                                                               if sym=nil then
                                                                 begin
                                                                    error(ill_field);
                                                                    disposetree(p1);
                                                                    p1:=genzeronode(errorn);
                                                                 end
                                                               else
                                                                 begin
                                                                    p1:=gensubscriptnode(sym,p1);
                                                                    pd:=sym^.definition;
                                                                 end;
                                                            end;
                                                      classdef :
                                                            begin
                                                               classh:=pclassdef(pd);
                                                               sym:=nil;
                                                               while assigned(classh) do
                                                                 begin
                                                                    sym:=pvarsym(classh^.publicsyms^.search(pattern));
                                                                    srsymtable:=classh^.publicsyms;
                                                                    if assigned(sym) then
                                                                      break;
                                                                    classh:=classh^.childof;
                                                                 end;
                                                               consume(ID);
                                                               if sym=nil then
                                                                 begin
                                                                    error(ill_field);
                                                                    disposetree(p1);
                                                                    p1:=genzeronode(errorn);
                                                                 end
                                                               else
                                                                 begin
                                                                    case sym^.typ of
                                                                       procsym : begin
                                                                                    p1:=genmethodcallnode(pprocsym(sym),
                                                                                      srsymtable,p1);
                                                                                    do_proc_call;
                                                                                 end;
                                                                       else
                                                                          begin
                                                                             p1:=gensubscriptnode(sym,p1);
                                                                             pd:=sym^.definition;
                                                                          end;
                                                                    end;
                                                                 end;
                                                            end
                                                         else
                                                            begin
                                                               error(invalid_qualifizier);
                                                               disposetree(p1);
                                                               p1:=genzeronode(errorn);
                                                            end;
                                                   end;
                                                end;
                                  else
                                    begin
                                       { Procedurenvariablen }
                                       if pd^.deftype=procdef then
                                         begin
                                            consume(LKLAMMER);
                                            if (pd^.deftype<>procdef) then
                                              begin
                                                 error(invalid_qualifizier);
                                                 disposetree(p1);
                                                 p1:=genzeronode(errorn);
                                                 { bei Fehler Paramter berlesen }
                                                 parse_paras;
                                              end
                                            else
                                              begin
                                              end;
                                            consume(RKLAMMER);
                                         end
                                       else nochmal:=false;
                                    end;
                               end;
                            end;
                       end;
            INTCONST : begin
                          val(pattern,l,code);
                          if code<>0 then
                            begin
                               error(error_in_integer);
                               l:=1;
                            end;
                          consume(INTCONST);
                          p1:=genordinalconstnode(l,s32bitdef);
                       end;
            REALNUMBER : begin
                          val(pattern,d,code);
                          if code<>0 then
                            begin
                               error(error_in_real);
                               d:=1.0;
                            end;
                          consume(REALNUMBER);
                          p1:=genrealconstnode(d);
                        end;
            CSTRING : begin
                         p1:=genstringconstnode(pattern);
                         consume(CSTRING);
                      end;
            CCHAR : begin
                         p1:=genordinalconstnode(ord(pattern[1]),cchardef);
                         consume(CCHAR);
                      end;
            KLAMMERAFFE : begin
                             consume(KLAMMERAFFE);
                             p1:=factor(true);
                             p1:=gensinglenode(addrn,p1);
                          end;
            LKLAMMER : begin
                          consume(LKLAMMER);
                          p1:=expr;
                          consume(RKLAMMER);
                       end;
            PLUS     : begin
                          consume(PLUS);
                          p1:=factor(false);
                       end;
            MINUS    : begin
                          consume(MINUS);
                          p1:=factor(false);
                          p1:=gensinglenode(umminusn,p1);
                       end;
            _NOT     : begin
                          consume(_NOT);
                          p1:=factor(false);
                          p1:=gensinglenode(notn,p1);
                       end;
            _TRUE    : begin
                          consume(_TRUE);
                          p1:=genordinalconstnode(1,booldef);
                       end;
            _FALSE    : begin
                          consume(_FALSE);
                          p1:=genordinalconstnode(0,booldef);
                       end;
            _NIL      : begin
                           consume(_NIL);
                           p1:=genzeronode(niln);
                        end;
            _SELF     : begin
                           consume(_SELF);
                           if not assigned(procinfo.class) then
                             begin
                                p1:=genzeronode(errorn);
                                error(self_not_in_method);
                             end
                           else
                             p1:=genselfnode(procinfo.class);
                        end;
            else
              begin
                 p1:=genzeronode(errorn);
                 error(error_in_expression);
              end;
         end;
         factor:=p1;
      end;

    function term : ptree;

      var
         p1,p2 : ptree;

      begin
         p1:=factor(false);
         repeat
           case token of
              STAR : begin
                        consume(STAR);
                        p2:=factor(false);
                        p1:=gennode(muln,p1,p2);
                     end;
              SLASH : begin
                        consume(SLASH);
                        p2:=factor(false);
                        p1:=gennode(slashn,p1,p2);
                      end;
              _DIV : begin
                        consume(_DIV);
                        p2:=factor(false);
                        p1:=gennode(divn,p1,p2);
                     end;
              _MOD : begin
                        consume(_MOD);
                        p2:=factor(false);
                        p1:=gennode(modn,p1,p2);
                     end;
              _AND : begin
                        consume(_AND);
                        p2:=factor(false);
                        p1:=gennode(andn,p1,p2);
                     end;
              _SHL : begin
                        consume(_SHL);
                        p2:=factor(false);
                        p1:=gennode(shln,p1,p2);
                     end;
              _SHR : begin
                        consume(_SHR);
                        p2:=factor(false);
                        p1:=gennode(shrn,p1,p2);
                     end;
              else break;
           end;
         until false;
         term:=p1;
      end;

    function simpl_expr : ptree;

      var
         p1,p2 : ptree;

      begin
         p1:=term;
         repeat
           case token of
              PLUS : begin
                        consume(PLUS);
                        p2:=term;
                        p1:=gennode(addn,p1,p2);
                     end;
              MINUS : begin
                        consume(MINUS);
                        p2:=term;
                        p1:=gennode(subn,p1,p2);
                     end;
              _OR : begin
                        consume(_OR);
                        p2:=term;
                        p1:=gennode(orn,p1,p2);
                     end;
              _XOR : begin
                        consume(_XOR);
                        p2:=term;
                        p1:=gennode(xorn,p1,p2);
                     end;
              else break;
           end;
         until false;
         simpl_expr:=p1;
      end;

    function simpl2_expr : ptree;

      var
         p1,p2 : ptree;

      begin
         p1:=simpl_expr;
         repeat
           case token of
              LT : begin
                      consume(LT);
                      p2:=simpl_expr;
                      p1:=gennode(ltn,p1,p2);
                   end;
              LTE : begin
                      consume(LTE);
                      p2:=simpl_expr;
                      p1:=gennode(lten,p1,p2);
                   end;
              GT : begin
                      consume(GT);
                      p2:=simpl_expr;
                      p1:=gennode(gtn,p1,p2);
                   end;
              GTE : begin
                      consume(GTE);
                      p2:=simpl_expr;
                      p1:=gennode(gten,p1,p2);
                   end;
              EQUAL : begin
                      consume(EQUAL);
                      p2:=simpl_expr;
                      p1:=gennode(equaln,p1,p2);
                   end;
              UNEQUAL : begin
                      consume(UNEQUAL);
                      p2:=simpl_expr;
                      p1:=gennode(unequaln,p1,p2);
                   end;
              _IN : begin
                      consume(_IN);
                      p2:=simpl_expr;
                      p1:=gennode(inn,p1,p2);
                   end;
              else break;
           end;
         until false;
         simpl2_expr:=p1;
      end;

    function expr : ptree;

      var
         p1,p2 : ptree;

      begin
         p1:=simpl2_expr;
         case token of
            POINTPOINT : begin
                            consume(POINTPOINT);
                            p2:=simpl2_expr;
                            p1:=gennode(rangen,p1,p2);
                         end;
            ASSIGNMENT : begin
                            consume(ASSIGNMENT);
                            p2:=expr;
                            p1:=gennode(assignn,p1,p2);
                         end
         end;
         expr:=p1;
      end;

    function parse : longint;

      var
         p : pgrunddef;
         st : psymtable;
         pd : pdef;
         hs : string[10];
         mac : pmacrosym;

      begin
         { Zielbetriebssystem als Symbol fr bedingte Compilierung }
         { festlegen                                               }
         case target of
            target_OS2 : hs:='OS2';
            target_DOS : hs:='DOS';
         end;
         mac:=new(pmacrosym,init(hs));
         mac^.defined:=true;
         macros^.insert(mac);
         { die beiden Stringcontainer fr das Makefile erzeugen }
         unitlist1.init;
         unitlist1.doubles:=false;
         unitlist2.init;

         { Codegenerator initialisieren }
         codegeninit;

         { Scanner initilisieren }
         token:=yylex;

         forwardsallowed:=false;

         { Definitionen jetzt nicht registrieren }
         registerdef:=false;

         { Fehlerdefinition erzeugen: }
         generrordef:=new(perrordef,init);

         { Definitionen fuer Konstanten erzeugen: }

         s32bitdef:=new(pgrunddef,init(s32bit,$80000000,$7fffffff));
         cstringdef:=new(pstringdef,init(255));
         cchardef:=new(pgrunddef,init(uchar,0,255));
         cs64realdef:=new(pgrunddef,init(s64real,0,0));

         { sonstige verwendete Definitionen: }
         voiddef:=new(pgrunddef,init(uvoid,0,0));
         u8bitdef:=new(pgrunddef,init(u8bit,0,255));
         u16bitdef:=new(pgrunddef,init(u16bit,0,65535));
         booldef:=new(pgrunddef,init(bool8bit,0,1));
         voidpointerdef:=new(ppointerdef,init(voiddef));

         { erst jetzt Definitionen registrieren, da vorher symtabletack }
         { auf jeden Fall ungueltig war, was mir einige Zeit Kopfzer-   }
         { brechen verursacht hat                                       }
         registerdef:=true;

         symtablestack:=nil;

         { falls nicht SYSTEM compiliert wird, SYSTEM laden }
         if not(compilesystem) then
           begin
              case target of
                 target_DOS : symtablestack:=readunit('SYSTEM');
                 target_OS2 : symtablestack:=readunit('SYSOS2');
                 target_LINUX : symtablestack:=readunit('SYSLINUX');
                 target_WIN32 : symtablestack:=readunit('SYSWIN32');
              end;
              insertinternsyms;
           end;
         { aktueller Rckgabetyp: void }
         procinfo.retdef:=voiddef;

         { lexikalisches Level zurcksetzen }
         lexlevel:=0;

         { Quelltext parsen }
         if token=_UNIT then proc_unit
           else proc_program;
         consume(_EOF);
         if codegeneration then
           begin
              if optimiere then simpljumpopt(mainasmlist);

              if debuginfo then
                begin
                   debuginfos.insert(gennasmrec(DIRECT,S_NO,'Ltext0:'));
                   mainasmlist.concatlist(@debuginfos);
                end;

              { nun Datensegment anhngen }
              mainasmlist.concatlist(@datasegment);

              constsegment.insert(gennasmrec(DIRECT,S_NO,'.align 4'));
              constsegment.insert(gennasmrec(DIRECT,S_NO,'.data'));

              { nun Constsegment anhngen }
              mainasmlist.concatlist(@constsegment);

              { und VMT's: }
              mainasmlist.concatlist(@vmtasmlist);

              { Symboltabellenstack lschen }
              while assigned(symtablestack) do
                dellexlevel;

              { Codegenerator beenden }
              codegendone;

              { Inputpuffer entfernen }
              donescanner;

              { Jetzt erst mainasmlist schreiben, da nun mindestens }
              { sourcesize Bytes fuer den Ausgabepuffer frei sind   }

              writemainasmlist;
           end;

         parse:=datasize;
      end;

end.
